#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "cvmx.h"
#include "cvmx-sysinfo.h"
#include "cvmx-llm.h"

#define MAX_ERRORS 20

#ifdef CVMX_ABI_N32
/* Global variables that define the min/max of the memory region set up for
    32 bit userspace access. We don't use it but it is required for linking */
uint64_t linux_mem32_offset = 0;
#endif

#ifdef LLM64
    #define LLM_WRITE cvmx_llm_write_wide
    #define LLM_READ  cvmx_llm_read_wide
    #define LLM_STEP  8
#else
static uint64_t read_llm_narrow_raw(cvmx_llm_address_t addr, int sep)
{
    cvmx_llm_data_t val;
    val = cvmx_llm_read_narrow(addr, sep);
    return(val.u64);
}
    #define LLM_WRITE cvmx_llm_write_narrow
    #define LLM_READ  read_llm_narrow_raw
    #define LLM_STEP  4
#endif

#define ULL unsigned long long

/* Global variable with the processor ID since we can't read it directly */
uint32_t cvmx_app_init_processor_id;

static void setup_system_info(void)
{
    cvmx_sysinfo_t *system_info = cvmx_sysinfo_get();
    memset(system_info, 0, sizeof(cvmx_sysinfo_t));

    system_info->core_mask = 0;
    system_info->init_core = cvmx_get_core_num();

    FILE *infile = fopen("/proc/octeon_info", "r");
    if (infile == NULL)
    {
        perror("Error opening /proc/octeon_info");
        exit(-1);
    }

    while (!feof(infile))
    {
        char buffer[80];
        if (fgets(buffer, sizeof(buffer), infile))
        {
            const char *field = strtok(buffer, " ");
            const char *valueS = strtok(NULL, " ");
            if (field == NULL)
                continue;
            if (valueS == NULL)
                continue;
            unsigned long long value;
            sscanf(valueS, "%lli", &value);

            if (strcmp(field, "dram_size:") == 0)
                system_info->system_dram_size = value;
            else if (strcmp(field, "phy_mem_desc_addr:") == 0)
                system_info->phy_mem_desc_ptr = cvmx_phys_to_ptr(value);
            else if (strcmp(field, "eclock_hz:") == 0)
                system_info->cpu_clock_hz = value;
            else if (strcmp(field, "dclock_hz:") == 0)
                system_info->dram_data_rate_hz = value * 2;
            else if (strcmp(field, "spi_clock_hz:") == 0)
                system_info->spi_clock_hz = value;
            else if (strcmp(field, "board_type:") == 0)
                system_info->board_type = value;
            else if (strcmp(field, "board_rev_major:") == 0)
                system_info->board_rev_major = value;
            else if (strcmp(field, "board_rev_minor:") == 0)
                system_info->board_rev_minor = value;
            else if (strcmp(field, "chip_type:") == 0)
                system_info->chip_type = value;
            else if (strcmp(field, "chip_rev_major:") == 0)
                system_info->chip_rev_major = value;
            else if (strcmp(field, "chip_rev_minor:") == 0)
                system_info->chip_rev_minor = value;
            else if (strcmp(field, "board_serial_number:") == 0)
                strncpy(system_info->board_serial_number, valueS, sizeof(system_info->board_serial_number));
            else if (strcmp(field, "mac_addr_base:") == 0)
            {
                int i;
                int m[6];
                sscanf(valueS, "%02x:%02x:%02x:%02x:%02x:%02x", m+0, m+1, m+2, m+3, m+4, m+5);
                for (i=0; i<6; i++)
                    system_info->mac_addr_base[i] = m[i];
            }
            else if (strcmp(field, "mac_addr_count:") == 0)
                system_info->mac_addr_count = value;
            else if (strcmp(field, "processor_id:") == 0)
                cvmx_app_init_processor_id = value;
        }
    }


#if 0
    printf("system_dram_size:       %llu\n", (unsigned long long)system_info->system_dram_size);
    printf("phy_mem_desc_ptr:       %p\n", system_info->phy_mem_desc_ptr);
    printf("init_core:              %u\n", system_info->init_core);
    printf("cpu_clock_hz:           %u\n", system_info->cpu_clock_hz);
    printf("dram_data_rate_hz:      %u\n", system_info->dram_data_rate_hz);
    printf("spi_clock_hz:           %u\n", system_info->spi_clock_hz);
    printf("board_type:             %u\n", system_info->board_type);
    printf("board_rev_major:        %u\n", system_info->board_rev_major);
    printf("board_rev_minor:        %u\n", system_info->board_rev_minor);
    printf("chip_type:              %u\n", system_info->chip_type);
    printf("chip_rev_major:         %u\n", system_info->chip_rev_major);
    printf("chip_rev_minor:         %u\n", system_info->chip_rev_minor);
    printf("mac_addr_base:          %02x:%02x:%02x:%02x:%02x:%02x\n",
               (int)system_info->mac_addr_base[0],
               (int)system_info->mac_addr_base[1],
               (int)system_info->mac_addr_base[2],
               (int)system_info->mac_addr_base[3],
               (int)system_info->mac_addr_base[4],
               (int)system_info->mac_addr_base[5]);
    printf("mac_addr_count:         %u\n", system_info->mac_addr_count);
    printf("board_serial_number:    %s\n", system_info->board_serial_number);
#endif
}


int main(int argc, char *argv[])
{
    cvmx_llm_address_t addr = {0};
    uint64_t val;
    uint64_t read_back;
    uint64_t errors = 0;
    FILE *in;

    setup_system_info();
    cvmx_llm_initialize();

    if (argc != 2)
    {
        printf("Usage: %s file_to_load\n", argv[0]);
        return -1;
    }

    in = fopen(argv[1], "r");
    if (in == NULL)
    {
        perror("open");
        return -1;
    }

    while (!feof(in))
    {
        fread(&val, 1, 8, in);
        LLM_WRITE(addr, val, 0);
        read_back = LLM_READ(addr, 0);
        if (val!=read_back && errors++ < MAX_ERRORS)
        {
            printf("LLM Write failed: addr=0x%016llx correct=0x%016llx data=0x%016llx\n",
                   (ULL)addr.u64, (ULL)val, (ULL)read_back);
        }
        addr.s.address+=LLM_STEP;
    }

    addr.s.address=0;
    fseek(in, 0, 0);
    while (!feof(in))
    {
        fread(&val, 1, 8, in);
        read_back = LLM_READ(addr, 0);
        if (val!=read_back && errors++ < MAX_ERRORS)
        {
            printf("LLM Write failed: addr=0x%016llx correct=0x%016llx data=0x%016llx\n",
                   (ULL)addr.u64, (ULL)val, (ULL)read_back);
        }
        addr.s.address+=LLM_STEP;
    }
    fclose(in);

    if (errors)
    {
        printf("...additional errors omitted...\n\n\
Error: the contents of file and memory did not match (%lld read errors)\n\       
       please make sure that data in the file is formatted for the specific width:\n\
         CN38XX : 64 bit (wide), 36 bit (narrow)\n\ 
         CN31XX : 64 bit (wide w/o ECC), 56 bit (wide w/ ECC), 32 bit (narrow).\n", CAST64(errors));
    }
    return errors ? -1 : 0;
}
