/*
 * Copyright (C) 2000 Lennert Buytenhek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

#include "libbridge.h"
#include "libbridge_private.h"
#include "br_hook_defs.h"

static int getethertypebyname(char *name)
{
	char buf[256];
	FILE *fd;
	int eth_type;

	snprintf(buf, sizeof(buf), "grep %s /etc/ethertypes | awk '{print $2}'", name);
	fd = popen(buf, "r");
	if (fd) {
		memset(buf, 0, sizeof(buf));
		fscanf(fd, "%s", buf);
		if (!strcmp(buf, "")) {
			pclose(fd);
			return -1;
		}
		sscanf(buf, "%x", &eth_type);
		pclose(fd);
		return eth_type;
	}

	return -1;
}

int br_add_bridge(const char *brname)
{
	int ret;

#ifdef SIOCBRADDBR
	ret = ioctl(br_socket_fd, SIOCBRADDBR, brname);
	if (ret < 0)
#endif
	{
		char _br[IFNAMSIZ];
		unsigned long arg[3] 
			= { BRCTL_ADD_BRIDGE, (unsigned long) _br };

		strncpy(_br, brname, IFNAMSIZ);
		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
	} 

	return ret < 0 ? errno : 0;
}

int br_del_bridge(const char *brname)
{
	int ret;

#ifdef SIOCBRDELBR	
	ret = ioctl(br_socket_fd, SIOCBRDELBR, brname);
	if (ret < 0)
#endif
	{
		char _br[IFNAMSIZ];
		unsigned long arg[3] 
			= { BRCTL_DEL_BRIDGE, (unsigned long) _br };

		strncpy(_br, brname, IFNAMSIZ);
		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
	} 
	return  ret < 0 ? errno : 0;
}

int br_add_rule(char **argv)
{
	int ret = 0;
	struct br_hook_ioctl ioc;
	int ether_type;
	int i;
	struct __target_map {
		char name[256];
		int number;
	} target_map[] = {{"ADD_VLAN_TAG_AND_CONTINUE", ADD_VLAN_TAG_AND_CONTINUE}, 
				{"REM_VLAN_TAG_AND_CONTINUE", REM_VLAN_TAG_AND_CONTINUE},
				{"ACCEPT_EAPOL", ACCEPT_EAPOL},
				{"", -1}};

	if (!strcmp(argv[1], "I")) {
		ioc.position = BR_FIRST;
	} else if (!strcmp(argv[1], "A")) {
		ioc.position = BR_NEXT;
	} else {
		return -1;
	}

	if (!strcmp(argv[2], "PREROUTING")) {
		ioc.hook = BR_PREROUTING;
	} else if (!strcmp(argv[2], "POSTROUTING")) {
		ioc.hook = BR_POSTROUTING;
	} else {
		return -1;
	}

	if (!strcmp(argv[3], "ANY")) {
		sprintf(ioc.bridge, "*");
	} else {
		snprintf(ioc.bridge, IFNAMSIZ, "%s", argv[3]);
	}

	if (!strcmp(argv[4], "ANY")) {
		sprintf(ioc.indev, "*");
	} else {
		snprintf(ioc.indev, IFNAMSIZ, "%s", argv[4]);
	}

	if (!strcmp(argv[5], "ANY")) {
		sprintf(ioc.outdev, "*");
	} else {
		snprintf(ioc.outdev, IFNAMSIZ, "%s", argv[5]);
	}

	if (argv[6][0] == '!') {
		ether_type = getethertypebyname((char *)&argv[6][1]);
		if (ether_type == -1)
			return -1;
		ioc.pkt_type = 0x10000000 | (unsigned short) ether_type;
	} else {
		ether_type = getethertypebyname((char *)argv[6]);
		if (ether_type == -1)
			return -1;
		ioc.pkt_type = (unsigned short) ether_type;
	}

	i = 0;
	ioc.target = -1;
	while(strcmp(target_map[i].name, "")) {
		if (!strcmp(target_map[i].name, argv[7])) {
			ioc.target = target_map[i].number;
			break;
		}
		i++;
	}
	if (ioc.target == -1)
		return -1;

	ioc.arg = atoi(argv[8]);

	{
		unsigned long arg[3] 
			= { BRCTL_ADD_RULE, (unsigned long)&ioc };

		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
		printf("\n");
	}

	return ret < 0 ? errno : 0;
}

int br_del_rule(char **argv)
{
	int ret = 0;
	struct br_hook_ioctl ioc;
	int ether_type;
	int i;
	struct __target_map {
		char name[256];
		int number;
	} target_map[] = {{"ANY", NULL_TARGET},
				{"ADD_VLAN_TAG_AND_CONTINUE", ADD_VLAN_TAG_AND_CONTINUE}, 
				{"REM_VLAN_TAG_AND_CONTINUE", REM_VLAN_TAG_AND_CONTINUE}, 
				{"ACCEPT_EAPOL", ACCEPT_EAPOL}, 
				{"", -1}};

	if (!strcmp(argv[1], "PREROUTING")) {
		ioc.hook = BR_PREROUTING;
	} else if (!strcmp(argv[1], "POSTROUTING")) {
		ioc.hook = BR_POSTROUTING;
	} else {
		return -1;
	}

	if (!strcmp(argv[2], "ANY")) {
		sprintf(ioc.bridge, "*");
	} else {
		snprintf(ioc.bridge, IFNAMSIZ, "%s", argv[2]);
	}

	if (!strcmp(argv[3], "ANY")) {
		sprintf(ioc.indev, "*");
	} else {
		snprintf(ioc.indev, IFNAMSIZ, "%s", argv[3]);
	}

	if (!strcmp(argv[4], "ANY")) {
		sprintf(ioc.outdev, "*");
	} else {
		snprintf(ioc.outdev, IFNAMSIZ, "%s", argv[4]);
	}

	if (!strcmp(argv[5], "ANY")) {
		ioc.pkt_type = 0;
	} else if (argv[5][0] == '!') {
		ether_type = getethertypebyname((char *)&argv[5][1]);
		if (ether_type == -1)
			return -1;
		ioc.pkt_type = 0x10000000 | (unsigned short) ether_type;
	} else {
		ether_type = getethertypebyname((char *)argv[5]);
		if (ether_type == -1)
			return -1;
		ioc.pkt_type = (unsigned short) ether_type;
	}

	i = 0;
	ioc.target = -1;
	while(strcmp(target_map[i].name, "")) {
		if (!strcmp(target_map[i].name, argv[6])) {
			ioc.target = target_map[i].number;
			break;
		}
		i++;
	}
	if (ioc.target == -1)
		return -1;

	ioc.arg = atoi(argv[7]);

	{
		unsigned long arg[3] 
			= { BRCTL_DEL_RULE, (unsigned long)&ioc };

		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
		printf("\n");
	} 

	return ret < 0 ? errno : 0;
}

int br_list_rule(char **argv)
{
	int ret = 0;
	struct br_hook_ioctl ioc;
	int cookie = 0;

	if (!strcmp(argv[1], "PREROUTING")) {
		ioc.hook = BR_PREROUTING;
	} else if (!strcmp(argv[1], "POSTROUTING")) {
		ioc.hook = BR_POSTROUTING;
	} else {
		return -1;
	}

	while (cookie != -1)
	{
		unsigned long arg[3] 
			= { BRCTL_LIST_RULE, (unsigned long)&cookie, (unsigned long)&ioc };

		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
		if (ret)
			break;

		if (cookie != -1) {
			printf("%d: bridge=%s indev=%s outdev=%s pkt_type=%x target=%d arg=%d\n",
					cookie,
					ioc.bridge, ioc.indev, ioc.outdev, 
					ioc.pkt_type, ioc.target, ioc.arg);
		}
	} 

	return ret < 0 ? errno : 0;
}

int br_add_interface(const char *bridge, const char *dev)
{
	struct ifreq ifr;
	int err;
	int ifindex = if_nametoindex(dev);

	if (ifindex == 0) 
		return ENODEV;
	
	strncpy(ifr.ifr_name, bridge, IFNAMSIZ);
#ifdef SIOCBRADDIF
	ifr.ifr_ifindex = ifindex;
	err = ioctl(br_socket_fd, SIOCBRADDIF, &ifr);
	if (err < 0)
#endif
	{
		unsigned long args[4] = { BRCTL_ADD_IF, ifindex, 0, 0 };
					  
		ifr.ifr_data = (char *) args;
		err = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
	}

	return err < 0 ? errno : 0;
}

int br_del_interface(const char *bridge, const char *dev)
{
	struct ifreq ifr;
	int err;
	int ifindex = if_nametoindex(dev);

	if (ifindex == 0) 
		return ENODEV;
	
	strncpy(ifr.ifr_name, bridge, IFNAMSIZ);
#ifdef SIOCBRDELIF
	ifr.ifr_ifindex = ifindex;
	err = ioctl(br_socket_fd, SIOCBRDELIF, &ifr);
	if (err < 0)
#endif		
	{
		unsigned long args[4] = { BRCTL_DEL_IF, ifindex, 0, 0 };
					  
		ifr.ifr_data = (char *) args;
		err = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
	}

	return err < 0 ? errno : 0;
}
