/*
 *  ebt_add1q
 *
 *  Copyright (c) Arada Systems, 1996-2007
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/ether.h>
#include <getopt.h>
#include "../include/ebtables_u.h"
#include <linux/netfilter_bridge/ebt_8021q.h>

static int vlanid_supplied;

#define ADD1Q_ID	'1'
#define ADD1Q_TARGET	'2'

static struct option opts_add1q[] =
{
	{ "add1q-id", required_argument, 0, ADD1Q_ID },
	{ "add1q-target", required_argument, 0, ADD1Q_TARGET },
	{ 0 }
};

static void print_help_add1q()
{
	printf(
	"add1q target options:\n"
	" --add1q-id vlan_id  : Insert 802.1Q header with vlan_id\n"
	" --add1q-target target : ACCEPT, DROP, RETURN or CONTINUE\n");
}

static void init_add1q(struct ebt_entry_target *target)
{
	struct ebt_8021q_info *add1q_info =
	   (struct ebt_8021q_info *)target->data;

	vlanid_supplied = 0;
	add1q_info->vlan_id = 1;
	add1q_info->target = EBT_ACCEPT;
}

#define OPT_ADD1Q_ID		0x01
#define OPT_ADD1Q_TARGET	0x02

static int parse_add1q(int c, char **argv, int argc,
   const struct ebt_u_entry *entry, unsigned int *flags,
   struct ebt_entry_target **target)
{
	struct ebt_8021q_info *add1q_info =
	   (struct ebt_8021q_info *)(*target)->data;
	char *end;

	switch (c) {
	case ADD1Q_ID:
		ebt_check_option2(flags, OPT_ADD1Q_ID);
		add1q_info->vlan_id = strtoul(optarg, &end, 0);
		if (*end != '\0' || end == optarg)
			ebt_print_error("Bad VLAN value '%s'", optarg);
		if (add1q_info->vlan_id > 4094)
			ebt_print_error("Bad VLAN value '%d'", add1q_info->vlan_id);
		vlanid_supplied = 1;
                break;
	case ADD1Q_TARGET:
		ebt_check_option2(flags, OPT_ADD1Q_TARGET);
		if (FILL_TARGET(optarg, add1q_info->target))
			ebt_print_error("Illegal --add1q-target target");
		break;
	 default:
		return 0;
	}
	return 1;
}

static void final_check_add1q(const struct ebt_u_entry *entry,
   const struct ebt_entry_target *target, const char *name,
   unsigned int hookmask, unsigned int time)
{
	struct ebt_8021q_info *add1q_info =
	   (struct ebt_8021q_info *)target->data;

	if (time == 0 && vlanid_supplied == 0)
		ebt_print_error("No VLAN value supplied");
	if (BASE_CHAIN && add1q_info->target == EBT_RETURN)
		ebt_print_error("--add1q-target RETURN not allowed on base chain");
}

static void print_add1q(const struct ebt_u_entry *entry,
   const struct ebt_entry_target *target)
{
	struct ebt_8021q_info *add1q_info =
	   (struct ebt_8021q_info *)target->data;

	printf("--add1q-id %d", add1q_info->vlan_id);
	if (add1q_info->target == EBT_ACCEPT)
		return;
	printf(" --add1q-target %s", TARGET_NAME(add1q_info->target));
}

static int compare_add1q(const struct ebt_entry_target *t1,
   const struct ebt_entry_target *t2)
{
	struct ebt_8021q_info *add1q_info1 =
	   (struct ebt_8021q_info *)t1->data;
	struct ebt_8021q_info *add1q_info2 =
	   (struct ebt_8021q_info *)t2->data;

	return add1q_info1->target == add1q_info2->target &&
	   add1q_info1->vlan_id == add1q_info2->vlan_id;
}

static struct ebt_u_target add1q_target =
{
	.name		= EBT_ADD1Q_TARGET,
	.size		= sizeof(struct ebt_8021q_info),
	.help		= print_help_add1q,
	.init		= init_add1q,
	.parse		= parse_add1q,
	.final_check	= final_check_add1q,
	.print		= print_add1q,
	.compare	= compare_add1q,
	.extra_ops	= opts_add1q,
};

void _init(void)
{
	ebt_register_target(&add1q_target);
}
