/*
 *	Wireless Tools
 *
 *		Jean II - HPL 99->04
 *
 * Main code for "iwevent". This listent for wireless events on rtnetlink.
 * You need to link this code against "iwcommon.c" and "-lm".
 *
 * Part of this code is from Alexey Kuznetsov, part is from Casey Carter,
 * I've just put the pieces together...
 * By the way, if you know a way to remove the root restrictions, tell me
 * about it...
 *
 * This file is released under the GPL license.
 *     Copyright (c) 1997-2004 Jean Tourrilhes <jt@hpl.hp.com>
 */

/***************************** INCLUDES *****************************/

#include "iwlib.h"		/* Header */

#include <linux/netlink.h>
#include <linux/rtnetlink.h>

#include <getopt.h>
#include <time.h>
#include <sys/time.h>

/* Ugly backward compatibility :-( */
#ifndef IFLA_WIRELESS
#define IFLA_WIRELESS	(IFLA_MASTER + 1)
#endif /* IFLA_WIRELESS */

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include "Trap.h"
#include <pthread.h>

/****************************** TYPES ******************************/

/*
 * Static information about wireless interface.
 * We cache this info for performance reason.
 */
typedef struct wireless_iface
{
  /* Linked list */
  struct wireless_iface *	next;

  /* Interface identification */
  int		ifindex;		/* Interface index == black magic */

  /* Interface data */
  char			ifname[IFNAMSIZ + 1];	/* Interface name */
  struct iw_range	range;			/* Wireless static data */
  int			has_range;
} wireless_iface;

/**************************** VARIABLES ****************************/

/* Cache of wireless interfaces */
struct wireless_iface *	interface_cache = NULL;


/* rogue AP*/
scanned_ap apList;
rogue_trap rogueList;
char knownMac[64][18];
char scanMac[64][18];
int scanApCount;
char rogue_vap[20] = {0};
void *rge_trap(void);

/************************ RTNETLINK HELPERS ************************/
/*
 * The following code is extracted from :
 * ----------------------------------------------
 * libnetlink.c	RTnetlink service routines.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 * -----------------------------------------------
 */

struct rtnl_handle
{
	int			fd;
	struct sockaddr_nl	local;
	struct sockaddr_nl	peer;
	__u32			seq;
	__u32			dump;
};

static inline void rtnl_close(struct rtnl_handle *rth)
{
	close(rth->fd);
}

static inline int rtnl_open(struct rtnl_handle *rth, unsigned subscriptions)
{
	int addr_len;

	memset(rth, 0, sizeof(rth));

	rth->fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (rth->fd < 0) {
		perror("Cannot open netlink socket");
		return -1;
	}

	memset(&rth->local, 0, sizeof(rth->local));
	rth->local.nl_family = AF_NETLINK;
	rth->local.nl_groups = subscriptions;

	if (bind(rth->fd, (struct sockaddr*)&rth->local, sizeof(rth->local)) < 0) {
		perror("Cannot bind netlink socket");
		return -1;
	}
	addr_len = sizeof(rth->local);
	if (getsockname(rth->fd, (struct sockaddr*)&rth->local,
			(socklen_t *) &addr_len) < 0) {
		perror("Cannot getsockname");
		return -1;
	}
	if (addr_len != sizeof(rth->local)) {
		fprintf(stderr, "Wrong address length %d\n", addr_len);
		return -1;
	}
	if (rth->local.nl_family != AF_NETLINK) {
		fprintf(stderr, "Wrong address family %d\n", rth->local.nl_family);
		return -1;
	}
	rth->seq = time(NULL);
	return 0;
}

/******************* WIRELESS INTERFACE DATABASE *******************/
/*
 * We keep a few information about each wireless interface on the
 * system. This avoid to query this info at each event, therefore
 * reducing overhead.
 *
 * Each interface is indexed by the 'ifindex'. As opposed to interface
 * names, 'ifindex' are never reused (even if you reactivate the same
 * hardware), so the data we cache will never apply to the wrong
 * interface.
 * Because of that, we are pretty lazy when it come to purging the
 * cache...
 */

/*------------------------------------------------------------------*/
/*
 * Get name of interface based on interface index...
 */
static inline int
index2name(int		skfd,
	   int		ifindex,
	   char *	name)
{
  struct ifreq	irq;
  int		ret = 0;

  memset(name, 0, IFNAMSIZ + 1);

  /* Get interface name */
  irq.ifr_ifindex = ifindex;
  if(ioctl(skfd, SIOCGIFNAME, &irq) < 0)
    ret = -1;
  else
    strncpy(name, irq.ifr_name, IFNAMSIZ);

  return(ret);
}

/*------------------------------------------------------------------*/
/*
 * Get interface data from cache or live interface
 */
static struct wireless_iface *
iw_get_interface_data(int	ifindex)
{
  struct wireless_iface *	curr;
  int				skfd = -1;	/* ioctl socket */

  /* Search for it in the database */
  curr = interface_cache;
  while(curr != NULL)
    {
      /* Match ? */
      if(curr->ifindex == ifindex)
	{
	  //printf("Cache : found %d-%s\n", curr->ifindex, curr->ifname);

	  /* Return */
	  return(curr);
	}
      /* Next entry */
      curr = curr->next;
    }

  /* Create a channel to the NET kernel. Doesn't happen too often, so
   * socket creation overhead is minimal... */
  if((skfd = iw_sockets_open()) < 0)
    {
      perror("iw_sockets_open");
      return(NULL);
    }

  /* Create new entry, zero, init */
  curr = calloc(1, sizeof(struct wireless_iface));
  if(!curr)
    {
      fprintf(stderr, "Malloc failed\n");
      return(NULL);
    }
  curr->ifindex = ifindex;

  /* Extract static data */
  if(index2name(skfd, ifindex, curr->ifname) < 0)
    {
      perror("index2name");
      free(curr);
      return(NULL);
    }
  curr->has_range = (iw_get_range_info(skfd, curr->ifname, &curr->range) >= 0);
  //printf("Cache : create %d-%s\n", curr->ifindex, curr->ifname);

  /* Done */
  iw_sockets_close(skfd);

  /* Link it */
  curr->next = interface_cache;
  interface_cache = curr;

  return(curr);
}

/*------------------------------------------------------------------*/
/*
 * Remove interface data from cache (if it exist)
 */
static void
iw_del_interface_data(int	ifindex)
{
  struct wireless_iface *	curr;
  struct wireless_iface *	prev = NULL;
  struct wireless_iface *	next;

  /* Go through the list, find the interface, kills it */
  curr = interface_cache;
  while(curr)
    {
      next = curr->next;

      /* Got a match ? */
      if(curr->ifindex == ifindex)
	{
	  /* Unlink. Root ? */
	  if(!prev)
	    interface_cache = next;
	  else
	    prev->next = next;
	  //printf("Cache : purge %d-%s\n", curr->ifindex, curr->ifname);

	  /* Destroy */
	  free(curr);
	}
      else
	{
	  /* Keep as previous */
	  prev = curr;
	}

      /* Next entry */
      curr = next;
    }
}

/********************* WIRELESS EVENT DECODING *********************/
/*
 * Parse the Wireless Event and print it out
 */

/*------------------------------------------------------------------*/
/*
 * Dump a buffer as a serie of hex
 * Maybe should go in iwlib...
 * Maybe we should have better formatting like iw_print_key...
 */
static char *
iw_hexdump(char *		buf,
	   size_t		buflen,
	   const unsigned char *data,
	   size_t		datalen)
{
  size_t	i;
  char *	pos = buf;

  for(i = 0; i < datalen; i++)
    pos += snprintf(pos, buf + buflen - pos, "%02X", data[i]);
  return buf;
}

/*------------------------------------------------------------------*/
/*
 * Print one element from the scanning results
 */
static inline int
print_event_token(struct iw_event *	event,		/* Extracted token */
		  struct iw_range *	iw_range,	/* Range info */
		  int			has_range)
		  
{
  char		buffer[128];	/* Temporary buffer */
  char		buffer2[30];	/* Temporary buffer */
  char *	prefix = (IW_IS_GET(event->cmd) ? "New" : "Set");
  /* Now, let's decode the event */
  switch(event->cmd)
    {
    case IWEVREGISTERED:
        {
          static int count=0;
	  struct iwreq      wrq;
          int skfd;
	  
	  if((skfd = iw_sockets_open()) < 0) {
              perror("iw_sockets_open");
              return(NULL);
          }
	  if(count < 64) {
              iw_saether_ntop(&event->u.addr, buffer);
              printf("count is less than one");             
              if(buffer != NULL)
           	  count++;    
          }      
          else {
             stat_Trap max_trap;
 	     double freq;
  	     char *delim = NULL;
             char buffer[128];

             struct sockaddr *sap =  (struct sockaddr *) &(event->u.addr);
             struct ether_addr *eth = (const struct ether_addr *) sap->sa_data;
               max_trap.maxStatMacAddr[0]   = eth->ether_addr_octet[0];
               max_trap.maxStatMacAddr[1]   = eth->ether_addr_octet[1];
               max_trap.maxStatMacAddr[2]   = eth->ether_addr_octet[2];
               max_trap.maxStatMacAddr[3]   = eth->ether_addr_octet[3];
               max_trap.maxStatMacAddr[4]   = eth->ether_addr_octet[4];
               max_trap.maxStatMacAddr[5]   = eth->ether_addr_octet[5];
        
               /*Get frequency */
  	       memset(&wrq, 0, sizeof(wrq));
               strncpy(wrq.ifr_name, rogue_vap, IFNAMSIZ);
               printf("wrq.ifr_name = %s\n",wrq.ifr_name);

               if (ioctl(skfd, SIOCGIWFREQ, &wrq) >= 0) {
                   freq = iw_freq2float(&(wrq.u.freq));
	      
		   iw_print_freq(buffer, sizeof(buffer), freq, -1, wrq.u.freq.flags);
                   printf("iw_print_freq = %s\n",buffer);
		   delim = strchr(buffer,':');
                   delim++;
                   delim = strtok(delim," "); /*to separate the channel info*/
                   if(strstr(delim,"2."))
                      max_trap.maxStationBand = 0;
                   else if(strstr(delim,"5."))
                      max_trap.maxStationBand = 1;
               }
               Trap_MaxStation_Req(&max_trap);
	   }
        }
        break;

    case IWEVEXPIRED:
            printf("print Expired node:%s\n",
	    iw_saether_ntop(&event->u.addr, buffer));
        break;
   
    case IWEVCUSTOM:
        {
          char custom[IW_CUSTOM_MAX+1];
          struct sockaddr * sap;
	  memset(custom, '\0', sizeof(custom));
          if(event->u.data.length == 0)
                printf("length is 0");

          if(event->u.data.pointer == NULL)
                printf("pointer is NULL");

          if ((event->u.data.pointer) && (event->u.data.length))
            {
                printf("Inside if\n");
                memcpy(custom, event->u.data.pointer, event->u.data.length);
                custom[event->u.data.length] = '\0';
            }
          printf("Custom driver event:%s\n", custom);
          switch(event->u.data.flags)
               {
                case IWEVCUST_STA_DENY:
                   {
                    acl_Trap ctrl_list;
                    struct iwreq      wrq;
                    struct wireless_info info;
                    memset((char *) &info, 0, sizeof(struct wireless_info));
                    int skfd;
                    char essid[IW_ESSID_MAX_SIZE+1]; 
		    memset(essid, '\0', sizeof(essid));
                    sap = (struct sockaddr *)custom;
                    printf("under IWEVCUST_STA_DENY ACL  \n");
      
		    if((skfd = iw_sockets_open()) < 0) {
      			  perror("iw_sockets_open");
      		          return(NULL);
                    }

                    memset(&ctrl_list, 0, sizeof(ctrl_list));

                    ctrl_list.aclMsMatchMacAddr[0] = sap->sa_data[0];
                    ctrl_list.aclMsMatchMacAddr[1] = sap->sa_data[1];
                    ctrl_list.aclMsMatchMacAddr[2] = sap->sa_data[2];
                    ctrl_list.aclMsMatchMacAddr[3] = sap->sa_data[3];
                    ctrl_list.aclMsMatchMacAddr[4] = sap->sa_data[4];
                    ctrl_list.aclMsMatchMacAddr[5] = sap->sa_data[5];
	
             	    memset(&wrq, 0, sizeof(wrq));
		    strncpy(wrq.ifr_name, rogue_vap, IFNAMSIZ);
                    printf("wrq.ifr_name = %s\n",wrq.ifr_name);

		    if (ioctl(skfd, SIOCGIWAP, &wrq) >= 0) {
                          memcpy(&(info.ap_addr), &(wrq.u.ap_addr), sizeof (sockaddr));

                          ctrl_list.aclMsMatchBssid[0] = info.ap_addr.sa_data[0];
                          ctrl_list.aclMsMatchBssid[1] = info.ap_addr.sa_data[1];
                          ctrl_list.aclMsMatchBssid[2] = info.ap_addr.sa_data[2];
                          ctrl_list.aclMsMatchBssid[3] = info.ap_addr.sa_data[3];
                          ctrl_list.aclMsMatchBssid[4] = info.ap_addr.sa_data[4];
                          ctrl_list.aclMsMatchBssid[5] = info.ap_addr.sa_data[5];
                     }

		    memset(&wrq, 0, sizeof(wrq));
		    strncpy(wrq.ifr_name,rogue_vap, IFNAMSIZ);
                    wrq.u.essid.pointer = (caddr_t)essid; 
                    wrq.u.essid.length = IW_ESSID_MAX_SIZE + 1;

                    if (ioctl(skfd, SIOCGIWESSID, &wrq) >= 0 ) {
                      strncpy(ctrl_list.aclMsMatchSsid,  essid, strlen(essid));
                      ctrl_list.aclMsMatchSsid[strlen(essid)] = '\0';
                    }

                    Trap_Aclmismatch_Req(&ctrl_list);

                    }
                    break;
                case IWEVCUST_BEACON_STUCK:
                      {
                        bcon_Trap bcon_stuck;
                        memset(&bcon_stuck, 0, sizeof(bcon_stuck));
                        struct iwreq wrq;
	                int skfd;
			struct wireless_info info;
                        memset((char *) &info, 0, sizeof(struct wireless_info));
                        if((skfd = iw_sockets_open()) < 0) {
                             perror("iw_sockets_open");
                             return(NULL);
                        }
 		        memset(&wrq, 0, sizeof(wrq));
                        strncpy(wrq.ifr_name,"wifi0vap0" , IFNAMSIZ);
                        printf("wrq.ifr_name = %s\n",wrq.ifr_name);

                        if (ioctl(skfd, SIOCGIWAP, &wrq) >= 0) {
                             memcpy(&(info.ap_addr), &(wrq.u.ap_addr), sizeof (sockaddr));
                           bcon_stuck.noBeaconBssid[0] = info.ap_addr.sa_data[0];
                           bcon_stuck.noBeaconBssid[1] = info.ap_addr.sa_data[1];
                           bcon_stuck.noBeaconBssid[2] = info.ap_addr.sa_data[2];
                           bcon_stuck.noBeaconBssid[3] = info.ap_addr.sa_data[3];
                           bcon_stuck.noBeaconBssid[4] = info.ap_addr.sa_data[4];
         		   bcon_stuck.noBeaconBssid[5] = info.ap_addr.sa_data[5];
                        }
                        Trap_noBeacon_Req(&bcon_stuck);
                      }
                      break;
                } 

            }
   default:
      printf("(Unknown Wireless event 0x%04X)\n", event->cmd);
    }	/* switch(event->cmd) */

  return(0);
}

/*------------------------------------------------------------------*/
/*
 * Print out all Wireless Events part of the RTNetlink message
 * Most often, there will be only one event per message, but
 * just make sure we read everything...
 */
static inline int
print_event_stream(int		ifindex,
		   char *	data,
		   int		len)
{
  struct iw_event	iwe;
  struct stream_descr	stream;
  int			i = 0;
  int			ret;
  char			buffer[64];
  struct timeval	recv_time;
  struct timezone	tz;
  struct wireless_iface *	wireless_data;

  /* Get data from cache */
  wireless_data = iw_get_interface_data(ifindex);
  if(wireless_data == NULL)
    return(-1);
  strcpy(rogue_vap,wireless_data->ifname);
  printf("vap name is = %s",rogue_vap);
  /* Print received time in readable form */
  gettimeofday(&recv_time, &tz);
  iw_print_timeval(buffer, sizeof(buffer), &recv_time, &tz);

  iw_init_event_stream(&stream, data, len);
  do
    {
      /* Extract an event and print it */
      ret = iw_extract_event_stream(&stream, &iwe,
				    wireless_data->range.we_version_compiled);
      if(ret != 0)
	{
	  if(i++ == 0)
	    printf("%s   %-8.16s ", buffer, wireless_data->ifname);
	  else
	    printf("                           ");
	  if(ret > 0)
	    print_event_token(&iwe,
			      &wireless_data->range, wireless_data->has_range);
	  else
	    printf("(Invalid event)\n");
	  /* Push data out *now*, in case we are redirected to a pipe */
	  fflush(stdout);
	}
    }
  while(ret > 0);

  return(0);
}

/*********************** RTNETLINK EVENT DUMP***********************/
/*
 * Dump the events we receive from rtnetlink
 * This code is mostly from Casey
 */

/*------------------------------------------------------------------*/
/*
 * Respond to a single RTM_NEWLINK event from the rtnetlink socket.
 */
static int
LinkCatcher(struct nlmsghdr *nlh)
{
  struct ifinfomsg* ifi;

#if 0
  fprintf(stderr, "nlmsg_type = %d.\n", nlh->nlmsg_type);
#endif

  ifi = NLMSG_DATA(nlh);

  /* Code is ugly, but sort of works - Jean II */

  /* If interface is getting destoyed */
  if(nlh->nlmsg_type == RTM_DELLINK)
    {
      /* Remove from cache (if in cache) */
      iw_del_interface_data(ifi->ifi_index);
      return 0;
    }

  /* Only keep add/change events */
  if(nlh->nlmsg_type != RTM_NEWLINK)
    return 0;

  /* Check for attributes */
  if (nlh->nlmsg_len > NLMSG_ALIGN(sizeof(struct ifinfomsg)))
    {
      int attrlen = nlh->nlmsg_len - NLMSG_ALIGN(sizeof(struct ifinfomsg));
      struct rtattr *attr = (void *) ((char *) ifi +
				      NLMSG_ALIGN(sizeof(struct ifinfomsg)));

      while (RTA_OK(attr, attrlen))
	{
	  /* Check if the Wireless kind */
	  if(attr->rta_type == IFLA_WIRELESS)
	    {
	      /* Go to display it */
	      print_event_stream(ifi->ifi_index,
				 (char *) attr + RTA_ALIGN(sizeof(struct rtattr)),
				 attr->rta_len - RTA_ALIGN(sizeof(struct rtattr)));
	    }
	  attr = RTA_NEXT(attr, attrlen);
	}
    }

  return 0;
}

/* ---------------------------------------------------------------- */
/*
 * We must watch the rtnelink socket for events.
 * This routine handles those events (i.e., call this when rth.fd
 * is ready to read).
 */
static inline void
handle_netlink_events(struct rtnl_handle *	rth)
{
  while(1)
    {
      struct sockaddr_nl sanl;
      socklen_t sanllen = sizeof(struct sockaddr_nl);

      struct nlmsghdr *h;
      int amt;
      char buf[8192];

      amt = recvfrom(rth->fd, buf, sizeof(buf), MSG_DONTWAIT, (struct sockaddr*)&sanl, &sanllen);
      if(amt < 0)
	{
	  if(errno != EINTR && errno != EAGAIN)
	    {
	      fprintf(stderr, "%s: error reading netlink: %s.\n",
		      __PRETTY_FUNCTION__, strerror(errno));
	    }
	  return;
	}

      if(amt == 0)
	{
	  fprintf(stderr, "%s: EOF on netlink??\n", __PRETTY_FUNCTION__);
	  return;
	}

      h = (struct nlmsghdr*)buf;
      while(amt >= (int)sizeof(*h))
	{
	  int len = h->nlmsg_len;
	  int l = len - sizeof(*h);

	  if(l < 0 || len > amt)
	    {
	      fprintf(stderr, "%s: malformed netlink message: len=%d\n", __PRETTY_FUNCTION__, len);
	      break;
	    }

	  switch(h->nlmsg_type)
	    {
	    case RTM_NEWLINK:
	    case RTM_DELLINK:
	      LinkCatcher(h);
	      break;
	    default:
#if 0
	      fprintf(stderr, "%s: got nlmsg of type %#x.\n", __PRETTY_FUNCTION__, h->nlmsg_type);
#endif
	      break;
	    }

	  len = NLMSG_ALIGN(len);
	  amt -= len;
	  h = (struct nlmsghdr*)((char*)h + len);
	}

      if(amt > 0)
	fprintf(stderr, "%s: remnant of size %d on netlink\n", __PRETTY_FUNCTION__, amt);
    }
}

/**************************** MAIN LOOP ****************************/

/* ---------------------------------------------------------------- */
/*
 * Wait until we get an event
 */
static inline int
wait_for_event(struct rtnl_handle *	rth)
{
#if 0
  struct timeval	tv;	/* Select timeout */
#endif
printf("\n Wait for Event : \n");
  /* Forever */
  while(1)
    {
      fd_set		rfds;		/* File descriptors for select */
      int		last_fd;	/* Last fd */
      int		ret;

      /* Guess what ? We must re-generate rfds each time */
      FD_ZERO(&rfds);
      FD_SET(rth->fd, &rfds);
      last_fd = rth->fd;

      /* Wait until something happens */
      ret = select(last_fd + 1, &rfds, NULL, NULL, NULL);

      /* Check if there was an error */
      if(ret < 0)
	{
	  if(errno == EAGAIN || errno == EINTR)
	    continue;
	  fprintf(stderr, "Unhandled signal - exiting...\n");
	  break;
	}

      /* Check if there was a timeout */
      if(ret == 0)
	{
	  continue;
	}

      /* Check for interface discovery events. */
      if(FD_ISSET(rth->fd, &rfds))
	handle_netlink_events(rth);
    }

  return(0);
}

/******************************* MAIN Thread ************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
void *
catch_event(void* unused)
{
  struct rtnl_handle	rth;
  pthread_t thrd_id;
  pthread_create (&thrd_id, NULL, &rge_trap, NULL);

  /* Open netlink Socket */
  if(rtnl_open(&rth, RTMGRP_LINK) < 0)
    {
      perror("Can't initialize rtnetlink socket");
      return(1);
    }

  fprintf(stderr, "Waiting for Wireless Events from interfaces...\n");

  /* Waiting for Events from Interfaces */
  wait_for_event(&rth);

  /* close the Netlink Socket */
  rtnl_close(&rth);

  return(0);
}

/*int getKnownApList(void)
{
    struct db_binding *bind = NULL;
    char *err = NULL;
    char name[64]="";
    int flags;
    dbcom_handle state;
    int i;
    int count = 0;
    int rv = 0;
    flags = DBCOM_ITERATE_CHILDREN;
    for(i=0; i<1; i++) {
       sprintf(name,"%s%d","system:apList:knownApTable:wlan",i);
       rv = dbcom_iterate (state,name,flags,&bind,&err);
    if(rv)   {	
       if(err) {	
           snmp_log(LOG_ERR,"\nerr %s\n",err);
           free(err);
       }
       if(bind)
           free(bind); 
       snmp_log(LOG_ERR, "Error in reading the database\n");  
       return -1;
     }
       while(bind != NULL) {
          snmp_log(LOG_ERR,"known mac address %s\n",bind->name);
          strncpy (knownMac[++count],bind->name, strlen(bind->name));		
          bind = bind->next;    
       }
    }
    return count;
}*/
int getKnownApList(void)
{
    #define BUFLEN 512

    FILE *fp;
    char temp[BUFLEN];
    char temp1[BUFLEN];
    char mac[BUFLEN];
    char *delim = NULL;
    int i=0, j;
    int count = 0;

    fp = fopen("/var/config","r");
    if (fp == NULL) {
        printf ("Unable to get the known aplist");
        return -1;
    }
    for(j=0; j<1; j++) {
		sprintf(temp,"%s%d:","system:apList:knownApTable:wlan",j);	
    	while( fgets(temp1, BUFLEN, fp))    {
        	if( strstr (temp1, temp))   {
            	i = 0;
            	delim = strrchr (temp1, ':');
            	delim++;
            	memset (mac, '\0', BUFLEN);
				strncpy (mac,delim, strlen(delim));
            	while (mac[i] != ' ')  {
                	i++;
            	}
	        mac[i] = '\0';
     	   strncpy (knownMac[++count], mac, strlen(mac));
        	}
    	}
	}
    fclose(fp);
    return count;
} 

int fillapList(struct iw_event *event, struct iw_range* iw_range, int has_range)
{
//printf("Inside fillapList\n");
  switch(event->cmd)
    {
	case SIOCGIWAP: 
          {
           int i =0 ;
   	   char buffer[128];
#if 0
           struct wireless_info info;
           memset((char *) &info, 0, sizeof(struct wireless_info));
                    memcpy(&(info.ap_addr), &(event->u.ap_addr), sizeof (sockaddr));
           for(i=0;i<6;i++) 
                          rogueList.rogue_MacAddr[i] = info.ap_addr.sa_data[i];
#endif          
 
   	   strncpy(rogueList.rogue_MacAddr, iw_pr_ether(buffer, event->u.ap_addr.sa_data),
           strlen(iw_pr_ether(buffer, event->u.ap_addr.sa_data)) );
             {
               int i;
               for(i=0; i<strlen(rogueList.rogue_MacAddr); i++)    {
                    if(islower(rogueList.rogue_MacAddr[i]))
                         rogueList.rogue_MacAddr[i] = toupper(rogueList.rogue_MacAddr[i]);
                         //printf("strlen of mac %d", strlen(rogueList.rogue_MacAddr));
                     }
               }
    }
    break;
        case SIOCGIWESSID:
        {
           char essid[IW_ESSID_MAX_SIZE+1];
           memset(essid, '\0', sizeof(essid));

           if((event->u.essid.pointer) && (event->u.essid.length)) {
               memcpy(essid, event->u.essid.pointer, event->u.essid.length);
           }
           essid[event->u.essid.length] = '\0';
           if(event->u.essid.flags)        {
               if(event->u.essid.length == 0)
                   strcpy(rogueList.rogue_Ssid,"0");
               else
					strncpy(rogueList.rogue_Ssid, essid, strlen(essid));
					rogueList.rogue_Ssid[strlen(essid)]='\0';
                                        snmp_log(LOG_ERR,"\n SIOCGIWESSID %s\n",rogueList.rogue_Ssid); 
				}
		}   
		break;
		case SIOCGIWFREQ:
		{
			double freq;
			int channel = -1;
			char *delim = NULL;
			char buffer[128];

			freq = iw_freq2float(&(event->u.freq));
			if(has_range)
				channel = iw_freq_to_channel(freq, iw_range);
				iw_print_freq(buffer, sizeof(buffer),
					freq, channel, event->u.freq.flags);
				delim = strchr(buffer,':');
				delim++;
				delim = strtok(delim," "); /*to separate the channel info*/
				if(strstr(delim,"2."))
					rogueList.rogue_band = 0;
				else if(strstr(delim,"5.")) 
					rogueList.rogue_band = 1;
					rogueList.rogue_ch = channel;
		}
		break;
		case SIOCGIWENCODE:
		{
			unsigned char   key[IW_ENCODING_TOKEN_MAX];

			if(event->u.data.pointer)
				memcpy(key, event->u.essid.pointer, event->u.data.length);
			else
				event->u.data.flags |= IW_ENCODE_NOKEY;
			if(event->u.data.flags & IW_ENCODE_DISABLED) {
				rogueList.rogue_prv = 0; /*privacy disabled */
			}
			else {
				rogueList.rogue_prv = 1; /*privacy enabled */
				printf("\n");
			}
		}
		break;
		case IWEVCUSTOM:
		{
			char custom[IW_CUSTOM_MAX+1];
			if((event->u.data.pointer) && (event->u.data.length))	{
				memcpy(custom, event->u.data.pointer, event->u.data.length);
			}
			custom[event->u.data.length] = '\0';
			if(event->u.data.flags == IWEVCUST_BCN_INT)	{
				char *delim = NULL;
				snmp_log(LOG_ERR,"beacon interval %s",custom);
				delim = strchr(custom,'=');
				delim++;
				rogueList.rogue_beacon = atoi(delim);
				snmp_log(LOG_ERR,"beacon interval %d",rogueList.rogue_beacon );
			}
			else if(event->u.data.flags == IWEVCUST_BCN_COUNT) {
				char *delim = NULL;
				delim = strchr(custom,'=');
				delim++;
				rogueList.rogue_numbeacon = atoi(delim);
				snmp_log(LOG_ERR,"beacon count %d",rogueList.rogue_numbeacon );
			}
			else if(event->u.data.flags == IWEVCUST_BCN_TIME)	{
				char *delim = NULL;
				delim = strchr(custom,'=');
				delim++;
				strncpy(rogueList.rogue_lastbeacon, delim, strlen(delim));
				snmp_log(LOG_ERR,"beacon time %s",rogueList.rogue_lastbeacon );
			}
			else if(event->u.data.flags == IWEVCUST_MAX_RATE)	{
				char *delim = NULL;
				snmp_log(LOG_ERR,"max rate %s\n",custom);
				delim = strchr(custom,'=');
				delim++;
				rogueList.rogue_rate = atoi(delim);
			}
        }
        break;
        default:
		   printf("");
	}
	return 1;	
}


int rogueApScan(void)
{
    struct iwreq wrq;    
    char *buffer;
    struct timeval tv;             /* Select timeout */
    int timeout = 5000000;      /* 5s */
	struct iw_range	range;
    int	has_range;
    int skfd;
#if 0
    (void) memset(&wrq, 0, sizeof(wrq));
    (void) strncpy(wrq.ifr_name, "wifi0vap0", sizeof(wrq.ifr_name)); 

    skfd = socket(AF_INET, SOCK_DGRAM, 0);
	if(skfd < 0)	{
	 	perror ("Error in opening the socket\n");
	    return -1;
	}
#endif
     if((skfd = iw_sockets_open()) < 0) {
             perror("iw_sockets_open");
             return(NULL);
     }

         memset(&wrq, 0, sizeof(wrq));
         strncpy(wrq.ifr_name, rogue_vap, IFNAMSIZ);
         printf("wrq.ifr_name under rogueApScan = %s\n",wrq.ifr_name);

    /* Init timeout value -> 250ms*/
    tv.tv_sec = 0;
    tv.tv_usec = 250000;
    wrq.u.param.flags = IW_SCAN_DEFAULT;
    wrq.u.param.value = 0;          
    if(ioctl(skfd, SIOCSIWSCAN, &wrq) < 0)	{
        printf("\n Start Scanning \n");
		if(errno != EPERM)	{ 
          fprintf (stderr, "Inteface does not support scanning \n");
		  close(skfd);		  
		  skfd = -1;
		  return -1;
		}
		tv.tv_usec = 0;	
	}
	timeout -= tv.tv_usec;
	buffer = (char *) calloc(IW_SCAN_MAX_DATA, sizeof(char));
	if(buffer == NULL)
	{	
		fprintf (stderr, "Not enough memory\n");
		close(skfd);		  
		skfd = -1;
		return -1;
	}
	/* Forever */
    while(1)
    {
// printf("Inside While LOOP\n");
        fd_set		rfds;		/* File descriptors for select */
        int		last_fd;	/* Last fd */
        int		ret;

        FD_ZERO(&rfds);
        last_fd = -1;

        /* Wait until something happens */
        ret = select(last_fd + 1, &rfds, NULL, NULL, &tv);

        /* Check if there was an error */
        if(ret < 0)	{
             if(errno == EAGAIN || errno == EINTR)
	    	    continue;
             fprintf(stderr, "Unhandled signal - exiting...\n");
             free(buffer);
             return(-1);
        }

        if(ret == 0)	{
             /* Try to read the results */
             wrq.u.data.pointer = buffer;
             wrq.u.data.flags = 0;
             wrq.u.data.length = IW_SCAN_MAX_DATA;
             if(ioctl(skfd, SIOCGIWSCAN, &wrq) < 0)  {
                  /* Bad error */
                  return(-2);
	         }
	         else
	             /* We have the results, go to process them */
	             break;
	     }
    }
    if(wrq.u.data.length)	 {			
        struct iw_event		iwe;
        struct stream_descr	stream;
        struct iwscan_state	state = { .ap_num = 1, .val_index = 0 };
        int ret;
        has_range = (iw_get_range_info(skfd, rogue_vap, &range) >= 0);
        iw_init_event_stream(&stream, buffer, wrq.u.data.length);
        do	{
             /* Extract an event and print it */
             ret = iw_extract_event_stream(&stream, &iwe, range.we_version_compiled);
             if(ret > 0) {
                 //printf("calling fillapList \n");
                 fillapList(&iwe, &range, has_range);
                 if(iwe.u.data.flags == IWEVCUST_BCN_TIME)	{
                    int i;
                    for(i=0; i<scanApCount; i++)
                    {
                        if(!strcmp(scanMac[i], rogueList.rogue_MacAddr))
                            break;	
                    }	
                    if(i == scanApCount)  {
	                unsigned short c[12] = {0};
			unsigned char buff[20] = {0};
                        int j , k = 0;
                        strcpy(scanMac[scanApCount++],rogueList.rogue_MacAddr);
                        snmp_log(LOG_ERR,"Rogue mac address %s\n %d\n",rogueList.rogue_MacAddr,strlen(rogueList.rogue_MacAddr));		
                        snmp_log(LOG_ERR,"Rogue ssid %s\n",rogueList.rogue_Ssid);
                        /* parse the mac address */
                        strcpy(buff,rogueList.rogue_MacAddr);
                        snmp_log(LOG_ERR,"mac Addr Rogue Trap = %s\n",buff); 
                        for(j=0;j <= strlen(buff);j++)
                           {
                                if (buff[j] == ':')
                                   continue;
                              if(buff[j] >= '0' && buff[j] <= '9')  
                                  c[k] = buff[j] - '0';
                              if(buff[j] >= 'a' && buff[j] <= 'f')
                                  c[k] = buff[j] - 'a' + 10;
                              if(buff[j] >= 'A' && buff[j] <= 'F')
                                  c[k] = (buff[j] - 'A') + 10;
                             k++;
                           }
                          k = 0;
                       for (j =0; j <= 12; j++)
                          {
                             rogueList.rogue_MacAddr[k] =  (c[j++] & 0x0F) << 4;
                             rogueList.rogue_MacAddr[k] |= (c[j] & 0x0F);
                             k++;
                          }
                             rogueList.rogue_MacAddr[6] = '\0';
                        Trap_Rogue_Req(&rogueList);
                    }
			  }
	       }
	} while(ret > 0);
    printf("\n");
   } 
    else  {
        free(buffer);
	    close(skfd);
	    return -1;
    }
        free(buffer);	
        close(skfd);	
        skfd = -1;
        return 0;
}

void * rge_trap() 
    {
  printf("\n ;Entered RogueTrap \n");
      while(1) {
        sleep(300);
        struct iwreq wrq;
        int skfd;
        struct wireless_info info;
        memset((char *) &info, 0, sizeof(struct wireless_info));
        char          buffer[128];    /* Temporary buffer */
        if((skfd = iw_sockets_open()) < 0) {
             perror("iw_sockets_open");
             return(NULL);
        }
          printf("\n rogue_vap  = %s\n",rogue_vap);
         memset(&wrq, 0, sizeof(wrq));
         strncpy(wrq.ifr_name, rogue_vap, IFNAMSIZ);
         printf("wrq.ifr_name 1 = %s\n",wrq.ifr_name);

         if (ioctl(skfd, SIOCGIWAP, &wrq) >= 0)
                   {
                    int i, j;
                    int knownApCount;
                    printf(" SIOCGIWAP \n");
                    strncpy(apList.scan_MacAddr, iw_pr_ether(buffer, wrq.u.ap_addr.sa_data),
                          strlen(iw_pr_ether(buffer, wrq.u.ap_addr.sa_data)) );
                    {
 		        printf("\n Inside SIOCGIWAP \n");
                        int i;
                        for(i=0; i<strlen(apList.scan_MacAddr); i++)    {
                             if(islower(apList.scan_MacAddr[i])) {
                                apList.scan_MacAddr[i] = toupper(apList.scan_MacAddr[i]);
                               printf("mac\n");
                             }
                        }
                    }
                    i = 0;
                    if((knownApCount = getKnownApList()) < 0)
                        snmp_log(LOG_ERR,"No knownAps in the database\n");
                    else        {
                        for(i=0; i<knownApCount; i++)
                        {
                               if(!strncmp(apList.scan_MacAddr, knownMac[i], strlen(apList.scan_MacAddr))) {
                                  snmp_log(LOG_ERR,"Don't call Rogueap Trap\n");
                                  break;
                               }
                        }
                        if(i == knownApCount )
                        {
                              printf("\n calling rogueAp \n");
                              rogueApScan();
                        }
                    }
                 }
      }
    } 
