/*************************************************************************
Cavium Octeon Ethernet Driver

Copyright (c) 2003-2005, Cavium Networks. All rights reserved.

This Software is the property of Cavium Networks.  The Software and all
accompanying documentation are copyrighted.  The Software made available
here constitutes the proprietary information of Cavium Networks.  You
agree to take reasonable steps to prevent the disclosure, unauthorized use
or unauthorized distribution of the Software.  You shall use this Software
solely with Cavium hardware.

Except as expressly permitted in a separate Software License Agreement
between You and Cavium Networks, you shall not modify, decompile,
disassemble, extract, or otherwise reverse engineer this Software.  You
shall not make any copy of the Software or its accompanying documentation,
except for copying incident to the ordinary and intended use of the
Software and the Underlying Program and except for the making of a single
archival copy.

This Software, including technical data, may be subject to U.S.  export
control laws, including the U.S.  Export Administration Act and its
associated regulations, and may be subject to export or import regulations
in other countries.  You warrant that You will comply strictly in all
respects with all such regulations and acknowledge that you have the
responsibility to obtain licenses to export, re-export or import the
Software.

TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
AND WITH ALL FAULTS AND CAVIUM MAKES NO PROMISES, REPRESENTATIONS OR
WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT
TO THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY
REPRESENTATION OR DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT
DEFECTS, AND CAVIUM SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES
OF TITLE, MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR
PURPOSE, LACK OF VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET
POSSESSION OR CORRESPONDENCE TO DESCRIPTION.  THE ENTIRE RISK ARISING OUT
OF USE OR PERFORMANCE OF THE SOFTWARE LIES WITH YOU.

*************************************************************************/
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/init.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/string.h>
#include <linux/ethtool.h>
#include <linux/mii.h>
#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include <net/dst.h>
#ifdef CONFIG_XFRM
#include <linux/xfrm.h>
#include <net/xfrm.h>
#endif  /* CONFIG_XFRM */

#undef OCTEON_MODEL
#define USE_RUNTIME_MODEL_CHECKS 1

#define printf printk
#include "cvmx.h"
#include "cvmx-atomic.h"
#include "cvmx-ciu.h"
#include "cvmx-pip.h"
#include "cvmx-ipd.h"
#include "cvmx-pko.h"
#include "cvmx-gmx.h"
#include "cvmx-spi.h"
#include "cvmx-bootmem.h"

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
#include <linux/tcp.h>
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
#include "cvmx-app-init.h"

#include "cavium-ethernet.h"

#ifndef CONFIG_CAVIUM_RESERVE32
#define CONFIG_CAVIUM_RESERVE32 0
#endif

/*
 * A few defines are used to control the operation of this driver:
 *  CONFIG_CAVIUM_RESERVE32
 *      This kernel config options controls the amount of memory configured
 *      in a wired TLB entry for all processes to share. If this is set, the
 *      driver will use this memory instead of kernel memory for pools. This
 *      allows 32bit userspace application to access the buffers, but also
 *      requires all received packets to be copied.
 *  CONFIG_CAVIUM_OCTEON_NUM_PACKET_BUFFERS
 *      This kernel config option allows the user to control the number of
 *      packet and work queue buffers allocated by the driver. If this is zero,
 *      the driver uses the default from below.
 *  NUM_PACKET_BUFFERS
 *      This define is set based on CONFIG_CAVIUM_OCTEON_NUM_PACKET_BUFFERS. It
 *      is used by the driver to determine the number of packet and work queue
 *      buffers.
 *  USE_SKBUFFS_IN_HW
 *      Tells the driver to populate the packet buffers with kernel skbuffs.
 *      This allows the driver to receive packets without copying them. It also
 *      means that 32bit userspace can't access the packet buffers.
 *  USE_32BIT_SHARED
 *      This define tells the driver to allocate memory for buffers from the
 *      32bit sahred region instead of the kernel memory space.
 *  USE_OPTIMIZED_PKO_BUFFER_SIZE
 *      Set the size of the PKO command buffers to an odd number of 64bit
 *      words. This allows the normal two word send to stay aligned and never
 *      span a comamnd word buffer. It is faster but incompatable with the
 *      three word variant. This MUST be used with the linux32-packet-io
 *      example.
 *  USE_HW_TCPUDP_CHECKSUM
 *      Controls if the Octeon TCP/UDP checksum engine is used for packet
 *      output. If this is zero, the kernel will perform the checksum in
 *      software.
 *  USE_MULTICORE_RECEIVE
 *      Process receive interrupts on multiple cores. This spreads the network
 *      load across the first 8 processors. If ths is zero, only one core
 *      processes incomming packets.
 *  FIX_IPD_PTR_ALIGNMENT
 *      Staggers PKT_PCNT and WQE_PCNT in the IPD to be optimal.
 *
 */
#if CONFIG_CAVIUM_RESERVE32
    #define USE_32BIT_SHARED        1
    #define USE_SKBUFFS_IN_HW       0
#else
    #define USE_32BIT_SHARED        0
    #define USE_SKBUFFS_IN_HW       1
#endif

#if CONFIG_CAVIUM_OCTEON_NUM_PACKET_BUFFERS
    #define NUM_PACKET_BUFFERS      CONFIG_CAVIUM_OCTEON_NUM_PACKET_BUFFERS
#else
    #define NUM_PACKET_BUFFERS      1024
#endif

#define INTERRUPT_LIMIT             10000   /* Max interrupts per second per core */
//#define INTERRUPT_LIMIT             0       /* Don't limit the number of interrupts */
#define USE_OPTIMIZED_PKO_BUFFER_SIZE 1
#define USE_HW_TCPUDP_CHECKSUM      1
#define USE_MULTICORE_RECEIVE       1
#define USE_RED                     1	/* Enable Random Early Dropping under load */
#define USE_10MBPS_PREAMBLE_WORKAROUND 1    /* Allow SW based preamble removal at 10Mbps to workaround PHYs giving us bad preambles */
#define DONT_WRITEBACK(x)           (x) /* Use this to have all FPA frees also tell the L2 not to write data to memory */
//#define DONT_WRITEBACK(x)         0   /* Use this to not have FPA frees control L2 */
#define IP_PROTOCOL_TCP             6
#define IP_PROTOCOL_UDP             0x11
#define FAU_NUM_PACKET_BUFFERS_TO_FREE (CVMX_FAU_REG_END - sizeof(uint32_t))
#define TOTAL_NUMBER_OF_PORTS       (CVMX_PIP_NUM_INPUT_PORTS+1)

#define INTERFACE(port) (port >> 4) /* Ports 0-15 are interface 0, 16-31 are interface 1 */
#define INDEX(port) (port & 0xf)
#define DEBUGPRINT(format, ...) do{if (__printk_ratelimit(HZ, 10)) printk(format, ##__VA_ARGS__);} while (0)

#ifndef CONFIG_SMP
#undef USE_MULTICORE_RECEIVE
#define USE_MULTICORE_RECEIVE 0
#endif

#ifdef CONFIG_CAVIUM_OCTEON_SIMULATOR
#define FIX_IPD_PTR_ALIGNMENT 0
#else
#define FIX_IPD_PTR_ALIGNMENT 1
#endif

#ifdef CONFIG_WN302T_DUAL_BAND_AP 
#define WN302T_PHY_FIX 1

int first_phy_id = 0;
module_param(first_phy_id, int, 0444);
MODULE_PARM_DESC(first_phy_id, "\n"
                 "\t\tPHY device id to use for MDIO read and write operation.\n");

#endif


int num_ethernet_devices = 32;
module_param(num_ethernet_devices, int, 0444);
MODULE_PARM_DESC(num_ethernet_devices, "\n"
                 "\t\tThe maximum number of ethernet ports to setup for Linux.\n"
                 "\t\tSet to zero to ignore all SPI and RGMII ethernet ports.\n"
                 "\t\tPorts not configured for Linux may be used with the\n"
                 "\t\tsimple executive. The purpose of this paramter is to allow\n"
                 "\t\tsetups where Linux configures some ports, but leaves others\n"
                 "\t\tfor alternative setups for other applications. Normally\n"
                 "\t\tyou want the kernel to do all the config, and have the\n"
                 "\t\tapplication tweak it afterwards. In this case, don't specify\n"
                 "\t\tthis paramter.");

int pow_receive_group = 15;
module_param(pow_receive_group, int, 0444);
MODULE_PARM_DESC(pow_receive_group, "\n"
                 "\t\tPOW group to receive packets from. All ethernet hardware\n"
                 "\t\twill be configured to send incomming packets to this POW\n"
                 "\t\tgroup. Also any other software can submit packets to this\n"
                 "\t\tgroup for the kernel to process.");

int pow_send_group = -1;
module_param(pow_send_group, int, 0644);
MODULE_PARM_DESC(pow_send_group, "\n"
                 "\t\tPOW group to send packets to other software on. This\n"
                 "\t\tcontrols the creation of the virtual device pow0.\n"
                 "\t\talways_use_pow also depends on this value.");

int always_use_pow = 0;
module_param(always_use_pow, int, 0444);
MODULE_PARM_DESC(always_use_pow, "\n"
                 "\t\tWhen set, always send to the pow group. This will cause\n"
                 "\t\tpackets sent to real ethernet devices to be sent to the\n"
                 "\t\tPOW group instead of the hardware. Unless some other\n"
                 "\t\tapplication changes the config, packets will still be\n"
                 "\t\treceived from the low level hardware. Use this option\n"
                 "\t\tto allow a CVMX app to intercept all packets from the\n"
                 "\t\tlinux kernel. You must specify pow_send_group along with\n"
                 "\t\tthis option.");

char pow_send_list[128] = "";
module_param_string(pow_send_list, pow_send_list, sizeof(pow_send_list), 0444);
MODULE_PARM_DESC(pow_send_list, "\n"
                 "\t\tComma separated list of ethernet devices that should use the\n"
                 "\t\tPOW for transmit instead of the actual ethernet hardware. This\n"
                 "\t\tis a per port version of always_use_pow. always_use_pow takes\n"
                 "\t\tprecedence over this list. For example, setting this to\n"
                 "\t\t\"eth2,spi3,spi7\" would cause these three devices to transmit\n"
                 "\t\tusing the pow_send_group.");

int skip_spi_interfaces = 0;
module_param(skip_spi_interfaces, int, 0444);
MODULE_PARM_DESC(skip_spi_interfaces, "\n"
                 "\t\tIf this parameter is set to 1, SPI-4 interfaces will not be configured\n"
                 "\t\tby the module. This mode is provided for testing purposes and to allow\n"
                 "\t\ta separate application / OS to manage SPI-4 interfaces.");

/**
 * Exported from the kernel so we can determine board information. It is
 * passed by the bootloader to the kernel.
 */
extern cvmx_bootinfo_t *octeon_bootinfo;

static struct net_device *cvm_oct_device[TOTAL_NUMBER_OF_PORTS];
static struct timer_list cvm_oct_poll_timer;/* Periodic timer to check RGMII auto negotiation */
static struct tasklet_struct cvm_oct_tasklet[NR_CPUS];
static spinlock_t global_register_lock;

static void cvm_oct_configure_rgmii_speed(struct net_device *dev);
static int cvm_oct_rx_poll(struct net_device *dev, int *budget);

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
/**
 * =============================================
 * = Cavium IP forwarding/NAT offload support. =
 * =============================================
 *
 * This feature is enabled by defining CVM_IPFWD_DRIVER_OFFLOAD
 * and setting ipfwd_driver_offload module parameter to 1. 
 *
 * Once enabled, the "active" TCP/UDP flows are forwarded (and
 * NATted) in the ethernet driver. 
 *
 */

/**
 * Global enable/disable for IP forwarding/NAT offload.
 *    0 = disable
 *    1 = enable
 */
int ipfwd_driver_offload = 1;
module_param(ipfwd_driver_offload, int, 0666);

#if 0
int ipfwd_slow_path_packets[16] = {0};
module_param_array(ipfwd_slow_path_packets, int, NULL, 0444);

int ipfwd_slow_path_refresh[16] = {0};
module_param_array(ipfwd_slow_path_refresh, int, NULL, 0444);
#endif

/**
 * =============================
 * IPFWD/NAT OFFLOAD DEFINITIONS
 * =============================
 */

#if 0
#define CVM_IPFWD_LUT_NAME "cvm_ipfwd_lut\0"
#endif
#define CVM_IPFWD_NOT_FOUND -1          /* Return code for "not found" */

#define CVM_IPFWD_LUT_BUCKETS 512       /* Number of hash buckets in the flow LUT (must be power of 2) */
#define CVM_IPFWD_LUT_MASK (CVM_IPFWD_LUT_BUCKETS - 1)
#define CVM_IPFWD_ENTRIES_PER_BUCKET 5	/* Currently, bucket is 5*24B + 8B header = 128B */
#define CVM_IPFWD_ENTRY_RETRIES 5	/* Number of retries for reads */

#define CVM_IPFWD_TIMER_TICK_SHFT       27  /* 1 timer tick is about .25 seconds @ 500MHz */
#define CVM_IPFWD_TIMER_GET_TIME        ((cvmx_get_cycle())>>CVM_IPFWD_TIMER_TICK_SHFT)
#define CVM_IPFWD_TIMER_NEW_LIFETIME    50
#define CVM_IPFWD_TIMER_OLD_LIFETIME    16  
#define CVM_IPFWD_TIMER_TOTAL_LIFETIME  (CVM_IPFWD_TIMER_NEW_LIFETIME + CVM_IPFWD_TIMER_OLD_LIFETIME)

#define CVM_IPFWD_EXPIRED 0         /* "Expired" entry state */
#define CVM_IPFWD_NEW     1         /* "New" entry state */
#define CVM_IPFWD_OLD     2         /* "Old" entry state */

#define CVM_IPFWD_PTR_FIFO_SIZE 16  /* Size of the buffer free queue -- must be a power of two */
#define CVM_IPFWD_PTR_FIFO_MASK (CVM_IPFWD_PTR_FIFO_SIZE - 1) 

#define CVM_IPFWD_RND_SHIFT 1       /* cycle counter shift count (for random_select()) */
#define CVM_IPFWD_RND_MSK   0x1F    /* cycle counter AND mask (for random_select()) */
#define CVM_IPFWD_RND_VAL   1       /* cycle counter compare value (for random_select()) */

#define CVM_IPFWD_COOKIE_VALUE ~0x12345ull  /* Packet info magic value */

#ifdef CVM_IPFWD_DYNAMIC_SHORT
#define CVM_IPFWD_IP_OFFSET 16
#endif /* CVM_IPFWD_DYNAMIC_SHORT */

/**
 *
 * IPFWD/NAT OFFLOAD DATA STRCUTURES
 *
 */

/* Ethernet header structure */
typedef struct {
    uint32_t enet_dst_hi;
    uint16_t enet_dst_lo;
    uint16_t enet_src_hi;
    uint32_t enet_src_lo;
    uint16_t enet_type;
} cvm_ipfwd_eh_t;

/* Flow entry structure (24 bytes) in the hash bucket */
typedef union {
    struct {
        uint64_t word0;
        uint64_t word1;
        uint64_t word2;
    } u;

    struct {
        /* Word 0 */
        uint8_t          id0       : 2;   /* synchronization id for word 0*/
        uint8_t          reserved0 : 6;
        uint8_t          ip_proto;        /* orig IP protocol */
        uint16_t         l4_sport;        /* orig L4 src port */
        uint32_t         ip_saddr;        /* orig IP src addr */

        /* Word 1 */
        uint16_t         id1       : 2;   /* synchronization id for word 1*/
        uint16_t         reserved1 : 14;
        uint16_t         l4_dport;        /* orig L4 dst port */
        uint32_t         ip_daddr;        /* orig IP dst addr */

        /* Word 2 */
        uint32_t         id2       : 2;   /* synchronization id for word 2*/
        uint32_t         reserved2 : 2;
        uint32_t         timestamp : 28;  /* lifetime timestamp */
        uint32_t         flow_info;       /* packet info ptr */
    } s;
} cvm_ipfwd_flow_entry_t;

/* Bucket header structure (64 bits) */
typedef struct {
    spinlock_t           wr_lock;        /* bucket lock */
    uint16_t             reserved;
    uint16_t             timestamp;      /* bucket timetamp */
} cvm_ipfwd_flow_bucket_header_t;

/* Hash bucket (i.e., flow LUT entry) */
typedef struct {
    cvm_ipfwd_flow_bucket_header_t hdr;
    cvm_ipfwd_flow_entry_t         entry[CVM_IPFWD_ENTRIES_PER_BUCKET];
} cvm_ipfwd_flow_bucket_t;

/* Flow info structure (pointed to by flow_info ptr in the flow entry
 * structure) */
typedef struct {
    uint32_t          enet_dst_hi;  /* new ethernet dst addr */
    uint16_t          enet_dst_lo;
    uint16_t          enet_src_hi;  /* new ethernet src addr */
    uint32_t          enet_src_lo;
    uint32_t          ip_saddr;     /* new ip src addr */
    uint32_t          ip_daddr;     /* new ip dst addr */
    uint8_t           ip_proto;     /* new ip protocol */
    uint32_t          ports;        /* new L4 ports */
    struct net_device *dev;
} cvm_ipfwd_flow_info_t;

/* Packet info structure (stored at the bottom of the packet buffer) */
typedef struct {
    uint64_t                reserved0;
    uint64_t                cookie;        /* magic number */
    cvm_ipfwd_flow_bucket_t *bucket;       /* bucket ptr */
    uint32_t                seq;           /* TCP sequence number (if present) */
    uint32_t                ack_seq;       /* TCP acknowledge number (if present) */
    uint32_t                reserved1; 
    struct iphdr            ip;            /* IP header */
    uint64_t                reserved2[4];  /* L4 ports and reserved */
} cvm_packet_info_t;

/**
 *
 * IPFWD/NAT GLOBAL VARIABLES
 *
 */

/* Shared pointer to the forwarding cache */
static cvm_ipfwd_flow_bucket_t *cvm_gbl_ipfwd_cache = NULL;

typedef struct {
    uint64_t res0 : 48;
    uint64_t idx  : 16;
    void     *cvm_ipfwd_ptr_fifo[CVM_IPFWD_PTR_FIFO_SIZE];
    uint8_t  res1[256 - (sizeof(uint64_t) + (sizeof(void *) * CVM_IPFWD_PTR_FIFO_SIZE))];
} cvm_ipfwd_fifo_t;

/* Per-core queue for freed flow_info buffers */
static cvm_ipfwd_fifo_t free_fifo[NR_CPUS] = {{0}};

/**
 *
 * IPFWD/NAT OFFLOAD INLINE FUNCTIONS
 *
 */

/**
 * Helper function to calculate IPv4 header checksum.
 * 
 * @param *ip pointer to the beginning of IP header
 * 
 * @return 16-bit one's complement IPv4 checksum.
 * No alignment requirements.
 */
static inline uint16_t cvm_ipfwd_calculate_ip_header_checksum(uint16_t *ip)
{
    uint64_t sum;
    uint16_t *ptr = ip;
    uint8_t *bptr = (uint8_t*) ip;

    sum  = ptr[0];
    sum += ptr[1];
    sum += ptr[2];
    sum += ptr[3];
    sum += ptr[4];
    // Skip checksum field
    sum += ptr[6];
    sum += ptr[7];
    sum += ptr[8];
    sum += ptr[9];

    // Check for options
    if (cvmx_unlikely(bptr[0] != 0x45)) goto slow_cksum_calc;

return_from_slow_cksum_calc:

    sum = (uint16_t) sum + (sum >> 16);
    sum = (uint16_t) sum + (sum >> 16);
    return ((uint16_t) (sum ^ 0xffff));

slow_cksum_calc:
    // Add IPv4 options into the checksum (if present)
    {
        uint64_t len = (bptr[0] & 0xf) - 5;
        ptr = &ptr[len<<1];

        while (len-- > 0) {
            sum += *ptr++;
            sum += *ptr++;
        }
    }

    goto return_from_slow_cksum_calc;
}

/**
 * Helper function to allocate flow_info buffer 
 *
 * @return pointer from the FPA pool (or NULL if pool is empty)
 */
static inline void *cvm_ipfwd_ptr_alloc(void)
{ 
    void *ptr = cvmx_fpa_alloc(CVMX_FPA_FLOW_INFO_POOL);
    return (ptr);
}

/**
 * Helper function to free flow_info buffers.
 *
 * Buffers are freed into a queue and eventually freed to
 * the FPA pool.
 *
 * @param *ptr flow_info buffer pointer to be freed
 *
 */
static inline void cvm_ipfwd_ptr_free(void *ptr)
{
    uint64_t coreid = cvmx_get_core_num();
    cvm_ipfwd_fifo_t *fifo = &free_fifo[coreid];
    void **tmpptr = &fifo->cvm_ipfwd_ptr_fifo[fifo->idx];

    if (*tmpptr)
        cvmx_fpa_free(*tmpptr, CVMX_FPA_FLOW_INFO_POOL, 0);

    *tmpptr = ptr;

    /* increment free fifo ptr */
    fifo->idx = (fifo->idx+1) & CVM_IPFWD_PTR_FIFO_MASK;
}

/**
 * Helper function to flush the freeing fifo.
 *
 */
static void cvm_ipfwd_ptr_flush(void)
{
    int i,j;
    cvm_ipfwd_fifo_t *fifo;
    void **tmpptr = NULL;

    for (i=0;i<NR_CPUS;i++)
    {
        fifo = &free_fifo[i];
        fifo->idx = 0;
        for (j=0;j<CVM_IPFWD_PTR_FIFO_SIZE;j++)
        {
            tmpptr = &fifo->cvm_ipfwd_ptr_fifo[j];
            if (*tmpptr)
                cvmx_fpa_free(*tmpptr, CVMX_FPA_FLOW_INFO_POOL, 0);
            *tmpptr = NULL;
        }
    }
}

/**
 * Helper function to check if the timestamp was just created 
 *
 * @param created timestamp to check
 *
 * @return 1: true -- 0: false
 *
 */
static inline int cvm_ipfwd_check_timestamp_just_created(uint64_t created)
{
    if (created > (CVM_IPFWD_TIMER_GET_TIME-4))
        return 1;
    else
        return 0;
}

/**
 * Helper function to check the timestamp.
 *
 * @param created timestamp to check
 *
 * @return 
 *    CVM_IPFWD_EXPIRED - timestamp is expired
 *    CVM_IPFWD_OLD     - timestamp is old (=not new and close to expiration)
 *    CVM_IPFWD_NEW     - timestamp is new
 */
static inline int cvm_ipfwd_check_timestamp(uint64_t created)
{
    uint64_t now;

    now = CVM_IPFWD_TIMER_GET_TIME;

    if (now < created)
        return CVM_IPFWD_EXPIRED;	/* Wrap around -- treat as expired */

    if (now < (created + CVM_IPFWD_TIMER_NEW_LIFETIME))
        return CVM_IPFWD_NEW;    

    if (now < (created + CVM_IPFWD_TIMER_TOTAL_LIFETIME))
        return CVM_IPFWD_OLD;

    return CVM_IPFWD_EXPIRED;
}

/**
 * Function to randomly return TRUE/FALSE.
 */
static inline int cvm_ipfwd_random_select(void)
{
    return (((cvmx_get_cycle() >> CVM_IPFWD_RND_SHIFT) & CVM_IPFWD_RND_MSK) == CVM_IPFWD_RND_VAL);
}

/**
 * Fast-path cache look-up function.
 *
 * Search a cache bucket to find an entry that matches
 * input packet's 5-tuple.
 *
 * @param *bucket ptr to hash bucket
 * @param *ip ptr to IP header of the input packet 
 * @param *entry flow_entry structure to be populated (output)
 *
 * @return index of the matching entry [0 .. (CVM_IPFWD_ENTRIES_PER_BUCKET-1)]
 * @return CVM_IPFWD_NOT_FOUND entry not found
 *
 */
static inline int cvm_ipfwd_find_entry(cvm_ipfwd_flow_bucket_t *bucket,  /* ptr to hash bucket (input) */
        struct iphdr* ip,                 /* ptr to IP header (input) */
        cvm_ipfwd_flow_entry_t *entry)    /* ptr to flow entry (output) */
{
    uint32_t saddr = ip->saddr, daddr = ip->daddr;
    uint32_t ports = *(((uint32_t*) ip) + ip->ihl);
    int count, idx;
    uint8_t  proto = ip->protocol;

    for (idx = 0; idx < CVM_IPFWD_ENTRIES_PER_BUCKET; idx++)
    {
        count = CVM_IPFWD_ENTRY_RETRIES;

        /* Read current entry */
        do {
            *entry = bucket->entry[idx];
        } while ((--count >= 0) && (entry->s.id0 != entry->s.id2));

        if (count >= 0)
        {
            /* Test current entry -- compare 5-tuple */
            if ((saddr != entry->s.ip_saddr) || (daddr != entry->s.ip_daddr) ||
                    (proto != entry->s.ip_proto) ||
                    (ports != ((entry->s.l4_sport << 16) | entry->s.l4_dport)))
                continue;

            /* Found */
            /* Prefetch flow info */
            if (entry->s.flow_info)
                CVMX_PREFETCH0(cvmx_phys_to_ptr((uint64_t)entry->s.flow_info<<7));
            return idx;
        }
    }

    /* If we get here, entry was not found */
    return CVM_IPFWD_NOT_FOUND;
}


/**
 *
 * IPFWD/NAT OFFLOAD FUNCTIONS
 *
 */

/**
 * Slow-path: Check bucket header timestamp to see whether bucket can be
 * updated. 
 *
 * @param *bucket ptr to hash bucket
 *
 * @return 0: this bucket cannot be updated
 * @return 1: this bucket can be updated
 */
static int cvm_ipfwd_check_bucket_timestamp(cvm_ipfwd_flow_bucket_t *bucket)
{
    uint64_t old = bucket->hdr.timestamp;
    uint64_t new = (cvmx_get_cycle() >> 16) & 0xffff;

    if ((new - old) < 8) 
        return 0;

    /* Time to punt packet to Linux --
     * update bucket header timestamp (without locking.) */
    *(uint16_t*)&bucket->hdr.timestamp = new;
    CVMX_SYNCWS;
    return 1;
}

/**
 * Slow-path: Check bucket timestamp.
 *
 * @param *bucket ptr to hash bucket
 *
 */
static void cvm_ipfwd_mark_packet(struct sk_buff *skb, cvmx_wqe_t *work, cvm_ipfwd_flow_bucket_t *bucket)
{
    struct iphdr *ip = (struct iphdr*) (skb->data + work->word2.s.ip_offset);
    struct tcphdr *tcp;
    cvm_packet_info_t *info = (cvm_packet_info_t *) ((((uint64_t) skb->end +
                    sizeof(struct skb_shared_info)) + 7) & ~7ull);

    //printk("cvm_ipfwd_mark_packet(skb=%p, skb->end=%p, bucket=%p, info=%p)\n", skb, skb->end, bucket, info);

    info->bucket = bucket;
    info->cookie = CVM_IPFWD_COOKIE_VALUE;
    info->ip.saddr = ip->saddr;
    info->ip.daddr = ip->daddr;
    info->ip.protocol = ip->protocol;
    info->ip.tot_len = ip->tot_len;
    info->ip.ihl = ip->ihl;
    ((uint32_t*)&info->ip)[ip->ihl] = ((uint32_t*)ip)[ip->ihl];

    /* If TCP, cache also sequence number and acknowlegment number */
    if (ip->protocol == IP_PROTOCOL_TCP)
    {
        tcp = (struct tcphdr *) ((uint32_t*)ip+ip->ihl);
        info->seq = tcp->seq;
        info->ack_seq = tcp->ack_seq;
    }
}

/**
 * Walks through all entries in a bucket.  Returns an index to the
 * oldest entry and populates a flow_entry structure.
 *
 * Note! Bucket must be locked prior to calling this function 
 *
 * @param *bucket bucket pointer
 * @param *entry pointer to a flow_entry structure (output).  The
 *               matching flow_entry will be copied into this structure.
 * @return idx matching entry index (within bucket)
 */
static int cvm_ipfwd_get_free_entry(cvm_ipfwd_flow_bucket_t *bucket, cvm_ipfwd_flow_entry_t *entry)
{
    int idx = 0, i;
    uint64_t min = ~0ull;
    cvm_ipfwd_flow_entry_t *entry_ptr;

    for (i=0;i<CVM_IPFWD_ENTRIES_PER_BUCKET;i++)
    {
        entry_ptr = &bucket->entry[i];
        if (entry_ptr->s.timestamp < min)
        {
            min = entry_ptr->s.timestamp;
            idx = i;
        }
    }

    /* Copy found entry */
    *entry = bucket->entry[idx];

    return idx;
}

/**
 * Format flow entry fields, based on an IP packet.
 *
 * Note! Bucket must be locked prior to calling this function 
 *
 * @param *ip pointer to an IP header
 * @param *entry pointer to a flow entry
 * @return none
 */
static void cvm_ipfwd_setup_entry(struct iphdr *ip, cvm_ipfwd_flow_entry_t *entry)
{
    uint8_t id = entry->s.id0 + 1;  /* Cache current id count */
    uint32_t ports;

    /* Free flow_info buf if one exists */
    if (entry->s.flow_info)
        cvm_ipfwd_ptr_free(cvmx_phys_to_ptr((uint64_t)entry->s.flow_info<<7));

    /* Clear entry */
    entry->u.word0 = 0;
    entry->u.word1 = 0;
    entry->u.word2 = 0;

    entry->s.id0 = id;
    entry->s.id1 = id;
    entry->s.id2 = id;
    entry->s.ip_proto = ip->protocol;
    entry->s.ip_saddr = ip->saddr;
    entry->s.ip_daddr = ip->daddr;
    ports = *((uint32_t*)ip + ip->ihl);

    entry->s.l4_sport = ports >> 16;
    entry->s.l4_dport = ports & 0xffff;
}

/**
 * Setup flow info structure.  Copies Ethernet dst/src, IP src/dst/proto,
 * and L4 ports from a packet pointed to by the skb into the flow info
 * structure.
 *
 * @param *info pointer to flow info structure
 * @param *skb  pointer to sk_buff
 * @param *dev  pointer to output device structure
 * @return none
 */
static void cvm_ipfwd_setup_flow_info(cvm_ipfwd_flow_info_t *info, struct sk_buff *skb, struct net_device *dev)
{
    cvm_ipfwd_eh_t *eh = (cvm_ipfwd_eh_t *) skb->data;
    struct iphdr *ip;

    info->enet_src_hi = eh->enet_src_hi;
    info->enet_src_lo = eh->enet_src_lo;
    info->enet_dst_hi = eh->enet_dst_hi;
    info->enet_dst_lo = eh->enet_dst_lo;

    ip = (struct iphdr*) (skb->data + 14);
    info->ip_saddr = ip->saddr;
    info->ip_daddr = ip->daddr;
    info->ip_proto = ip->protocol;
    info->ports = *((uint32_t*) ip + ip->ihl);
    info->dev = dev;
}

/**
 * Cache flow routine.
 *
 * If packet info (at the end of the packet buffer) is valid, a new
 * flow entry (and flow info) will be created for the packet.
 *
 * The following must be true for the flow to be cached:
 *    - packet_info->cookie must be valid
 *    - packet_info->bucket must be valid (LUT bucket ptr)
 *    - packet length must not have changed during its traversal
 *      through the Linux stack
 *    - If TCP, sequence number and sequence acknowlegement numbers
 *      must not have changed
 *
 * If any of the above conditions is not true, then the flow associated with
 * this packet will not be cached.
 *
 * @param *skb pointer to sk_buff
 * @param *dev pointer to output interface's device structe
 */
static int cvm_ipfwd_cache_flow(struct sk_buff *skb, struct net_device *dev, cvm_packet_info_t *info)
{
    int idx, status;
    cvm_ipfwd_flow_entry_t entry;
    cvm_ipfwd_flow_bucket_t *bucket;

    /* iph points to the new header -- ip points to the cached header */
    struct iphdr *ipnew = skb->nh.iph, *ip = &info->ip;
    struct tcphdr *l4 = (struct tcphdr *) ((uint32_t*)ipnew + ipnew->ihl);

    /* Check whether this buffer has valid packet_info */
    if ((info->cookie != CVM_IPFWD_COOKIE_VALUE) || (info->bucket == NULL)) {
        //DEBUGPRINT("failed(info->cookie=%ld, info->bucket=%p)\n", info->cookie, info->bucket);
        return 0;
    }

    bucket = info->bucket;

#if 0
    DEBUGPRINT("cvm_ipfwd_cache_flow -- skb=%p, skb->end=%p, info=%p, cookie = %lx -- ", skb, skb->end, info, info->cookie);
    printf("ip: len = %d, proto = %d\n", ip->tot_len, ip->protocol);
    printf("ipnew: len = %d, proto = %d\n", ipnew->tot_len, ipnew->protocol);
#endif

    /* Verify length */
    if (ip->tot_len != ipnew->tot_len)
        return 0;

    /* If TCP, check whether sequence number or acknowlegment number changed */
    if (ipnew->protocol == IP_PROTOCOL_TCP)
    {
        if (info->seq != l4->seq)
            return 0;
        if (info->ack_seq != l4->ack_seq)
            return 0;
    }

#if 0
    if (spin_trylock(&bucket->hdr.wr_lock))
    {
        /* Lock is taken -- abort */
        return 0;
    }
#else
    /* packet_info valid -> lock bucket */
    spin_lock(&bucket->hdr.wr_lock);
#endif

    /* See if a matching entry already exists */
    idx = cvm_ipfwd_find_entry(bucket, ip, &entry);

    if (idx == CVM_IPFWD_NOT_FOUND)
    {
        idx = cvm_ipfwd_get_free_entry(bucket, &entry); /* Find available entry (replace oldest) */

#if 1
        /* If the entry-to-be-replaced was just created, don't replace it */
        if (cvm_ipfwd_check_timestamp_just_created(entry.s.timestamp)) {
            spin_unlock(&bucket->hdr.wr_lock);
            return 0;
        }
#endif

        //printk("%s: CVM_IPFWD_NOT_FOUND -- bucket = %p -- replacing entry %d\n", __func__, bucket, idx);
        cvm_ipfwd_setup_entry(ip, &entry);     /* Set entry fields -- free
                                                * flow info ptr if one exists */
    } else {
        status = cvm_ipfwd_check_timestamp(entry.s.timestamp);
        if (status == CVM_IPFWD_NEW)
            goto done;

        /* OLD & EXPIRED => setup new flow info (at least for now -- add
         * conditional update later, if necessary) */
        if (entry.s.flow_info) {
            cvm_ipfwd_ptr_free(cvmx_phys_to_ptr((uint64_t)entry.s.flow_info<<7));
            entry.s.flow_info = 0;
        }
    }

    /* See if need to allocate info buffer */
    if (entry.s.flow_info == 0) {
        entry.s.flow_info = ((uint64_t) cvmx_ptr_to_phys(cvm_ipfwd_ptr_alloc())) >> 7;
        if (entry.s.flow_info == 0)
        {
            DEBUGPRINT("%s: Failed to allocate flow info buffer\n", __func__); 
            spin_unlock(&bucket->hdr.wr_lock);
            return (0);
        }

        /* Setup info */
        cvm_ipfwd_setup_flow_info(cvmx_phys_to_ptr((uint64_t)entry.s.flow_info<<7), skb, dev);
    }

    /* Refresh time stamp */
    entry.s.timestamp = CVM_IPFWD_TIMER_GET_TIME;

    /* Store entry */
    bucket->entry[idx] = entry;

done:
    spin_unlock(&bucket->hdr.wr_lock);
    return (0);
}
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

/**
 * Return the version string. Generated from CVS tags. Turns
 * NAME_V_V_V_build_B into "V.V.V, build B". If the tag isn't of
 * this format then the tag is returned. If there isn't a tag
 * then the build date is returned as "Internal __DATE__".
 *
 * @return Version string
 */
static inline const char *cvm_oct_get_version(void)
{
    static char version[80];
    const char *cvs_tag = "$Name:  $";

    if (cvs_tag[7] == ' ')
    {
        snprintf(version, sizeof(version), "Internal %s", __DATE__);
    }
    else
    {
        char *major = NULL;
        char *minor1 = NULL;
        char *minor2 = NULL;
        char *build = NULL;
        char *buildnum = NULL;
        char *end = NULL;
        char buf[80];

        strncpy(buf, cvs_tag, sizeof(buf));
        buf[sizeof(buf)-1] = 0;

        major = strchr(buf, '_');
        if (major)
        {
            major++;
            minor1 = strchr(major, '_');
            if (minor1)
            {
                *minor1 = 0;
                minor1++;
                minor2 = strchr(minor1, '_');
                if (minor2)
                {
                    *minor2 = 0;
                    minor2++;
                    build = strchr(minor2, '_');
                    if (build)
                    {
                        *build = 0;
                        build++;
                        buildnum = strchr(build, '_');
                        if (buildnum)
                        {
                            *buildnum = 0;
                            buildnum++;
                            end = strchr(buildnum, ' ');
                            if (end)
                                *end = 0;
                        }
                    }
                }
            }
        }

        if (major && minor1 && minor2 && build && buildnum && (strcmp(build, "build") == 0))
            snprintf(version, sizeof(version), "%s.%s.%s, build %s", major, minor1, minor2, buildnum);
        else
            snprintf(version, sizeof(version), "%s", cvs_tag);
    }

    return version;
}

#define CVM_OCT_CN31XX  0x000d0100
#define CVM_OCT_CN30XX  0x000d0200
static inline uint32_t cvm_oct_get_chipid(void)
{
    uint32_t id;
    asm volatile ("mfc0 %0, $15,0" : "=r" (id));
    return id;
}

static int cvm_oct_has_4ports(void)
{
    uint32_t chipid = cvm_oct_get_chipid();
    chipid &= 0xffffff00;
    switch (chipid)
    {
        case CVM_OCT_CN31XX:
        case CVM_OCT_CN30XX:
            return 0;
        default:
            return 1;
    }
}

static int cvm_oct_is_gmii_mode(void)
{
    cvmx_gmxx_inf_mode_t mode;

    if (cvm_oct_has_4ports())
        return 0;

    mode.u64 = cvmx_read_csr(CVMX_GMXX_INF_MODE(0));
    return mode.s.type; /* GMII mode */
}


#if USE_SKBUFFS_IN_HW

/**
 * Put an skbuff into a hardware pool
 *
 * @param pool   Pool to populate
 * @param size   Size of the buffer needed for the pool
 * @return Zero on success
 */
static inline int cvm_oct_free_hw_buffer(uint64_t pool, uint64_t size)
{
    struct sk_buff *skb = dev_alloc_skb(size + 128);
#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    char *src, *dst;
    cvm_packet_info_t *info;
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

    if (unlikely(skb == NULL))
        return -1;

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    /* 
     * Make room for our packet_info structure to enable
     * marking the packet.
     *
     * This is done by bumping up the skb_shinfo struct 
     * which is at the end of the buffer. 
     *
     * For non-marked packets, skb->end - skb->head should
     * always equal skb->truesize - sizeof(struct sk_buff).
     *
     */
    src = skb->end;
    /* Bump it up by sizeof(cvm_ipfwd_packet_info_t) + 16 bytes. */
    skb->end -= (sizeof(cvm_packet_info_t) + 16);
    dst = skb->end;

    /* Copy shared info structure to its new location */
    memcpy(dst, src, sizeof(struct skb_shared_info));

    /* Initialize packet_info structure */
    info = (cvm_packet_info_t *) ((((uint64_t) skb->end + sizeof(struct skb_shared_info)) + 7) & ~7ull);
    info->bucket = NULL;
    info->cookie = 0;
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
    
    skb_reserve(skb, 128 - (((unsigned long)skb->data) & 0x7f));
    *(struct sk_buff **)(skb->data - sizeof(void*)) = skb;
    /* Do a syncw to make sure the updates to the skb header are visiblle when
        another core reads this packet in */
    cvmx_fpa_free(skb->data, pool, DONT_WRITEBACK(size/128));
    return 0;
}


/**
 * Fill the supplied hardware pool with skbuffs
 *
 * @param pool     Pool to allocate an skbuff for
 * @param size     Size of the buffer needed for the pool
 * @param elements Number of buffers to allocate
 */
static void cvm_oct_fill_hw_skbuff(uint64_t pool, uint64_t size, uint64_t elements)
{
    while (elements--)
        if (cvm_oct_free_hw_buffer(pool, size))
            panic("Failed to allocate skb for hardware pool %lu\n", pool);
}


/**
 * Free the supplied hardware pool of skbuffs
 *
 * @param pool     Pool to allocate an skbuff for
 * @param size     Size of the buffer needed for the pool
 * @param elements Number of buffers to allocate
 */
static void cvm_oct_free_hw_skbuff(uint64_t pool, uint64_t size, int elements)
{
    char *memory;

    do
    {
        memory = cvmx_fpa_alloc(pool);
        if (memory)
        {
            struct sk_buff *skb = *(struct sk_buff **)(memory - sizeof(void*));
            elements--;
            dev_kfree_skb(skb);
        }
    } while (memory);

    if (elements < 0)
        printk("Warning: Freeing of pool %lu had too many skbuffs (%d)\n", pool, elements);
    else if (elements > 0)
        printk("Warning: Freeing of pool %lu is missing %d skbuffs\n", pool, elements);
}

#endif


/**
 * Given a packet data address, return a pointer to the
 * beginning of the packet buffer.
 *
 * @param packet_ptr Packet data hardware address
 * @return Packet buffer pointer
 */
static inline void *cvm_oct_get_buffer_ptr(cvmx_buf_ptr_t packet_ptr)
{
    return cvmx_phys_to_ptr(((packet_ptr.s.addr >> 7) - packet_ptr.s.back) << 7);
}


/**
 * Enqueue a packet into the PKO output queue
 *
 * @param dev    Linux device to enqueue for
 * @param pko_command
 *               PKO command encoding of the packet
 * @param packet The packet data
 * @return Zero on success
 */
static inline int cvm_oct_low_level_send(struct net_device *dev,
                                         cvmx_pko_command_word0_t pko_command,
                                         cvmx_buf_ptr_t packet)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    uint64_t tx_command_state;
    uint64_t *tx_command_ptr;
    uint64_t tx_command_index;

    /* Get the queue command ptr location from the FAU */
    CVMX_SYNCIOBDMA;
    tx_command_state = cvmx_scratch_read64(CVMX_SCR_SCRATCH);
    tx_command_ptr = cvmx_phys_to_ptr(tx_command_state>>CVMX_PKO_INDEX_BITS);
    tx_command_index = tx_command_state & CVMX_PKO_INDEX_MASK;
    tx_command_ptr += tx_command_index;

    /* Check if we are at the end of the buffer and need to chain the next one */
#if USE_OPTIMIZED_PKO_BUFFER_SIZE
    if (likely(tx_command_index < CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE/8 - 4))
#else
    if (likely(tx_command_index < CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE/8 - 3))
#endif
    {
        /* No buffer needed. Output the command and go */
        tx_command_ptr[0] = pko_command.u64;
        tx_command_ptr[1] = packet.u64;
    }
    else
    {
        uint64_t *newbuf;
        /* We need a new output buffer. At this point we have either 2 or 3
            words left. The first word will always be the pko command */
        tx_command_ptr[0] = pko_command.u64;

        /* Get a buffer for the new command buffer. We don't use scratch in
            case some user app is doing something. It's possible the user app
            is using the scrach when the interrupt happens. Reading scratch and
            fetching the next buffer isn't atomic */
        newbuf = cvmx_fpa_alloc(CVMX_FPA_OUTPUT_BUFFER_POOL);
        if (unlikely(!newbuf))
            return -1;

#if USE_OPTIMIZED_PKO_BUFFER_SIZE == 0
        /* If we only have two words left we have to split the command over the
            command buffers */
        if (CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE/8 - tx_command_index == 2)
        {
            tx_command_ptr[1] = cvmx_ptr_to_phys(newbuf);
            *newbuf = packet.u64;
            cvmx_fau_atomic_write64(CVMX_FAU_REG_OQ_ADDR_INDEX + priv->queue*8, (tx_command_ptr[1]<<CVMX_PKO_INDEX_BITS) + 1);
        }
        else
#endif
        {
            tx_command_ptr[1] = packet.u64;
            tx_command_ptr[2] = cvmx_ptr_to_phys(newbuf);
            cvmx_fau_atomic_write64(CVMX_FAU_REG_OQ_ADDR_INDEX + priv->queue*8, tx_command_ptr[2]<<CVMX_PKO_INDEX_BITS);
        }
    }

    cvmx_pko_doorbell(priv->port, priv->queue, 2);
    return 0;
}


/**
 * Packet transmit
 *
 * @param skb    Packet to send
 * @param dev    Device info structure
 * @return Always returns zero
 */
static int cvm_oct_xmit(struct sk_buff *skb, struct net_device *dev)
{
    cvmx_pko_command_word0_t    pko_command;
    cvmx_buf_ptr_t              hw_buffer;
    uint64_t                    old_scratch;
    uint64_t                    old_scratch2;
    int                         dropped;
    cvm_oct_private_t*          priv = (cvm_oct_private_t*)dev->priv;
    int32_t in_use;
    int32_t buffers_to_free;
#if USE_SKBUFFS_IN_HW
    unsigned char *fpa_head;
#endif
#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    cvm_packet_info_t *pinfo;
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

    /* Prefetch the private data structure. It is larger that one cache line */
    CVMX_PREFETCH(priv, 0);
    CVMX_PREFETCH(priv, 128);

    /* Start off assuming no drop */
    dropped = 0;

    /* Save scratch in case userspace is using it */
    CVMX_SYNCIOBDMA;
    old_scratch = cvmx_scratch_read64(CVMX_SCR_SCRATCH);
    old_scratch2 = cvmx_scratch_read64(CVMX_SCR_SCRATCH+8);

    /* Assume we're going to be able t osend this packet. Fetch and increment
        the number of pending packets for output */
    cvmx_fau_async_fetch_and_add32(CVMX_SCR_SCRATCH+8, FAU_NUM_PACKET_BUFFERS_TO_FREE, 0);
    cvmx_fau_async_fetch_and_add32(CVMX_SCR_SCRATCH, priv->fau, 1);

    /* Build the PKO buffer pointer */
    hw_buffer.u64 = 0;
    hw_buffer.s.addr = cvmx_ptr_to_phys(skb->data);
    hw_buffer.s.pool = 0;
    hw_buffer.s.size = (uint64_t)skb->end - (uint64_t)skb->head;

    /* Build the PKO command */
    pko_command.u64 = 0;
    pko_command.s.segs = 1;
    pko_command.s.total_bytes = skb->len;
    pko_command.s.size0 = CVMX_FAU_OP_SIZE_32;
    pko_command.s.subone0 = 1;

    /* See if we can put this skb in the FPA pool */
#if USE_SKBUFFS_IN_HW
    fpa_head = skb->head + 128 - ((uint64_t)skb->head&0x7f);
    if ((skb->data >= fpa_head) && (skb->end - fpa_head >= CVMX_FPA_PACKET_POOL_SIZE) &&
        !(skb_shared(skb) || skb_cloned(skb) || skb_header_cloned(skb)))
    {
        pko_command.s.dontfree = 0;
        /* This will reduce the number of buffers we need to free by one */
        pko_command.s.reg0 = FAU_NUM_PACKET_BUFFERS_TO_FREE;

        hw_buffer.s.back = (skb->data - fpa_head)>>7;
        *(struct sk_buff **)(fpa_head-8) = skb;
        /* The skbuff will be reused without ever being freed. We must free the
            dst structure. FIXME: Likely need other stuff here as well */
        dst_release(skb->dst);
        skb->dst = NULL;
#ifdef CONFIG_XFRM
        secpath_put(skb->sp);
        skb->sp = NULL;
#endif
        nf_reset(skb);
#ifdef CONFIG_BRIDGE_NETFILTER
        nf_bridge_put(skb->nf_bridge);
#endif /* CONFIG_BRIDGE_NETFILTER */
#ifdef CONFIG_NET_SCHED
        skb->tc_index = 0;
#ifdef CONFIG_NET_CLS_ACT
        skb->tc_verd = 0;
#endif /* CONFIG_NET_CLS_ACT */
#endif /* CONFIG_NET_SCHED */
    }
    else
    {
        pko_command.s.dontfree = 1;
        pko_command.s.reg0 = priv->fau;
    }
#else
    pko_command.s.reg0 = priv->fau;
    pko_command.s.dontfree = 1;
#endif /* USE_SKBUFFS_IN_HW */

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    pinfo = (cvm_packet_info_t *) ((((uint64_t) skb->end + sizeof(struct skb_shared_info)) + 7) & ~7ull);
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

    /* Check if we can use the hardware checksumming */
    if (USE_HW_TCPUDP_CHECKSUM && (skb->protocol == htons(ETH_P_IP)) &&
        (skb->nh.iph->version == 4) && (skb->nh.iph->ihl == 5) &&
        ((skb->nh.iph->frag_off == 0) || (skb->nh.iph->frag_off == 1<<14)) &&
        ((skb->nh.iph->protocol == IP_PROTOCOL_TCP) || (skb->nh.iph->protocol == IP_PROTOCOL_UDP)))
    {
#ifdef CVM_IPFWD_DRIVER_OFFLOAD
        /* Try to cache (or update) this flow */
        if (ipfwd_driver_offload) 
            cvm_ipfwd_cache_flow(skb, dev, pinfo);
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

        /* Use hardware checksum calc */
        pko_command.s.ipoffp1 = sizeof(struct ethhdr) + 1;
    }

    /* This lock is here so that two cores cannot increment the PKO output
        queue position at the same time. Looking at the data structures it
        might appear that this lock could happen later, but it can't because
        it is also the lock for the PKO command words */
    spin_lock(&priv->tx_free_list.lock);

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    pinfo->cookie = 0;
    pinfo->bucket = NULL;
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

    /* Get the number of skbuffs in use by the hardware */
    CVMX_SYNCIOBDMA;
    in_use = cvmx_scratch_read64(CVMX_SCR_SCRATCH);
    buffers_to_free = cvmx_scratch_read64(CVMX_SCR_SCRATCH+8);

    /* Get the position for writing the PKO command. This will be used when we
        call cvm_oct_low_level_send. If we drop, this must be decremented */
    cvmx_fau_async_fetch_and_add64(CVMX_SCR_SCRATCH, CVMX_FAU_REG_OQ_ADDR_INDEX + priv->queue*8, 2);

    /* If we're sending faster than the receive can free them then don't do
        the HW free */
    if ((buffers_to_free<-100) && !pko_command.s.dontfree)
    {
        pko_command.s.dontfree = 1;
        pko_command.s.reg0 = priv->fau;
    }

    /* Free skbuffs not in use by the hardware, two at a time. Have 3 packets
        left between the delete and insert to allow for timing slop */
    if (skb_queue_len(&priv->tx_free_list) > in_use + 4)
    {
        dev_kfree_skb(__skb_dequeue(&priv->tx_free_list));
        dev_kfree_skb(__skb_dequeue(&priv->tx_free_list));
    }

    /* Drop this packet if we have too many already queued to the HW */
    if (unlikely(skb_queue_len(&priv->tx_free_list) >= dev->tx_queue_len))
    {
        //DEBUGPRINT("%s: Tx dropped. Too many queued\n", dev->name);
        /* We need to make sure the fetch and add for the PKO position
            finished. Normally the low level send does this, but in this
            case we don't call it */
        CVMX_SYNCIOBDMA;
        dropped=1;
    }
    /* Send the packet to the output queue */
    else if (unlikely(cvm_oct_low_level_send(dev, pko_command, hw_buffer)))
    {
        DEBUGPRINT("%s: Failed to send the packet\n", dev->name);
        dropped=1;
    }

    /* Restore the scratch area */
    cvmx_scratch_write64(CVMX_SCR_SCRATCH, old_scratch);
    cvmx_scratch_write64(CVMX_SCR_SCRATCH+8, old_scratch2);

    if (unlikely(dropped))
    {
        dev_kfree_skb_any(skb);
        cvmx_fau_atomic_add64(CVMX_FAU_REG_OQ_ADDR_INDEX + priv->queue*8, -2);
        cvmx_fau_atomic_add32(priv->fau, -1);
        priv->stats.tx_dropped++;
    }
    else
    {
#if USE_SKBUFFS_IN_HW
        /* Put this packet on the queue to be freed later */
        if (pko_command.s.dontfree)
            __skb_queue_tail(&priv->tx_free_list, skb);
        else
            cvmx_fau_atomic_add32(priv->fau, -1);
#else
        /* Put this packet on the queue to be freed later */
        __skb_queue_tail(&priv->tx_free_list, skb);
#endif /* USE_SKBUFFS_IN_HW */
    }

    spin_unlock(&priv->tx_free_list.lock);

    return 0;
}


/**
 * Packet transmit to the POW
 *
 * @param skb    Packet to send
 * @param dev    Device info structure
 * @return Always returns zero
 */
static int cvm_oct_xmit_pow(struct sk_buff *skb, struct net_device *dev)
{
    cvm_oct_private_t*  priv = (cvm_oct_private_t*)dev->priv;
    void *              packet_buffer;
    void *              copy_location;

    /* Get a work queue entry */
    cvmx_wqe_t *work = cvmx_fpa_alloc(CVMX_FPA_WQE_POOL);
    if (unlikely(work == NULL))
    {
        DEBUGPRINT("%s: Failed to allocate a work queue entry\n", dev->name);
        priv->stats.tx_dropped++;
        dev_kfree_skb(skb);
        return 0;
    }

    /* Get a packet buffer */
    packet_buffer = cvmx_fpa_alloc(CVMX_FPA_PACKET_POOL);
    if (unlikely(packet_buffer == NULL))
    {
        DEBUGPRINT("%s: Failed to allocate a packet buffer\n", dev->name);
        cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));
        priv->stats.tx_dropped++;
        dev_kfree_skb(skb);
        return 0;
    }

    /* Calculate where we need to copy the data to. We need to leave 8 bytes
        for a next pointer (unused). We also need to include any configure
        skip. Then we need to align the IP packet src and dest into the same
        64bit word. The below calculation may add a little extra, but that
        doesn't hurt */
    copy_location = packet_buffer + sizeof(uint64_t);
    copy_location += ((CVMX_HELPER_FIRST_MBUFF_SKIP+7)&0xfff8) + 6;

    /* We have to copy the packet since whoever processes this packet
        will free it to a hardware pool. We can't use the trick of
        counting outstanding packets like in cvm_oct_xmit */
    memcpy(copy_location, skb->data, skb->len);

    /* Fill in some of the work queue fields. We may need to add more
        if the software at the other end needs them */
    work->hw_chksum     = skb->csum;
    work->len           = skb->len;
    work->ipprt         = priv->port;
    work->qos           = priv->port & 0x7;
    work->grp           = pow_send_group;
    work->tag_type      = CVMX_HELPER_INPUT_TAG_TYPE;
    work->tag           = pow_send_group; /* FIXME */
    work->word2.u64     = 0;    /* Default to zero. Sets of zero later are commented out */
    work->word2.s.bufs  = 1;
    work->packet_ptr.u64 = 0;
    work->packet_ptr.s.addr = cvmx_ptr_to_phys(copy_location);
    work->packet_ptr.s.pool = CVMX_FPA_PACKET_POOL;
    work->packet_ptr.s.size = CVMX_FPA_PACKET_POOL_SIZE;
    work->packet_ptr.s.back = (copy_location - packet_buffer)>>7;

    if (skb->protocol == htons(ETH_P_IP))
    {
        work->word2.s.ip_offset     = 14;
        //work->word2.s.vlan_valid  = 0; /* FIXME */
        //work->word2.s.vlan_cfi    = 0; /* FIXME */
        //work->word2.s.vlan_id     = 0; /* FIXME */
        //work->word2.s.dec_ipcomp  = 0; /* FIXME */
        work->word2.s.tcp_or_udp    = (skb->nh.iph->protocol == IP_PROTOCOL_TCP) || (skb->nh.iph->protocol == IP_PROTOCOL_UDP);
        //work->word2.s.dec_ipsec   = 0; /* FIXME */
        //work->word2.s.is_v6       = 0; /* We only support IPv4 right now */
        //work->word2.s.software    = 0; /* Hardware would set to zero */
        //work->word2.s.L4_error    = 0; /* No error, packet is internal */
        work->word2.s.is_frag       = !((skb->nh.iph->frag_off == 0) || (skb->nh.iph->frag_off == 1<<14));
        //work->word2.s.IP_exc      = 0;  /* Assume Linux is sending a good packet */
        work->word2.s.is_bcast      = (skb->pkt_type==PACKET_BROADCAST);
        work->word2.s.is_mcast      = (skb->pkt_type==PACKET_MULTICAST);
        //work->word2.s.not_IP      = 0; /* This is an IP packet */
        //work->word2.s.rcv_error   = 0; /* No error, packet is internal */
        //work->word2.s.err_code    = 0; /* No error, packet is internal */

        /* When copying the data, include 4 bytes of the ethernet header to
            align the same way hardware does */
        memcpy(work->packet_data, skb->data + 10, sizeof(work->packet_data));
    }
    else
    {
        //work->word2.snoip.vlan_valid  = 0; /* FIXME */
        //work->word2.snoip.vlan_cfi    = 0; /* FIXME */
        //work->word2.snoip.vlan_id     = 0; /* FIXME */
        //work->word2.snoip.software    = 0; /* Hardware would set to zero */
        work->word2.snoip.is_rarp       = skb->protocol == htons(ETH_P_RARP);
        work->word2.snoip.is_arp        = skb->protocol == htons(ETH_P_ARP);
        work->word2.snoip.is_bcast      = (skb->pkt_type==PACKET_BROADCAST);
        work->word2.snoip.is_mcast      = (skb->pkt_type==PACKET_MULTICAST);
        work->word2.snoip.not_IP        = 1; /* IP was done up above */
        //work->word2.snoip.rcv_error   = 0; /* No error, packet is internal */
        //work->word2.snoip.err_code    = 0; /* No error, packet is internal */
        memcpy(work->packet_data, skb->data, sizeof(work->packet_data));
    }

    /* Submit the packet to the POW */
    cvmx_pow_work_submit(work, work->tag, work->tag_type, work->qos, work->grp);
    priv->stats.tx_packets++;
    priv->stats.tx_bytes += skb->len;
    dev_kfree_skb(skb);
    return 0;
}


/**
 * Interrupt handler used for quickly detecting RGMII link
 * status changes. This interrupt is setup to occur when the
 * inband status changes for the RGMII ports. It may be call
 * due to other RML interrupts as well.
 *
 * @param cpl
 * @param dev_id
 * @param regs
 * @return
 */
static irqreturn_t cvm_oct_rml_interrupt(int cpl, void *dev_id, struct pt_regs *regs)
{
    cvmx_npi_rsl_int_blocks_t rsl_int_blocks;

    /* Check and see if this interrupt was caused by the GMX block */
    rsl_int_blocks.u64 = cvmx_read_csr(CVMX_NPI_RSL_INT_BLOCKS);
    if (rsl_int_blocks.s.gmx0 || rsl_int_blocks.s.gmx1)
    {
        int interface;
        for (interface=0; interface<2; interface++)
        {
            int index;
            int max_port;
            cvmx_gmxx_inf_mode_t mode;

            /* Skip interfaces that aren't enabled */
            mode.u64 = cvmx_read_csr(CVMX_GMXX_INF_MODE(interface));
            if (!mode.s.en)
                continue;

            /* Figure out how many ports we have */
            if (cvm_oct_is_gmii_mode())
                max_port = 2;
            else if (cvm_oct_has_4ports())
                max_port = 4;
            else
                max_port = 3;

            for (index=0; index<max_port; index++)
            {
                uint64_t ena = cvmx_read_csr(CVMX_GMXX_RXX_INT_EN(index, interface));
                uint64_t sum = cvmx_read_csr(CVMX_GMXX_RXX_INT_REG(index, interface));
                if (ena & sum)
                {
                    int port = interface*16 + index;
                    int queue;
                    int stop_queue;
                    unsigned long flags;
                    cvmx_gmxx_tx_ovr_bp_t ovr_bp;

#ifdef WN302T_PHY_FIX
		    cvmx_smi_cmd_t          smi_cmd;
		    cvmx_smi_rd_dat_t       smi_rd;
		    
		    smi_cmd.u64 = 0;
		    smi_cmd.s.phy_op = 1;
		    smi_cmd.s.phy_adr = first_phy_id;
		    smi_cmd.s.reg_adr = 0x5;
		    cvmx_write_csr(CVMX_SMI_CMD, smi_cmd.u64);

		    do
		    {
			if (!in_interrupt())
			    yield();
			smi_rd.u64 = cvmx_read_csr(CVMX_SMI_RD_DAT);
		    } while (smi_rd.s.pending);
		    if((smi_rd.s.dat & 0x3E0) != 0) {
			cvmx_gmxx_rxx_int_en_t gmx_en;
			gmx_en.u64 = cvmx_read_csr(CVMX_GMXX_RXX_INT_EN(index, interface));
			gmx_en.s.phy_dupx = 0;
			gmx_en.s.phy_link = 0;
			gmx_en.s.phy_spd = 0;
			cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmx_en.u64);
			cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), gmx_en.u64);
			return IRQ_NONE;
		    }
#endif /* WN302T_PHY_FIX */
                    /* Take the global register lock since we are going to touch
                        registers that affect more than one port */
                    spin_lock_irqsave(&global_register_lock, flags);

                    /* Disable the lowest level output from the port */
                    cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(interface),
                                   cvmx_read_csr(CVMX_ASXX_TX_PRT_EN(interface)) & ~(1<<index));

                    /* Disable the lowest level input from the port. Note that
                        this also disables inband status */
                    cvmx_write_csr(CVMX_ASXX_RX_PRT_EN(interface),
                                   cvmx_read_csr(CVMX_ASXX_RX_PRT_EN(interface)) & ~(1<<index));

                    /* Disable packet output by disabling all queues for this port */
                    queue = cvmx_pko_get_base_queue(port);
                    stop_queue = queue + cvmx_pko_get_num_queues(port);
                    while (queue < stop_queue)
                    {
                        cvmx_pko_mem_queue_qos_t pko_qos;
                        pko_qos.u64 = 0;
                        pko_qos.s.pid = port;
                        pko_qos.s.qid = queue;
                        pko_qos.s.qos_mask = 0;
                        cvmx_write_csr(CVMX_PKO_REG_READ_IDX, queue);
                        cvmx_write_csr(CVMX_PKO_MEM_QUEUE_QOS, pko_qos.u64);
                        queue++;
                    }

                    /* Disable backpressure for this port */
                    ovr_bp.u64 = cvmx_read_csr(CVMX_GMXX_TX_OVR_BP(interface));
                    ovr_bp.s.en |= 1<<index;
                    ovr_bp.s.bp &= ~(1<<index);
                    cvmx_write_csr(CVMX_GMXX_TX_OVR_BP(interface), ovr_bp.u64);

                    /* Disable the interrupt for this port. It will be enabled
                        once the port is brought back up */
                    cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), 0);

                    /* We're done with the global registers */
                    spin_unlock_irqrestore(&global_register_lock, flags);

                    /* Inform the user and kernel of the change */
                    if (cvm_oct_device[port])
                    {
                        printk("%s: Link down\n", cvm_oct_device[port]->name);
                        if (netif_carrier_ok(cvm_oct_device[port]))
                            netif_carrier_off(cvm_oct_device[port]);
                    }
                }
            }
        }
        return IRQ_HANDLED;
    }
    return IRQ_NONE;
}


/**
 * Interrupt handler. The interrupt occurs whenever the POW
 * transitions from 0->1 packets in our group.
 *
 * @param cpl
 * @param dev_id
 * @param regs
 * @return
 */
static irqreturn_t cvm_oct_do_interrupt(int cpl, void *dev_id, struct pt_regs *regs)
{
    struct net_device *dev;
    /* Acknowledge the interrupt */
    int port =0;
#if INTERRUPT_LIMIT
    cvmx_write_csr(CVMX_POW_WQ_INT, 1<<pow_receive_group);
#else
    cvmx_write_csr(CVMX_POW_WQ_INT, 0x10001<<pow_receive_group);
#endif
    //cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0x1ful<<24);
    preempt_disable();
    for (port=0; port<TOTAL_NUMBER_OF_PORTS; port++) {
        if (cvm_oct_device[port]) {
            dev = cvm_oct_device[port];    
            break;
        }
    }
   if (netif_rx_schedule_prep(dev)) {
    	cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0);
        __netif_rx_schedule(dev);
    }
    //tasklet_schedule(cvm_oct_tasklet + smp_processor_id());
    preempt_enable();
    return IRQ_HANDLED;
}


#ifdef CONFIG_NET_POLL_CONTROLLER
/**
 * This is called when the kernel needs to manually poll the
 * device. For Octeon, this is simply calling the interrupt
 * handler. We actually poll all the devices, not just the
 * one supplied.
 *
 * @param dev    Device to poll. Unused
 */
static void cvm_oct_poll_controller(struct net_device *dev)
{
    preempt_disable();
    tasklet_schedule(cvm_oct_tasklet + smp_processor_id());
    preempt_enable();
}
#endif


/**
 * Poll Function
 *
 * @param unused
 */
static int cvm_oct_rx_poll(struct net_device *dev, int *budget)
{
    const uint64_t      coreid = cvmx_get_core_num();
    uint64_t            old_group_mask;
    uint64_t            old_scratch;
    uint64_t            rx_count = 0;
    int rx_limit = dev->quota;
    int early_stop = 0;

#if USE_SKBUFFS_IN_HW
    int                 number_to_free;
#endif

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    cvm_ipfwd_eh_t *eh;
    cvm_ipfwd_flow_entry_t entry;
    cvm_ipfwd_flow_bucket_t *bucket = NULL;
    struct iphdr *ip;
    struct tcphdr *th;
    cvm_ipfwd_flow_info_t *info;
    int idx, status;
    int skip_fastpath;
#endif

    /* Prefetch cvm_oct_device since we know we need it soon */
    CVMX_PREFETCH(cvm_oct_device, 0);

    /* Save scratch in case userspace is using it */
    CVMX_SYNCIOBDMA;
    old_scratch = cvmx_scratch_read64(CVMX_SCR_SCRATCH);

    /* Only allow work for our group */
    old_group_mask = cvmx_read_csr(CVMX_POW_PP_GRP_MSKX(coreid));
    cvmx_write_csr(CVMX_POW_PP_GRP_MSKX(coreid), 1<<pow_receive_group);

    cvmx_pow_work_request_async(CVMX_SCR_SCRATCH, CVMX_POW_NO_WAIT);

    while (1)
    {
        void *start_of_buffer;
        struct sk_buff *skb;
        cvm_oct_callback_result_t callback_result;
        cvmx_wqe_t *work = cvmx_pow_work_response_async(CVMX_SCR_SCRATCH);
        if (work == NULL)
            break;

        /* Limit each core to processing 60 packets without a break. This way
            the RX can't starve the TX task. */
#if 0
#ifndef CVM_IPFWD_DRIVER_OFFLOAD
        if (likely(rx_count < 48))
#else
        if (likely(rx_count < 48))
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
#endif
        if (rx_limit-- >= 0) {
            cvmx_pow_work_request_async_nocheck(CVMX_SCR_SCRATCH, CVMX_POW_NO_WAIT);
        }
        else
        {
            early_stop = 1;
            cvmx_scratch_write64(CVMX_SCR_SCRATCH, 0x8000000000000000ull);
            cvmx_pow_tag_sw_null_nocheck();
        }
        rx_count++;


#ifdef CVM_IPFWD_DRIVER_OFFLOAD
        if (ipfwd_driver_offload) {
            bucket = &cvm_gbl_ipfwd_cache[work->tag & CVM_IPFWD_LUT_MASK];
            CVMX_PREFETCH0(bucket);
        }
#ifdef CVM_IPFWD_DYNAMIC_SHORT
        if (work->word2.s.bufs == 0)
        {
            /* Dynamic short pre-adjustment */
            work->packet_ptr.u64 = 0;
            work->packet_ptr.s.size = CVMX_FPA_WQE_POOL_SIZE;
            work->packet_ptr.s.pool = CVMX_FPA_WQE_POOL;
            if (work->word2.s.not_IP)
            {
                /* Non-IP */
                work->packet_ptr.s.addr = (uint64_t)(&work->packet_data);
            }
            else
            {
                if(!work->word2.s.is_v6)
                {
                    /* IPv4 (shifted four bytes) */
                    work->packet_ptr.s.addr = (uint64_t)(&work->packet_data) + CVM_IPFWD_IP_OFFSET + 4 - work->word2.s.ip_offset;
                }
                else
                    /* IPv6 */
                    work->packet_ptr.s.addr = (uint64_t)(&work->packet_data) + CVM_IPFWD_IP_OFFSET - work->word2.s.ip_offset;
            }
        }
#endif /* CVM_IPFWD_DYNAMIC_SHORT */
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

        CVMX_PREFETCH(cvm_oct_device[work->ipprt], 0);
        start_of_buffer = cvm_oct_get_buffer_ptr(work->packet_ptr);
        CVMX_PREFETCH(start_of_buffer, -8);

        /* Immediately throw away all packets with receive errors except for
            CVMX_PIP_OVER_ERR. CVMX_PIP_OVER_ERR means the packet was good,
            just oversized */
        if (USE_10MBPS_PREAMBLE_WORKAROUND && ((work->word2.snoip.err_code == 5) || (work->word2.snoip.err_code == 7)))
        {
            /* We received a packet with either an alignment error or a
                FCS error. This may be signalling that we are running
                10Mbps with GMXX_RXX_FRM_CTL[PRE_CHK] off. If this is the
                case we need to parse the packet to determine if we can
                remove a non spec preamble and generate a correct packet */
            int interface = INTERFACE(work->ipprt);
            int index = INDEX(work->ipprt);
            cvmx_gmxx_rxx_frm_ctl_t gmxx_rxx_frm_ctl;
            gmxx_rxx_frm_ctl.u64 = cvmx_read_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface));
            if (gmxx_rxx_frm_ctl.s.pre_chk == 0)
            {
                uint8_t *ptr = cvmx_phys_to_ptr(work->packet_ptr.s.addr);
                int i = 0;
                while (i<work->len-1)
                {
                    if (*ptr != 0x55)
                        break;
                    ptr++;
                    i++;
                }
                if (*ptr == 0xd5)
                {
                    //DEBUGPRINT("Port %d received 0xd5 preamble\n", work->ipprt);
                    work->packet_ptr.s.addr += i+1;
                    work->len -= i+5;
                }
                else if ((*ptr & 0xf) == 0xd)
                {
                    //DEBUGPRINT("Port %d received 0x?d preamble\n", work->ipprt);
                    work->packet_ptr.s.addr += i;
                    work->len -= i+4;
                    for (i=0; i<work->len; i++)
                    {
                        *ptr = ((*ptr&0xf0)>>4) | ((*(ptr+1)&0xf)<<4);
                        ptr++;
                    }
                }
                else
                {
                    DEBUGPRINT("Port %d unknown preamble, packet dropped\n", work->ipprt);
                    //cvmx_helper_dump_packet(work);
                    cvm_oct_free_work(work);
                    continue;
                }
            }
        }
        else if (unlikely(work->word2.snoip.rcv_error && (work->word2.snoip.err_code != CVMX_PIP_OVER_ERR)))
        {
            //DEBUGPRINT("Port %d receive error code %d, packet dropped\n", work->ipprt, work->word2.snoip.err_code);
            cvm_oct_free_work(work);
            continue;
        }

        /* We can only use the zero copy path if skbuffs are in the FPA pool
            and the packet fit in a single buffer */
#if USE_SKBUFFS_IN_HW
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
	if (likely(work->word2.s.bufs <= 1))
#else
        if (likely(work->word2.s.bufs == 1))
#endif
        {
            skb = *(struct sk_buff **)(start_of_buffer - sizeof(void*));
            /* This calculation was changed in case the skb header is using a
                different address aliasing type than the buffer. It doesn't make
                any differnece now, but the new one is more correct */
            skb->data = skb->head + work->packet_ptr.s.addr - cvmx_ptr_to_phys(skb->head);
            CVMX_PREFETCH(skb->data, 0);
            skb->len = work->len;
            skb->tail = skb->data + skb->len;

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
            /* 
             * Fast path: Do a flow-based IPFWD lookup 
             *    Packet requirements: 
             *       1) IPv4
             *       2) TCP/UDP (if TCP, no SYN, FIN, or RESET)
             *       3) non-fragment (at least for now)
             */

            skip_fastpath = 0;
            ip = (struct iphdr*) (skb->data + 14); //XXX work->word2.s.ip_offset);
            th = (struct tcphdr *) ((uint32_t*)ip + ip->ihl);
            if ((ip->protocol == IP_PROTOCOL_TCP) && ((th->syn) || (th->fin) || (th->rst)))
                skip_fastpath = 1;

            if ((ipfwd_driver_offload) &&
                    ((!work->word2.s.is_v6) &&
                     (work->word2.s.tcp_or_udp) &&
                     (!work->word2.s.is_frag) &&
                     (!skip_fastpath)))
            {
#if 0
                {
                    cvm_packet_info_t *pinfo = (cvm_packet_info_t *) ((((uint64_t) skb->end +
                                    sizeof(struct skb_shared_info)) + 7) & ~7ull);
                    if (pinfo->cookie == CVM_IPFWD_COOKIE_VALUE)
                        printf("ERROR: rx packet has a valid cookie!\n");
                }
#endif

                //printk("%s: cvm_ipfwd_find_entry(bucket = %p)\n", __func__, bucket);
                idx = cvm_ipfwd_find_entry(bucket, ip, &entry);

                if (idx == CVM_IPFWD_NOT_FOUND)
                    goto not_found;

                /* Match found */

                /* Check its timestamp */
                status = cvm_ipfwd_check_timestamp(entry.s.timestamp);

                if (status == CVM_IPFWD_EXPIRED) {
#if 0
                    ipfwd_slow_path_packets[cvmx_get_core_num()]++;
#endif
                    goto not_found;
                }

                if ((status != CVM_IPFWD_NEW) && (cvm_ipfwd_random_select()))
                {
                    /* Old match found and this packet was selected to go through
                     * the Linux stack.  Make sure there is not already someone
                     * in this bucket doing this. */
                    if (cvm_ipfwd_check_bucket_timestamp(bucket))
                    {
#if 0
                        ipfwd_slow_path_refresh[cvmx_get_core_num()]++;
#endif
                        goto not_found;
                    }
                }

                /* If we get here, entry is not expired */
                if (entry.s.flow_info)
                {
                    /* match found -- format header */
                    info = (cvm_ipfwd_flow_info_t*) cvmx_phys_to_ptr((uint64_t)entry.s.flow_info << 7);
                    eh = (cvm_ipfwd_eh_t *) skb->data;
                    eh->enet_dst_hi = info->enet_dst_hi;
                    eh->enet_dst_lo = info->enet_dst_lo;
                    eh->enet_src_hi = info->enet_src_hi;
                    eh->enet_src_lo = info->enet_src_lo;
                    ip = (struct iphdr*) (skb->data + work->word2.s.ip_offset);
                    ip->ttl--;
                    ip->saddr = info->ip_saddr;
                    ip->daddr = info->ip_daddr;
                    ip->protocol = info->ip_proto;
                    *((uint32_t*) ip + ip->ihl) = info->ports;

                    /* compute IP checksum */
                    ip->check = cvm_ipfwd_calculate_ip_header_checksum((uint16_t*) ip);

                    /* Use the cvm_oct_transmit() to send the packet */
                    cvm_oct_transmit(info->dev, work, 1);

                    /* done */
                    continue;
                }

not_found:
                /* 
                 * No match found (or old match found) -- 
                 * mark the packet (buffer) so that the flow can be
                 * cached on the way out.  
                 */
                cvm_ipfwd_mark_packet(skb, work, bucket);
            } 
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
        }
        else
#endif
        {
            /* We have to copy the packet. First allocate an skbuff for it */
            skb = dev_alloc_skb(work->len);
            if (!skb)
            {
                DEBUGPRINT("Port %d failed to allocate skbuff, packet dropped\n", work->ipprt);
                cvm_oct_free_work(work);
                continue;
            }

            /* Check if we've received a packet that was entirely stored the
                work entry. This is untested */
            if (unlikely(work->word2.s.bufs == 0))
            {
                DEBUGPRINT("Port %d received a work with work->word2.s.bufs=0, untested\n", work->ipprt);
                memcpy(skb_put(skb, work->len), work->packet_data, work->len);
                /* No packet buffers to free */
            }
            else
            {
                int segments = work->word2.s.bufs;
                cvmx_buf_ptr_t segment_ptr = work->packet_ptr;
                int len = work->len;
                while (segments--)
                {
                    cvmx_buf_ptr_t next_ptr = *(cvmx_buf_ptr_t*)cvmx_phys_to_ptr(segment_ptr.s.addr-8);
                    /* Octeon Errata PKI-100: The segment size is wrong. Until it
                        is fixed, calculate the segment size based on the packet
                        pool buffer size. When it is fixed, the following line should
                        be replaced with this one:
                            int segment_size = segment_ptr.s.size; */
                    int segment_size = CVMX_FPA_PACKET_POOL_SIZE - (segment_ptr.s.addr - (((segment_ptr.s.addr >> 7) - segment_ptr.s.back) << 7));
                    /* Don't copy more than what is left in the packet */
                    if (segment_size > len)
                        segment_size = len;
                    /* Copy the data into the packet */
                    memcpy(skb_put(skb, segment_size), cvmx_phys_to_ptr(segment_ptr.s.addr), segment_size);
#if 0
                    {
                        unsigned long *dst = skb_put(skb, segment_size);
                        unsigned long *src = cvmx_phys_to_ptr(segment_ptr.s.addr);
                        int bytes = segment_size;
						printk("-%d \n",bytes);
                        while (bytes  >= 8) {   
                            *dst = *src;
                            dst++;
                            src++;
                            bytes -=8;
                        }
                        if (bytes) {
                            memcpy(dst, src, bytes);
                        }
                    }
#endif
                    /* Reduce the amount of bytes left to copy */
                    len -= segment_size;
                    segment_ptr = next_ptr;
                }
            }
        }

        if (likely((work->ipprt < TOTAL_NUMBER_OF_PORTS) && cvm_oct_device[work->ipprt]))
        {
            struct net_device *dev = cvm_oct_device[work->ipprt];
            cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

            /* Only accept packets for devices that are currently up */
            if (likely(priv->up))
            {
                skb->protocol = eth_type_trans(skb, dev);
                skb->dev = dev;

                if (unlikely(work->word2.s.not_IP || work->word2.s.IP_exc || work->word2.s.L4_error))
                    skb->ip_summed = CHECKSUM_NONE;
                else
                    skb->ip_summed = CHECKSUM_UNNECESSARY;
                /* Increment RX stats for virtual ports */
                if (work->ipprt >= CVMX_PIP_NUM_INPUT_PORTS)
                {
                    cvmx_atomic_add64_nosync(&priv->stats.rx_packets, 1);
                    cvmx_atomic_add64_nosync(&priv->stats.rx_bytes, skb->len);
                }

                if (priv->intercept_cb)
                {
                    callback_result = priv->intercept_cb(dev, work, skb);
                    switch (callback_result)
                    {
                        case CVM_OCT_PASS:
                            //netif_receive_skb(skb);
                            netif_rx(skb);
                            break;
                        case CVM_OCT_DROP:
                            dev_kfree_skb_irq(skb);
                            cvmx_atomic_add64_nosync(&priv->stats.rx_dropped, 1);
                            break;
                        case CVM_OCT_TAKE_OWNERSHIP_WORK:
                            /* Interceptor stole our work, but we need to free
                                the skbuff */
#if USE_SKBUFFS_IN_HW
                            if (likely(work->word2.s.bufs == 1))
                            {
                                /* We can't free the skbuff since its data is
                                    the same as the work. In this case we don't
                                    do anything */
                            }
                            else
                                dev_kfree_skb_irq(skb);
#else
                            dev_kfree_skb_irq(skb);
#endif
                            break;
                        case CVM_OCT_TAKE_OWNERSHIP_SKB:
                            /* Interceptor stole our packet */
                            break;
                    }
                }
                else
                {
                    //netif_receive_skb(skb);
                    netif_rx(skb);
                    callback_result = CVM_OCT_PASS;
                }
            }
            else
            {
                /* Drop any packet received for a device that isn't up */
                //DEBUGPRINT("%s: Device not up, packet dropped\n", dev->name);
                cvmx_atomic_add64_nosync(&priv->stats.rx_dropped, 1);
                dev_kfree_skb_irq(skb);
                callback_result = CVM_OCT_DROP;
            }
        }
        else
        {
            /* Drop any packet received for a device that doesn't exist */
            DEBUGPRINT("Port %d not controlled by Linux, packet dropped\n", work->ipprt);
            dev_kfree_skb_irq(skb);
            callback_result = CVM_OCT_DROP;
        }

        /* We only need to free the work if the interceptor didn't take over
            ownership of it */
        if (callback_result != CVM_OCT_TAKE_OWNERSHIP_WORK)
        {
#if USE_SKBUFFS_IN_HW
            /* Check to see if the skbuff and work share the same packet buffer */
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
            if (likely(work->word2.s.bufs <= 1))
#else
            if (likely(work->word2.s.bufs == 1))
#endif
            {
                /* This buffer needs to be replaced, increment the number of buffers we need to free by one */
                cvmx_fau_atomic_add32(FAU_NUM_PACKET_BUFFERS_TO_FREE, 1);

#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
                if (likely(work->word2.s.bufs != 0))
                    cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));
#else
                cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));
#endif
            }
            else
                cvm_oct_free_work(work);
#else
            cvm_oct_free_work(work);
#endif
        }
    }

    /* Restore the original POW group mask */
    cvmx_write_csr(CVMX_POW_PP_GRP_MSKX(coreid), old_group_mask);
    /* Restore the scratch area */
    cvmx_scratch_write64(CVMX_SCR_SCRATCH, old_scratch);

#if USE_SKBUFFS_IN_HW
    /* Refill the packet buffer pool */
    number_to_free = cvmx_fau_fetch_and_add32(FAU_NUM_PACKET_BUFFERS_TO_FREE, 0);
    if (number_to_free>0)
    {
        cvmx_fau_atomic_add32(FAU_NUM_PACKET_BUFFERS_TO_FREE, -number_to_free);
        while (number_to_free--)
            cvm_oct_free_hw_buffer(CVMX_FPA_PACKET_POOL, CVMX_FPA_PACKET_POOL_SIZE);
    }
#endif
    *budget -= rx_count;
    if (!early_stop) {
        netif_rx_complete(dev);
	    cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0x1ful<<24);
	}

	return early_stop;
}

/**
 * Tasklet function that is scheduled on a core when an interrupt occurs.
 *
 * @param unused
 */
static void cvm_oct_tasklet_rx(unsigned long unused)
{
    const uint64_t      coreid = cvmx_get_core_num();
    uint64_t            old_group_mask;
    uint64_t            old_scratch;
    uint64_t            rx_count = 0;
#if USE_SKBUFFS_IN_HW
    int                 number_to_free;
#endif

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    cvm_ipfwd_eh_t *eh;
    cvm_ipfwd_flow_entry_t entry;
    cvm_ipfwd_flow_bucket_t *bucket = NULL;
    struct iphdr *ip;
    struct tcphdr *th;
    cvm_ipfwd_flow_info_t *info;
    int idx, status;
    int skip_fastpath;
#endif

    /* Prefetch cvm_oct_device since we know we need it soon */
    CVMX_PREFETCH(cvm_oct_device, 0);

    /* Save scratch in case userspace is using it */
    CVMX_SYNCIOBDMA;
    old_scratch = cvmx_scratch_read64(CVMX_SCR_SCRATCH);

    /* Only allow work for our group */
    old_group_mask = cvmx_read_csr(CVMX_POW_PP_GRP_MSKX(coreid));
    cvmx_write_csr(CVMX_POW_PP_GRP_MSKX(coreid), 1<<pow_receive_group);

    cvmx_pow_work_request_async(CVMX_SCR_SCRATCH, CVMX_POW_NO_WAIT);

    while (1)
    {
        void *start_of_buffer;
        struct sk_buff *skb;
        cvm_oct_callback_result_t callback_result;
        cvmx_wqe_t *work = cvmx_pow_work_response_async(CVMX_SCR_SCRATCH);
        if (work == NULL)
            break;

        /* Limit each core to processing 60 packets without a break. This way
            the RX can't starve the TX task. */
#ifndef CVM_IPFWD_DRIVER_OFFLOAD
        if (likely(rx_count < 48))
#else
        if (likely(rx_count < 48))
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
            cvmx_pow_work_request_async_nocheck(CVMX_SCR_SCRATCH, CVMX_POW_NO_WAIT);
        else
        {
            cvmx_scratch_write64(CVMX_SCR_SCRATCH, 0x8000000000000000ull);
            cvmx_pow_tag_sw_null_nocheck();
        }
        rx_count++;

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
        if (ipfwd_driver_offload) {
            bucket = &cvm_gbl_ipfwd_cache[work->tag & CVM_IPFWD_LUT_MASK];
            CVMX_PREFETCH0(bucket);
        }
#ifdef CVM_IPFWD_DYNAMIC_SHORT
        if (work->word2.s.bufs == 0)
        {
            /* Dynamic short pre-adjustment */
            work->packet_ptr.u64 = 0;
            work->packet_ptr.s.size = CVMX_FPA_WQE_POOL_SIZE;
            work->packet_ptr.s.pool = CVMX_FPA_WQE_POOL;
            if (work->word2.s.not_IP)
            {
                /* Non-IP */
                work->packet_ptr.s.addr = (uint64_t)(&work->packet_data);
            }
            else
            {
                if(!work->word2.s.is_v6)
                {
                    /* IPv4 (shifted four bytes) */
                    work->packet_ptr.s.addr = (uint64_t)(&work->packet_data) + CVM_IPFWD_IP_OFFSET + 4 - work->word2.s.ip_offset;
                }
                else
                    /* IPv6 */
                    work->packet_ptr.s.addr = (uint64_t)(&work->packet_data) + CVM_IPFWD_IP_OFFSET - work->word2.s.ip_offset;
            }
        }
#endif /* CVM_IPFWD_DYNAMIC_SHORT */
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

        CVMX_PREFETCH(cvm_oct_device[work->ipprt], 0);
        start_of_buffer = cvm_oct_get_buffer_ptr(work->packet_ptr);
        CVMX_PREFETCH(start_of_buffer, -8);

        /* Immediately throw away all packets with receive errors except for
            CVMX_PIP_OVER_ERR. CVMX_PIP_OVER_ERR means the packet was good,
            just oversized */
        if (USE_10MBPS_PREAMBLE_WORKAROUND && ((work->word2.snoip.err_code == 5) || (work->word2.snoip.err_code == 7)))
        {
            /* We received a packet with either an alignment error or a
                FCS error. This may be signalling that we are running
                10Mbps with GMXX_RXX_FRM_CTL[PRE_CHK] off. If this is the
                case we need to parse the packet to determine if we can
                remove a non spec preamble and generate a correct packet */
            int interface = INTERFACE(work->ipprt);
            int index = INDEX(work->ipprt);
            cvmx_gmxx_rxx_frm_ctl_t gmxx_rxx_frm_ctl;
            gmxx_rxx_frm_ctl.u64 = cvmx_read_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface));
            if (gmxx_rxx_frm_ctl.s.pre_chk == 0)
            {
                uint8_t *ptr = cvmx_phys_to_ptr(work->packet_ptr.s.addr);
                int i = 0;
                while (i<work->len-1)
                {
                    if (*ptr != 0x55)
                        break;
                    ptr++;
                    i++;
                }
                if (*ptr == 0xd5)
                {
                    //DEBUGPRINT("Port %d received 0xd5 preamble\n", work->ipprt);
                    work->packet_ptr.s.addr += i+1;
                    work->len -= i+5;
                }
                else if ((*ptr & 0xf) == 0xd)
                {
                    //DEBUGPRINT("Port %d received 0x?d preamble\n", work->ipprt);
                    work->packet_ptr.s.addr += i;
                    work->len -= i+4;
                    for (i=0; i<work->len; i++)
                    {
                        *ptr = ((*ptr&0xf0)>>4) | ((*(ptr+1)&0xf)<<4);
                        ptr++;
                    }
                }
                else
                {
                    DEBUGPRINT("Port %d unknown preamble, packet dropped\n", work->ipprt);
                    //cvmx_helper_dump_packet(work);
                    cvm_oct_free_work(work);
                    continue;
                }
            }
        }
        else if (unlikely(work->word2.snoip.rcv_error && (work->word2.snoip.err_code != CVMX_PIP_OVER_ERR)))
        {
            //DEBUGPRINT("Port %d receive error code %d, packet dropped\n", work->ipprt, work->word2.snoip.err_code);
            cvm_oct_free_work(work);
            continue;
        }

        /* We can only use the zero copy path if skbuffs are in the FPA pool
            and the packet fit in a single buffer */
#if USE_SKBUFFS_IN_HW
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
	if (likely(work->word2.s.bufs <= 1))
#else
        if (likely(work->word2.s.bufs == 1))
#endif
        {
            skb = *(struct sk_buff **)(start_of_buffer - sizeof(void*));
            /* This calculation was changed in case the skb header is using a
                different address aliasing type than the buffer. It doesn't make
                any differnece now, but the new one is more correct */
            skb->data = skb->head + work->packet_ptr.s.addr - cvmx_ptr_to_phys(skb->head);
            CVMX_PREFETCH(skb->data, 0);
            skb->len = work->len;
            skb->tail = skb->data + skb->len;

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
            /* 
             * Fast path: Do a flow-based IPFWD lookup 
             *    Packet requirements: 
             *       1) IPv4
             *       2) TCP/UDP (if TCP, no SYN, FIN, or RESET)
             *       3) non-fragment (at least for now)
             */

            skip_fastpath = 0;
            ip = (struct iphdr*) (skb->data + 14); //XXX work->word2.s.ip_offset);
            th = (struct tcphdr *) ((uint32_t*)ip + ip->ihl);
            if ((ip->protocol == IP_PROTOCOL_TCP) && ((th->syn) || (th->fin) || (th->rst)))
                skip_fastpath = 1;

            if ((ipfwd_driver_offload) &&
                    ((!work->word2.s.is_v6) &&
                     (work->word2.s.tcp_or_udp) &&
                     (!work->word2.s.is_frag) &&
                     (!skip_fastpath)))
            {
#if 0
                {
                    cvm_packet_info_t *pinfo = (cvm_packet_info_t *) ((((uint64_t) skb->end +
                                    sizeof(struct skb_shared_info)) + 7) & ~7ull);
                    if (pinfo->cookie == CVM_IPFWD_COOKIE_VALUE)
                        printf("ERROR: rx packet has a valid cookie!\n");
                }
#endif

                //printk("%s: cvm_ipfwd_find_entry(bucket = %p)\n", __func__, bucket);
                idx = cvm_ipfwd_find_entry(bucket, ip, &entry);

                if (idx == CVM_IPFWD_NOT_FOUND)
                    goto not_found;

                /* Match found */

                /* Check its timestamp */
                status = cvm_ipfwd_check_timestamp(entry.s.timestamp);

                if (status == CVM_IPFWD_EXPIRED) {
#if 0
                    ipfwd_slow_path_packets[cvmx_get_core_num()]++;
#endif
                    goto not_found;
                }

                if ((status != CVM_IPFWD_NEW) && (cvm_ipfwd_random_select()))
                {
                    /* Old match found and this packet was selected to go through
                     * the Linux stack.  Make sure there is not already someone
                     * in this bucket doing this. */
                    if (cvm_ipfwd_check_bucket_timestamp(bucket))
                    {
#if 0
                        ipfwd_slow_path_refresh[cvmx_get_core_num()]++;
#endif
                        goto not_found;
                    }
                }

                /* If we get here, entry is not expired */
                if (entry.s.flow_info)
                {
                    /* match found -- format header */
                    info = (cvm_ipfwd_flow_info_t*) cvmx_phys_to_ptr((uint64_t)entry.s.flow_info << 7);
                    eh = (cvm_ipfwd_eh_t *) skb->data;
                    eh->enet_dst_hi = info->enet_dst_hi;
                    eh->enet_dst_lo = info->enet_dst_lo;
                    eh->enet_src_hi = info->enet_src_hi;
                    eh->enet_src_lo = info->enet_src_lo;
                    ip = (struct iphdr*) (skb->data + work->word2.s.ip_offset);
                    ip->ttl--;
                    ip->saddr = info->ip_saddr;
                    ip->daddr = info->ip_daddr;
                    ip->protocol = info->ip_proto;
                    *((uint32_t*) ip + ip->ihl) = info->ports;

                    /* compute IP checksum */
                    ip->check = cvm_ipfwd_calculate_ip_header_checksum((uint16_t*) ip);

                    /* Use the cvm_oct_transmit() to send the packet */
                    cvm_oct_transmit(info->dev, work, 1);

                    /* done */
                    continue;
                }

not_found:
                /* 
                 * No match found (or old match found) -- 
                 * mark the packet (buffer) so that the flow can be
                 * cached on the way out.  
                 */
                cvm_ipfwd_mark_packet(skb, work, bucket);
            } 
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
        }
        else
#endif
        {
            /* We have to copy the packet. First allocate an skbuff for it */
            skb = dev_alloc_skb(work->len);
            if (!skb)
            {
                DEBUGPRINT("Port %d failed to allocate skbuff, packet dropped\n", work->ipprt);
                cvm_oct_free_work(work);
                continue;
            }

            /* Check if we've received a packet that was entirely stored the
                work entry. This is untested */
            if (unlikely(work->word2.s.bufs == 0))
            {
                DEBUGPRINT("Port %d received a work with work->word2.s.bufs=0, untested\n", work->ipprt);
                memcpy(skb_put(skb, work->len), work->packet_data, work->len);
                /* No packet buffers to free */
            }
            else
            {
                int segments = work->word2.s.bufs;
                cvmx_buf_ptr_t segment_ptr = work->packet_ptr;
                int len = work->len;
                while (segments--)
                {
                    cvmx_buf_ptr_t next_ptr = *(cvmx_buf_ptr_t*)cvmx_phys_to_ptr(segment_ptr.s.addr-8);
                    /* Octeon Errata PKI-100: The segment size is wrong. Until it
                        is fixed, calculate the segment size based on the packet
                        pool buffer size. When it is fixed, the following line should
                        be replaced with this one:
                            int segment_size = segment_ptr.s.size; */
                    int segment_size = CVMX_FPA_PACKET_POOL_SIZE - (segment_ptr.s.addr - (((segment_ptr.s.addr >> 7) - segment_ptr.s.back) << 7));
                    /* Don't copy more than what is left in the packet */
                    if (segment_size > len)
                        segment_size = len;
                    /* Copy the data into the packet */
                    memcpy(skb_put(skb, segment_size), cvmx_phys_to_ptr(segment_ptr.s.addr), segment_size);
                    /* Reduce the amount of bytes left to copy */
                    len -= segment_size;
                    segment_ptr = next_ptr;
                }
            }
        }

        if (likely((work->ipprt < TOTAL_NUMBER_OF_PORTS) && cvm_oct_device[work->ipprt]))
        {
            struct net_device *dev = cvm_oct_device[work->ipprt];
            cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

            /* Only accept packets for devices that are currently up */
            if (likely(priv->up))
            {
                skb->protocol = eth_type_trans(skb, dev);
                skb->dev = dev;

                if (unlikely(work->word2.s.not_IP || work->word2.s.IP_exc || work->word2.s.L4_error))
                    skb->ip_summed = CHECKSUM_NONE;
                else
                    skb->ip_summed = CHECKSUM_UNNECESSARY;
                /* Increment RX stats for virtual ports */
                if (work->ipprt >= CVMX_PIP_NUM_INPUT_PORTS)
                {
                    cvmx_atomic_add64_nosync(&priv->stats.rx_packets, 1);
                    cvmx_atomic_add64_nosync(&priv->stats.rx_bytes, skb->len);
                }

                if (priv->intercept_cb)
                {
                    callback_result = priv->intercept_cb(dev, work, skb);
                    switch (callback_result)
                    {
                        case CVM_OCT_PASS:
                            netif_receive_skb(skb);
                            break;
                        case CVM_OCT_DROP:
                            dev_kfree_skb_irq(skb);
                            cvmx_atomic_add64_nosync(&priv->stats.rx_dropped, 1);
                            break;
                        case CVM_OCT_TAKE_OWNERSHIP_WORK:
                            /* Interceptor stole our work, but we need to free
                                the skbuff */
#if USE_SKBUFFS_IN_HW
                            if (likely(work->word2.s.bufs == 1))
                            {
                                /* We can't free the skbuff since its data is
                                    the same as the work. In this case we don't
                                    do anything */
                            }
                            else
                                dev_kfree_skb_irq(skb);
#else
                            dev_kfree_skb_irq(skb);
#endif
                            break;
                        case CVM_OCT_TAKE_OWNERSHIP_SKB:
                            /* Interceptor stole our packet */
                            break;
                    }
                }
                else
                {
                    netif_receive_skb(skb);
                    callback_result = CVM_OCT_PASS;
                }
            }
            else
            {
                /* Drop any packet received for a device that isn't up */
                //DEBUGPRINT("%s: Device not up, packet dropped\n", dev->name);
                cvmx_atomic_add64_nosync(&priv->stats.rx_dropped, 1);
                dev_kfree_skb_irq(skb);
                callback_result = CVM_OCT_DROP;
            }
        }
        else
        {
            /* Drop any packet received for a device that doesn't exist */
            DEBUGPRINT("Port %d not controlled by Linux, packet dropped\n", work->ipprt);
            dev_kfree_skb_irq(skb);
            callback_result = CVM_OCT_DROP;
        }

        /* We only need to free the work if the interceptor didn't take over
            ownership of it */
        if (callback_result != CVM_OCT_TAKE_OWNERSHIP_WORK)
        {
#if USE_SKBUFFS_IN_HW
            /* Check to see if the skbuff and work share the same packet buffer */
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
            if (likely(work->word2.s.bufs <= 1))
#else
            if (likely(work->word2.s.bufs == 1))
#endif
            {
                /* This buffer needs to be replaced, increment the number of buffers we need to free by one */
                cvmx_fau_atomic_add32(FAU_NUM_PACKET_BUFFERS_TO_FREE, 1);

#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
                if (likely(work->word2.s.bufs != 0))
                    cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));
#else
                cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));
#endif
            }
            else
                cvm_oct_free_work(work);
#else
            cvm_oct_free_work(work);
#endif
        }
    }

    /* Restore the original POW group mask */
    cvmx_write_csr(CVMX_POW_PP_GRP_MSKX(coreid), old_group_mask);
    /* Restore the scratch area */
    cvmx_scratch_write64(CVMX_SCR_SCRATCH, old_scratch);

#if USE_SKBUFFS_IN_HW
    /* Refill the packet buffer pool */
    number_to_free = cvmx_fau_fetch_and_add32(FAU_NUM_PACKET_BUFFERS_TO_FREE, 0);
    if (number_to_free>0)
    {
        cvmx_fau_atomic_add32(FAU_NUM_PACKET_BUFFERS_TO_FREE, -number_to_free);
        while (number_to_free--)
            cvm_oct_free_hw_buffer(CVMX_FPA_PACKET_POOL, CVMX_FPA_PACKET_POOL_SIZE);
    }
#endif
    cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0x1ful<<24);
}


/**
 * Open a device for use. Device should be able to send and
 * receive packets after this is called.
 *
 * @param dev    Device to bring up
 * @return Zero on success
 */
static int cvm_oct_open(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    priv->up = 1;

    /* For RGMII ports, we enable them when they come up */
    if (priv->isRGMII)
    {
        cvmx_gmxx_prtx_cfg_t gmx_cfg;
        gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(INDEX(priv->port), INTERFACE(priv->port)));
        gmx_cfg.s.en = 1;
        cvmx_write_csr(CVMX_GMXX_PRTX_CFG(INDEX(priv->port), INTERFACE(priv->port)), gmx_cfg.u64);
    }

    return 0;
}


/**
 * Stop an ethernet device. No more packets should be
 * received from this device.
 *
 * @param dev    Device to bring down
 * @return Zero on success
 */
static int cvm_oct_stop(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

    /* For RGMII ports, Disable the port when it goes down */
    if (priv->isRGMII)
    {
        cvmx_gmxx_prtx_cfg_t gmx_cfg;
        gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(INDEX(priv->port), INTERFACE(priv->port)));
        gmx_cfg.s.en = 0;
        cvmx_write_csr(CVMX_GMXX_PRTX_CFG(INDEX(priv->port), INTERFACE(priv->port)), gmx_cfg.u64);
    }

    priv->up = 0;
    return 0;
}


/**
 * Get the low level ethernet statistics
 *
 * @param dev    Device to get the statistics from
 * @return Pointer to the statistics
 */
static struct net_device_stats *cvm_oct_get_stats(struct net_device *dev)
{
    cvmx_pip_port_status_t rx_status;
    cvmx_pko_port_status_t tx_status;
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

    if (priv->port < CVMX_PIP_NUM_INPUT_PORTS)
    {
#ifndef CONFIG_CAVIUM_OCTEON_SIMULATOR
        cvmx_pip_get_port_status(priv->port, 1, &rx_status);
        cvmx_pko_get_port_status(priv->port, 1, &tx_status);
#else
        /* The simulator doesn't support statistics */
        memset(&rx_status, 0, sizeof(rx_status));
        memset(&tx_status, 0, sizeof(tx_status));
#endif

        priv->stats.rx_packets      += rx_status.inb_packets;
        priv->stats.tx_packets      += tx_status.packets;
        priv->stats.rx_bytes        += rx_status.inb_octets;
        priv->stats.tx_bytes        += tx_status.octets;
        priv->stats.multicast       += rx_status.multicast_packets;
        priv->stats.rx_crc_errors   += rx_status.inb_errors;
        priv->stats.rx_frame_errors += rx_status.fcs_align_err_packets;

        /* The drop counter must be incremented atomically since the RX
            tasklet also increments it */
        cvmx_atomic_add64_nosync(&priv->stats.rx_dropped, rx_status.dropped_packets);
    }

    return &priv->stats;
}


/**
 * Periodic timer tick for slow management operations
 *
 * @param arg    Device to check
 */
static void cvm_do_timer(unsigned long arg)
{
    int port;

    /* Loop through all the ports and check the RGMII link status */
    for (port=0; port<CVMX_PIP_NUM_INPUT_PORTS; port++)
    {
        if (cvm_oct_device[port])
        {
#ifndef CONFIG_CAVIUM_OCTEON_SIMULATOR
            cvm_oct_private_t* priv = (cvm_oct_private_t*)cvm_oct_device[port]->priv;
            /* Check link negotiation */
            if (priv->isRGMII)
                cvm_oct_configure_rgmii_speed(cvm_oct_device[port]);
#endif
            cvm_oct_get_stats(cvm_oct_device[port]);
        }

    }

    /* Repeat every two seconds */
    mod_timer(&cvm_oct_poll_timer, jiffies + HZ);
}


/**
 * Set the multicast list. Currently unimplemented.
 *
 * @param dev    Device to work on
 */
static void cvm_oct_set_multicast_list(struct net_device *dev)
{
    cvmx_gmxx_prtx_cfg_t gmx_cfg;
    int interface;
    int index;
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

    if (priv->isRGMII)
    {
        uint64_t control = 1;   /* Allow broadcast MAC addresses */

        if (dev->mc_list || (dev->flags&IFF_ALLMULTI) || (dev->flags&IFF_PROMISC))
            control |= 2<<1;    /* Force accept multicast packets */
        else
            control |= 1<<1;    /* Force reject multicat packets */

        if(dev->flags&IFF_PROMISC)
            control |= 0<<3;    /* Reject matches if promisc. Since CAM is shut off, should accept everything */
        else
            control |= 1<<3;    /* Filter packets based on the CAM */

        interface = INTERFACE(priv->port);
        index = INDEX(priv->port);
        gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));
        cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64 & ~1ull);

        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CTL(index, interface), control);
        if (dev->flags&IFF_PROMISC)
            cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM_EN(index, interface), 0);
        else
            cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM_EN(index, interface), 1);

        cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);
    }
}


/**
 * Set the hardware MAC address for a device
 *
 * @param dev    Device to change the MAC address for
 * @param addr   Address structure to change it too. MAC address is addr + 2.
 * @return Zero on success
 */
static int cvm_oct_set_mac_address(struct net_device *dev, void *addr)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    cvmx_gmxx_prtx_cfg_t gmx_cfg;
    int interface;
    int index;

    memcpy(dev->dev_addr, addr + 2, 6);

    if (priv->isRGMII)
    {
        int i;
        uint8_t *ptr = addr;
        uint64_t mac = 0;
        for (i=0; i<6; i++)
            mac = (mac<<8) | (uint64_t)(ptr[i+2]);

        interface = INTERFACE(priv->port);
        index = INDEX(priv->port);
        gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));
        cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64 & ~1ull);

        cvmx_write_csr(CVMX_GMXX_SMACX(index, interface), mac);
        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM0(index, interface), ptr[2]);
        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM1(index, interface), ptr[3]);
        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM2(index, interface), ptr[4]);
        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM3(index, interface), ptr[5]);
        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM4(index, interface), ptr[6]);
        cvmx_write_csr(CVMX_GMXX_RXX_ADR_CAM5(index, interface), ptr[7]);
        cvm_oct_set_multicast_list(dev);
        cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);
    }
    return 0;
}


/**
 * Change the link MTU. Unimplemented
 *
 * @param dev     Device to change
 * @param new_mtu The new MTU
 * @return Zero on success
 */
static int cvm_oct_change_mtu(struct net_device *dev, int new_mtu)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

    /* Limit the MTU to make sure the ethernet packets are between 64 bytes
        and 65535 bytes */
    if ((new_mtu + 14 + 4 < 64) || (new_mtu + 14 + 4 > 65392))
    {
        printk("MTU must be between %d and %d.\n", 64-14-4, 65392-14-4);
        return -EINVAL;
    }
    dev->mtu = new_mtu;

    if (priv->isRGMII)
    {
        /* Set the hardware to truncate packets larger than the MTU. The jabber
            register must be set to a multiple of 8 bytes, so round up */
        int interface = INTERFACE(priv->port);
        int index = INDEX(priv->port);
        cvmx_write_csr(CVMX_GMXX_RXX_JABBER(index, interface), ((new_mtu + 14 + 4) + 7) & ~7u);
    }

    if (octeon_is_pass1() && (new_mtu>CVMX_FPA_PACKET_POOL_SIZE - CVMX_HELPER_FIRST_MBUFF_SKIP - 32))
        printk("Warning: Octeon Pass 1 has problems with chained buffers. You may lose buffers with jumbo frames\n");

    return 0;
}


/**
 * Perform an MII read. Called by the generic MII routines
 *
 * @param dev      Device to perform read for
 * @param phy_id   The MII phy id
 * @param location Register location to read
 * @return Result from the read or zero on failure
 */
static int cvm_oct_mdio_read(struct net_device *dev, int phy_id, int location)
{
    cvmx_smi_cmd_t          smi_cmd;
    cvmx_smi_rd_dat_t       smi_rd;

    smi_cmd.u64 = 0;
    smi_cmd.s.phy_op = 1;
    smi_cmd.s.phy_adr = phy_id;
    smi_cmd.s.reg_adr = location;
    cvmx_write_csr(CVMX_SMI_CMD, smi_cmd.u64);

    do
    {
        if (!in_interrupt())
            yield();
        smi_rd.u64 = cvmx_read_csr(CVMX_SMI_RD_DAT);
    } while (smi_rd.s.pending);

    if (smi_rd.s.val)
        return smi_rd.s.dat;
    else
        return 0;
}


/**
 * Perform an MII write. Called by the generic MII routines
 *
 * @param dev      Device to perform write for
 * @param phy_id   The MII phy id
 * @param location Register location to write
 * @param val      Value to write
 */
static void cvm_oct_mdio_write(struct net_device *dev, int phy_id, int location, int val)
{
    cvmx_smi_cmd_t          smi_cmd;
    cvmx_smi_wr_dat_t       smi_wr;

    smi_wr.u64 = 0;
    smi_wr.s.dat = val;
    cvmx_write_csr(CVMX_SMI_WR_DAT, smi_wr.u64);

    smi_cmd.u64 = 0;
    smi_cmd.s.phy_op = 0;
    smi_cmd.s.phy_adr = phy_id;
    smi_cmd.s.reg_adr = location;
    cvmx_write_csr(CVMX_SMI_CMD, smi_cmd.u64);

    do
    {
        if (!in_interrupt())
            yield();
        smi_wr.u64 = cvmx_read_csr(CVMX_SMI_WR_DAT);
    } while (smi_wr.s.pending);
}


/**
 * IOCTL support for PHY control
 *
 * @param dev    Device to change
 * @param rq     the request
 * @param cmd    the command
 * @return Zero on success
 */
static int cvm_oct_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
    cvm_oct_private_t*      priv = (cvm_oct_private_t*)dev->priv;
    struct mii_ioctl_data*  data = if_mii(rq);
    unsigned int            duplex_chg;

    if (priv->isRGMII)
        return generic_mii_ioctl(&priv->mii_info, data, cmd, &duplex_chg);
    else
        return -EOPNOTSUPP;
}


static void cvm_oct_get_drvinfo(struct net_device *dev, struct ethtool_drvinfo *info)
{
    strcpy(info->driver, "cavium-ethernet");
    strcpy(info->version, cvm_oct_get_version());
    strcpy(info->bus_info, "Builtin");
}


static int cvm_oct_get_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    return mii_ethtool_gset(&priv->mii_info, cmd);
}


static int cvm_oct_set_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    return mii_ethtool_sset(&priv->mii_info, cmd);
}


static int cvm_oct_nway_reset(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    return mii_nway_restart(&priv->mii_info);
}


static u32 cvm_oct_get_link(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    return mii_link_ok(&priv->mii_info);
}


static struct ethtool_ops cvm_oct_ethtool_ops = {
    .get_drvinfo    = cvm_oct_get_drvinfo,
    .get_settings   = cvm_oct_get_settings,
    .set_settings   = cvm_oct_set_settings,
    .nway_reset     = cvm_oct_nway_reset,
    .get_link       = cvm_oct_get_link,
    .get_sg         = ethtool_op_get_sg,
    .get_tx_csum    = ethtool_op_get_tx_csum,
};


/**
 * Per network device initialization
 *
 * @param dev    Device to initialize
 * @return Zero on success
 */
static int __init cvm_oct_init(struct net_device *dev)
{
    static int count = 0;

    char mac[8] = {0x00,0x00,
        octeon_bootinfo->mac_addr_base[0],
        octeon_bootinfo->mac_addr_base[1],
        octeon_bootinfo->mac_addr_base[2],
        octeon_bootinfo->mac_addr_base[3],
        octeon_bootinfo->mac_addr_base[4],
        octeon_bootinfo->mac_addr_base[5] + count};
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

    if (priv->queue != -1)
    {
        dev->hard_start_xmit= cvm_oct_xmit;
        if (USE_HW_TCPUDP_CHECKSUM)
            dev->features |= NETIF_F_IP_CSUM;
        count++;
#ifndef CONFIG_CAVIUM_OCTEON_SIMULATOR
        if (count > octeon_bootinfo->mac_addr_count)
            printk("ERROR: Board was allocated %d mac addresses, currently using %d\n",
                   octeon_bootinfo->mac_addr_count, count);
#endif
    }
    else
    {
        dev->hard_start_xmit= cvm_oct_xmit_pow;
        memset(mac, 0, sizeof(mac));
        mac[7] = pow_send_group;
    }

    dev->get_stats          = cvm_oct_get_stats;
    dev->open               = cvm_oct_open;
    dev->stop               = cvm_oct_stop;
    dev->weight             = 32;
    dev->set_mac_address    = cvm_oct_set_mac_address;
    dev->set_multicast_list = cvm_oct_set_multicast_list;
    dev->change_mtu         = cvm_oct_change_mtu;
    dev->do_ioctl           = cvm_oct_ioctl;
    dev->poll               = cvm_oct_rx_poll;
#ifdef CONFIG_NET_POLL_CONTROLLER
    dev->poll_controller    = cvm_oct_poll_controller;
#endif
    if (priv->isRGMII)
        SET_ETHTOOL_OPS(dev, &cvm_oct_ethtool_ops);
    dev->features           |= NETIF_F_LLTX; /* We do our own locking, Linux doesn't need to */
    cvm_oct_set_mac_address(dev, mac);
    return 0;
}


/**
 * This function fills a hardware pool with memory. Depending
 * on the config defines, this memory might come from the
 * kernel or global 32bit memory allocated with
 * cvmx_bootmem_alloc.
 *
 * @param pool     Pool to populate
 * @param size     Size of each buffer in the pool
 * @param elements Number of buffers to allocate
 */
static void cvm_oct_fill_hw_memory(uint64_t pool, uint64_t size, uint64_t elements)
{
    char *memory;

#if USE_32BIT_SHARED
    extern uint64_t octeon_reserve32_memory;

    memory = cvmx_bootmem_alloc_range(elements*size, 128, octeon_reserve32_memory, octeon_reserve32_memory + (CONFIG_CAVIUM_RESERVE32<<20) - 1);
    if (memory == NULL)
        panic("Unable to allocate %lu bytes for FPA pool %ld\n", elements*size, pool);
    printk("Memory range %p - %p reserved for hardware\n", memory, memory + elements*size - 1);
    while (elements--)
    {
        cvmx_fpa_free(memory, pool, 0);
        memory += size;
    }

#else
    while (elements--)
    {
        /* We need to force alignment to 128 bytes here */
        memory = kmalloc(size + 127, GFP_ATOMIC);
        memory = (char*)(((unsigned long)memory+127) & -128);
        if (memory == NULL)
            panic("Unable to allocate %lu bytes for FPA pool %ld\n", elements*size, pool);
        cvmx_fpa_free(memory, pool, 0);
    }
#endif
}


/**
 * Free memory previously allocated with cvm_oct_fill_hw_memory
 *
 * @param pool     FPA pool to free
 * @param size     Size of each buffer in the pool
 * @param elements Number of buffers that should be in the pool
 */
static void cvm_oct_free_hw_memory(uint64_t pool, uint64_t size, int elements)
{
#if USE_32BIT_SHARED
    printk("Warning: 32 shared memory is not freeable\n");
#else
    uint64_t top_bits;
    char *memory = kmalloc(size, GFP_ATOMIC);
    if (memory == NULL)
        panic("Unable to get memory for top bits reference\n");
    top_bits = ((uint64_t)memory >> 36) << 36;
    kfree(memory);
    do
    {
        memory = cvmx_fpa_alloc(pool);
        if (memory)
        {
            elements--;
            memory = (char*)(top_bits | ((uint64_t)memory & 0xfffffffffull));
            kfree(memory);
        }
    } while (memory);

    if (elements < 0)
        printk("Freeing of pool %lu had too many buffers (%d)\n", pool, elements);
    else if (elements > 0)
        printk("Warning: Freeing of pool %lu is missing %d buffers\n", pool, elements);
#endif
}


#ifdef USE_RED
/**
 * Setup Random Early Drop to automatically begin dropping packets.
 *
 * @param pass_thresh
 *               Packets will begin slowly dropping when there are less than
 *               this many packet buffers free in FPA 0.
 * @param drop_thresh
 *               All incomming packets will be dropped when there are less
 *               than this many free packet buffers in FPA 0.
 * @return Zero on success. Negative on failure
 */
static int cvm_oct_setup_red(int pass_thresh, int drop_thresh)
{
    /* Disable backpressure based on queued buffers. It needs SW support */
    cvmx_ipd_qos_red_marks_t red_marks;
    cvmx_ipd_red_quex_param_t red_param;
    cvmx_ipd_red_port_enable_t red_port_enable;

    /* Set RED to begin dropping packets when there are 1024 buffers left. It
        will linearly drop more packets until reaching 512 buffers */
    red_marks.u64 = 0;
    red_marks.s.drop = drop_thresh;
    red_marks.s.pass = pass_thresh;
    cvmx_write_csr(CVMX_IPD_QOS0_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS1_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS2_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS3_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS4_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS5_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS6_RED_MARKS, red_marks.u64);
    cvmx_write_csr(CVMX_IPD_QOS7_RED_MARKS, red_marks.u64);

    /* Use the actual queue 0 counter, not the average */
    red_param.u64 = 0;
    red_param.s.prb_con = (255ul<<24) / (red_marks.s.pass - red_marks.s.drop);
    red_param.s.avg_con = 1;
    red_param.s.new_con = 255;
    red_param.s.use_pcnt = 1;
    cvmx_write_csr(CVMX_IPD_RED_QUE0_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE1_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE2_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE3_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE4_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE5_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE6_PARAM, red_param.u64);
    cvmx_write_csr(CVMX_IPD_RED_QUE7_PARAM, red_param.u64);

    /* Shutoff the dropping based on the per port page count. SW isn't
        decrementing it right now */
    cvmx_write_csr(CVMX_IPD_BP_PRT_RED_END, 0);

    red_port_enable.u64 = 0;
    red_port_enable.s.prt_enb = 0xfffffffffull;
    red_port_enable.s.avg_dly = 10000;
    red_port_enable.s.prb_dly = 10000;
    cvmx_write_csr(CVMX_IPD_RED_PORT_ENABLE, red_port_enable.u64);

    return 0;
}
#endif


/**
 * Configure common hardware for all interfaces
 */
static void cvm_oct_configure_common_hw(void)
{
    /* Setup the FPA */
    cvmx_fpa_enable();

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
#ifdef CVM_IPFWD_DYNAMIC_SHORT
    cvm_oct_fill_hw_skbuff(CVMX_FPA_WQE_POOL, CVMX_FPA_WQE_POOL_SIZE, NUM_PACKET_BUFFERS);
#else
    cvm_oct_fill_hw_memory(CVMX_FPA_WQE_POOL, CVMX_FPA_WQE_POOL_SIZE, NUM_PACKET_BUFFERS);
#endif
    /* Setup FPA pool for flow info buffers */
    cvm_oct_fill_hw_memory(CVMX_FPA_FLOW_INFO_POOL, CVMX_FPA_FLOW_INFO_POOL_SIZE,
	  (CVM_IPFWD_LUT_BUCKETS*CVM_IPFWD_ENTRY_RETRIES));
#else
    cvm_oct_fill_hw_memory(CVMX_FPA_WQE_POOL, CVMX_FPA_WQE_POOL_SIZE, NUM_PACKET_BUFFERS);
#endif
#if CVMX_FPA_OUTPUT_BUFFER_POOL != CVMX_FPA_PACKET_POOL
    cvm_oct_fill_hw_memory(CVMX_FPA_OUTPUT_BUFFER_POOL, CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE, 128);
#endif

#if USE_SKBUFFS_IN_HW
    cvm_oct_fill_hw_skbuff(CVMX_FPA_PACKET_POOL, CVMX_FPA_PACKET_POOL_SIZE, NUM_PACKET_BUFFERS);
#else
    cvm_oct_fill_hw_memory(CVMX_FPA_PACKET_POOL, CVMX_FPA_PACKET_POOL_SIZE, NUM_PACKET_BUFFERS);
#endif

#ifdef USE_RED
    cvm_oct_setup_red(NUM_PACKET_BUFFERS/4, NUM_PACKET_BUFFERS/8);
#endif

    if (num_ethernet_devices)
    {
        cvmx_pko_pool_cfg_t config;
        cvmx_ipd_config(CVMX_FPA_PACKET_POOL_SIZE/8,
                        CVMX_HELPER_FIRST_MBUFF_SKIP/8,
                        CVMX_HELPER_NOT_FIRST_MBUFF_SKIP/8,
                        CVMX_HELPER_FIRST_MBUFF_SKIP / 128,
                        CVMX_HELPER_NOT_FIRST_MBUFF_SKIP / 128,
                        CVMX_FPA_WQE_POOL,
                        CVMX_IPD_OPC_MODE_STF,
                        CVMX_HELPER_ENABLE_BACK_PRESSURE);

        /* Set the PKO to think command buffers are an odd length. This makes
            it so we never have to divide a comamnd across two buffers */
        config.u64 = 0;
        config.s.pool = CVMX_FPA_OUTPUT_BUFFER_POOL;
#if USE_OPTIMIZED_PKO_BUFFER_SIZE
        config.s.size = CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE / 8 - 1;
#else
        config.s.size = CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE / 8;
#endif
        cvmx_write_csr(CVMX_PKO_REG_CMD_BUF, config.u64);

        if (cvm_oct_has_4ports())
        {
            /* PKO registers */
            cvmx_pko_reg_gmx_port_mode_t pko_mode;
            pko_mode.u64 = 0;
            pko_mode.s.mode0 = 2;   /* 16 >> 2 == 4 ports */
            pko_mode.s.mode1 = 2;   /* 16 >> 2 == 4 ports */
            cvmx_write_csr(CVMX_PKO_REG_GMX_PORT_MODE, pko_mode.u64);
        }
        cvmx_pko_enable();
    }

    /* Pass 1 PKI-12 Errata: Ignore ipv4 header checksum violations */
    if (octeon_is_pass1())
    {
        cvmx_pip_gbl_ctl_t global_control;
        global_control.u64 = cvmx_read_csr(CVMX_PIP_GBL_CTL);
        global_control.s.ip_chk = 0;
        cvmx_write_csr(CVMX_PIP_GBL_CTL, global_control.u64);
    }

#ifndef CONFIG_CAVIUM_OCTEON_SIMULATOR
    /* Enable the MII interface */
    cvmx_write_csr(CVMX_SMI_EN, 1);
#endif

    /* Register an IRQ hander for to receive POW interrupts */
    request_irq(8 + pow_receive_group, cvm_oct_do_interrupt, SA_SHIRQ, "Ethernet", cvm_oct_device);

    /* Register an IRQ hander for to receive GMX change interrupts */
    request_irq(8 + 46, cvm_oct_rml_interrupt, SA_SHIRQ, "Ethernet RML", cvm_oct_device);

#if USE_MULTICORE_RECEIVE
    preempt_disable();
    {
        int cpu;
        for (cpu=0; cpu<NR_CPUS; cpu++)
        {
            if (cpu_online(cpu) && (cpu != smp_processor_id()))
            {
                cvmx_ciu_intx0_t en;
                en.u64 = cvmx_read_csr(CVMX_CIU_INTX_EN0(cpu_logical_map(cpu)*2));
                en.s.workq |= (1<<pow_receive_group);
                cvmx_write_csr(CVMX_CIU_INTX_EN0(cpu_logical_map(cpu)*2), en.u64);
            }
        }
    }
    preempt_enable();
#endif
}


/**
 * Configure per interface settings
 *
 * @param interface Interface to configure (0 or 1)
 */
static int cvm_oct_configure_interface(int interface, int isRGMII)
{
    if (isRGMII)
    {
        if (!cvm_oct_has_4ports())
            cvmx_write_csr(CVMX_GMXX_TX_PRTS(interface), 3);
        else
            cvmx_write_csr(CVMX_GMXX_TX_PRTS(interface), 4);
        /* Disable ASX until auto negotiation is complete. It will
            be enabled when the PHY speeds is checked */
        cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(interface), 0);
        cvmx_write_csr(CVMX_ASXX_RX_PRT_EN(interface), 0);
        return 0;
    }
    else
    {
        int result = cvmx_spi_start_interface(interface, CVMX_SPI_MODE_DUPLEX, 10);
        if (result == 0)
        {
            result = cvmx_spi4000_initialize(interface);
            if (result)
            {
                cvmx_pko_reg_crc_enable_t enable;
                printk("\t\tSPI4000 not found\n");
                enable.u64 = cvmx_read_csr(CVMX_PKO_REG_CRC_ENABLE);
                enable.s.enable |= 0xffff << (interface*16);
                cvmx_write_csr(CVMX_PKO_REG_CRC_ENABLE, enable.u64);
            }

            return 0;
        }
        else
            return -1;
    }
}


/**
 * Configure the RGMII port for the negotiated speed
 *
 * @param dev    Linux device for the RGMII port
 */
static void cvm_oct_configure_rgmii_speed(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    int interface = INTERFACE(priv->port);
    int index = INDEX(priv->port);
    const char *link_duplex;
    const char *link_speed;
    cvmx_gmxx_prtx_cfg_t gmx_cfg;
#ifdef WN302T_PHY_FIX
     unsigned short volatile phy_val=0x00;
#endif

 
#ifdef CONFIG_CAVIUM_OCTEON_SIMULATOR

    link_duplex = "Full";
    link_speed = "  1Gbs";

    /* Enable ASX */
    cvmx_write_csr(CVMX_ASXX_RX_PRT_EN(interface), cvmx_read_csr(CVMX_ASXX_RX_PRT_EN(interface)) | (1<<index));
    cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(interface), cvmx_read_csr(CVMX_ASXX_TX_PRT_EN(interface)) | (1<<index));

    /* Enable GMX */
    gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));
    gmx_cfg.s.en = priv->up;
    cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);

#else

    int mdio;
    int queue;
    int stop_queue;
    cvmx_gmxx_rxx_int_en_t gmx_en;
    cvmx_gmxx_rxx_rx_inbnd_t link;
    cvmx_gmxx_tx_ovr_bp_t ovr_bp;
    unsigned long flags;

    /* If the 10Mbps preamble workaround is supported and we're at 10Mbps
        we may need to do some special checking */
    if (USE_10MBPS_PREAMBLE_WORKAROUND && (cvmx_read_csr(CVMX_GMXX_TXX_CLK(index, interface)) == 50))
    {
        /* Read the GMXX_RXX_INT_REG[PCTERR] bit and see if we are getting
            preamble errors */
        int interface = INTERFACE(priv->port);
        int index = INDEX(priv->port);
        cvmx_gmxx_rxx_int_reg_t gmxx_rxx_int_reg;
        gmxx_rxx_int_reg.u64 = cvmx_read_csr(CVMX_GMXX_RXX_INT_REG(index, interface));
        if (gmxx_rxx_int_reg.s.pcterr)
        {
            /* We are getting preamble errors at 10Mbps. Most likely the
                PHY is giving us packets with mis aligned preambles. In
                order to get these packets we need to disable preamble
                checking and do it in software */
            cvmx_gmxx_rxx_frm_ctl_t gmxx_rxx_frm_ctl;
            cvmx_ipd_sub_port_fcs_t ipd_sub_port_fcs;

            /* Disable preamble checking */
            gmxx_rxx_frm_ctl.u64 = cvmx_read_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface));
            gmxx_rxx_frm_ctl.s.pre_chk = 0;
            cvmx_write_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface), gmxx_rxx_frm_ctl.u64);
            /* Disable FCS stripping */
            ipd_sub_port_fcs.u64 = cvmx_read_csr(CVMX_IPD_SUB_PORT_FCS);
            ipd_sub_port_fcs.s.port_bit &= 0xffffffffull ^ (1ull<<priv->port);
            cvmx_write_csr(CVMX_IPD_SUB_PORT_FCS, ipd_sub_port_fcs.u64);
            /* Clear any error bits */
            cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmxx_rxx_int_reg.u64);
            DEBUGPRINT("%s: Using 10Mbps with software preamble removal\n", dev->name);
        }
    }

    /* Check to see if ASX TX is enabled. If it is, the link is up and doesn't
        need any changes. The RML interrupt handler would have disabled ASX if
        the status changed */
    if (likely(cvmx_read_csr(CVMX_ASXX_TX_PRT_EN(interface)) & (1<<index))) {
#ifdef WN302T_PHY_FIX 
	if(priv->port ==0x0) {
		phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0xa);
		if( phy_val & 0xC00 ) {
		    gmx_en.u64 = cvmx_read_csr(CVMX_GMXX_RXX_INT_EN(index, interface));
		    if((gmx_en.s.phy_dupx!=1) && (gmx_en.s.phy_dupx!=1) && (gmx_en.s.phy_spd!=1)) {
			gmx_en.s.phy_dupx = 1;
			gmx_en.s.phy_link = 1;
			gmx_en.s.phy_spd = 1;
			cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmx_en.u64);
			cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), gmx_en.u64);
		    }
		 }
		 else {		
		    phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0x5);
		    if( phy_val & 0x180 ) {
			gmx_en.u64 = cvmx_read_csr(CVMX_GMXX_RXX_INT_EN(index, interface));
			if((gmx_en.s.phy_dupx!=1) && (gmx_en.s.phy_link!=1) && (gmx_en.s.phy_spd!=1)) {
			    gmx_en.s.phy_dupx = 1;
			    gmx_en.s.phy_link = 1;
			    gmx_en.s.phy_spd = 1;
			    cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmx_en.u64);
			    cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), gmx_en.u64);
			}
		    }
		    else if( !(phy_val & 0x20) && !(phy_val & 0x40)) {
			gmx_en.u64 = cvmx_read_csr(CVMX_GMXX_RXX_INT_EN(index, interface));
			if((gmx_en.s.phy_dupx!=1) && (gmx_en.s.phy_link!=1) && (gmx_en.s.phy_spd!=1)) {
			    gmx_en.s.phy_dupx = 1;
			    gmx_en.s.phy_link = 1;
			    gmx_en.s.phy_spd = 1;
			    cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmx_en.u64);
			    cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), gmx_en.u64);
			}
			
			/*disable link itself, because we lose a disconnect interrupt before enabling interrupt mask for link status.*/
			    /* Disable the lowest level output from the port */
			    cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(interface),
					   cvmx_read_csr(CVMX_ASXX_TX_PRT_EN(interface)) & ~(1<<index));

			    /* Disable the lowest level input from the port. Note that
				this also disables inband status */
			    cvmx_write_csr(CVMX_ASXX_RX_PRT_EN(interface),
					   cvmx_read_csr(CVMX_ASXX_RX_PRT_EN(interface)) & ~(1<<index));

			    /* Disable packet output by disabling all queues for this port */
			    queue = cvmx_pko_get_base_queue(priv->port);
			    stop_queue = queue + cvmx_pko_get_num_queues(priv->port);
			    while (queue < stop_queue)
			    {
				cvmx_pko_mem_queue_qos_t pko_qos;
				pko_qos.u64 = 0;
				pko_qos.s.pid = priv->port;
				pko_qos.s.qid = queue;
				pko_qos.s.qos_mask = 0;
				cvmx_write_csr(CVMX_PKO_REG_READ_IDX, queue);
				cvmx_write_csr(CVMX_PKO_MEM_QUEUE_QOS, pko_qos.u64);
				queue++;
			    }

			    /* Disable backpressure for this port */
			    ovr_bp.u64 = cvmx_read_csr(CVMX_GMXX_TX_OVR_BP(interface));
			    ovr_bp.s.en |= 1<<index;
			    ovr_bp.s.bp &= ~(1<<index);
			    cvmx_write_csr(CVMX_GMXX_TX_OVR_BP(interface), ovr_bp.u64);

			    /* Disable the interrupt for this port. It will be enabled
				once the port is brought back up */
			    cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), 0);

			    /* We're done with the global registers */
			    spin_unlock_irqrestore(&global_register_lock, flags);

			    /* Inform the user and kernel of the change */
			    if (cvm_oct_device[priv->port])
			    {
				printk("%s: Link down : MDIO probe\n", cvm_oct_device[priv->port]->name);
				if (netif_carrier_ok(cvm_oct_device[priv->port]))
				    netif_carrier_off(cvm_oct_device[priv->port]);
			    }	    
			}   /* if link is 10M from reading 0x5 */
		    }   /* if link is not Giga from reading 0xA */
	}    /* likely - cvmx_read_csr */

#endif  /* WN302T_PHY_FIX */
        return;
    }

    /*PATCH1*/
    if(netif_carrier_ok(dev))
        netif_carrier_off(dev);

#ifdef WN302T_PHY_FIX 
     /* Read the values from BCM PHY and simulate the values like Marvell Phy */
    if(priv->port ==0x00) {
        phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0x02);
        if(phy_val != 0x143) {   
#ifdef WN302T_PHY_FIX_DEBUG 
	        printk("Could not read Phy ID...Bailing Out \r\n");
#endif
        }	
        /* Read Status  register */
        phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0x01);

#ifdef WN302T_PHY_FIX_DEBUG 
        printk("Status reg is %x \r\n",phy_val);
#endif
        /* return if Auto Negotioataion is in process */
        if((phy_val & (1<<5))) {
	        /* Auto Negotiate Complete */
        } else {

#ifdef WN302T_PHY_FIX_DEBUG 
	        printk("Phy Auto Negotitaion is in progress \r\n");
#endif
	        return;
        }

        /* Check if link is up, If not Restart the Auto Negotitaion process */
        if(phy_val & (1<<2)) {
            /* link is up */
        } else {
            /* Restart the Auto Negotitaion */
            // phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0x00);
            // phy_val |= ( 1<<12);
            // cvm_oct_mdio_write(dev, priv->mii_info.phy_id, 0x00,phy_val);
            // if (netif_carrier_ok(dev))
            //   netif_carrier_off(dev);

    #ifdef WN302T_PHY_FIX_DEBUG 
            printk("Returning...Looks like Link is down \r\n");
    #endif
            return;

        }

        /* If we are here then link is up */
        mdio = ((1<<10)|(1<<11));

        /* read link status register */
        phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0xA);

    #ifdef WN302T_PHY_FIX_DEBUG 
        printk("gigabit status is %x\r\n",phy_val); 
    #endif  
        if(phy_val & 0x800) {
            /* 1Gbps full duplex */
            mdio |= (2<<14)|(1<<13);

        } else if(phy_val & 0x400) {
            /* 1Gbps Half duplex */
            mdio |= (2<<14);
        }else {

            phy_val = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 0x5);
     
    #ifdef WN302T_PHY_FIX_DEBUG 
            printk("100Mbps  status is %x\r\n",phy_val); 
    #endif       
            if(phy_val & 0x100) {
                 /* 100 Mbps full duplex */
                mdio |= ((1<<14)|(1<<13));
       
            } else if(phy_val & 0x80) {
                 /* 100 Mbps Half duplex */
                mdio |= (1<<14);
            }else if(phy_val & 0x40) {

                 /* 10 Mbps Full duplex */
                mdio |= (1<<13);
            }else {
                /* 10 Mbps half duplex */
            }
        }
    } else {
        mdio = ((1<<13) /* Full duplex */
                | (2<<14) /* 1Gbps */
                | (1<<11) /* resolved */
                | (1<<10)); /* link up */
    }

#else

    /* ASX RX is not enabled. We can't finish auto negotiation until the link
        is valid. First we need to poll the PHY and see if auto neg is complete */
    mdio = cvm_oct_mdio_read(dev, priv->mii_info.phy_id, 17);

    /* Force MDIO result for failed reads */
    if ((mdio == 0) || (mdio == 0xffff))
        mdio = ((1<<13) /* Full duplex */
                | (2<<14) /* 1Gbps */
                | (1<<11) /* resolved */
                | (1<<10)); /* link up */

    /* Return if the PHY hasn't resolved yet (Register 17.11) */
    if (!(mdio & (1<<11)))
        return;

    /* Return if the link is down (Register 17.10) */
    /*PATCH2*/
    if (!(mdio & (1<<10)))
        return;

#endif

    /* Take the global register lock since we are going to touch
        registers that affect more than one port */
    spin_lock_irqsave(&global_register_lock, flags);

    /* If the 10Mbps preamble workaround is allowed we need to on preamble checking,
        FCS stripping, and clear error bits on every speed change. If errors occur
        during 10Mbps operation the above cod will change this stuff */
    if (USE_10MBPS_PREAMBLE_WORKAROUND)
    {
        cvmx_gmxx_rxx_frm_ctl_t gmxx_rxx_frm_ctl;
        cvmx_ipd_sub_port_fcs_t ipd_sub_port_fcs;
        cvmx_gmxx_rxx_int_reg_t gmxx_rxx_int_reg;
        int interface = INTERFACE(priv->port);
        int index = INDEX(priv->port);

        /* Enable preamble checking */
        gmxx_rxx_frm_ctl.u64 = cvmx_read_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface));
        gmxx_rxx_frm_ctl.s.pre_chk = 1;
        cvmx_write_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface), gmxx_rxx_frm_ctl.u64);
        /* Enable FCS stripping */
        ipd_sub_port_fcs.u64 = cvmx_read_csr(CVMX_IPD_SUB_PORT_FCS);
        ipd_sub_port_fcs.s.port_bit |= 1ull<<priv->port;
        cvmx_write_csr(CVMX_IPD_SUB_PORT_FCS, ipd_sub_port_fcs.u64);
        /* Clear any error bits */
        gmxx_rxx_int_reg.u64 = cvmx_read_csr(CVMX_GMXX_RXX_INT_REG(index, interface));
        cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmxx_rxx_int_reg.u64);
    }

    if (!(cvmx_read_csr(CVMX_ASXX_RX_PRT_EN(interface)) & (1<<index)))
    {
        /* Poll the GMX state machine. If it isn't idle, we can't reconfigure
            yet. The current state is avaialble through the magic debug register */
        cvmx_write_csr(CVMX_NPI_DBG_SELECT, interface*0x800 + index*0x100 + 0x880);
        if ((cvmx_read_csr(CVMX_DBG_DATA) & 0x7) != 0)
        {
            printk("%s: GMX state machine 1 not idle\n", dev->name);
            spin_unlock_irqrestore(&global_register_lock, flags);
            return;
        }
        cvmx_write_csr(CVMX_NPI_DBG_SELECT, interface*0x800 + index*0x100 + 0x8c1);
        if ((cvmx_read_csr(CVMX_DBG_DATA) & 0xf) != 0)
        {
            printk("%s: GMX state machine 2 not idle\n", dev->name);
            spin_unlock_irqrestore(&global_register_lock, flags);
            return;
        }
    }

    /* Disable GMX before a reconfig */
    gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));
    gmx_cfg.s.en = 0;
    cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);
    cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));

    /* If ASX RX is disabled, enable it and return. This gives inband
        status a chance to become valid. We'll finish setup on the next call */
    if (!(cvmx_read_csr(CVMX_ASXX_RX_PRT_EN(interface)) & (1<<index)))
    {
        cvmx_write_csr(CVMX_ASXX_RX_PRT_EN(interface),
                       cvmx_read_csr(CVMX_ASXX_RX_PRT_EN(interface)) | (1<<index));
        spin_unlock_irqrestore(&global_register_lock, flags);
        //DEBUGPRINT("%s: Enabling inband status\n", dev->name);
        return;
    }

    /* Check for a couple special cases where MDIO doesn't give us the correct
        status */
    if ((octeon_bootinfo->board_type == CVMX_BOARD_TYPE_TRANTOR) ||
        (cvm_oct_is_gmii_mode() && (priv->port == 1)))
    {
        /* Force to gig for Trantor board  and GMII */
        mdio = ((1<<13) /* Full duplex */
                | (2<<14)) /* 1Gbps */;
    }
    else if (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_CN3005_EVB_HS5)
    {
        /* WAN PHY MDIO does not work, so fake it here.
        ** Set LAN(switch) to 1 Gig, FDX */
        mdio = 0;
        if (priv->port == 0)
        {
            link.u64 = cvmx_read_csr(CVMX_GMXX_RXX_RX_INBND(index, interface));
            mdio = (link.s.speed << 14) | (link.s.duplex << 13) | (link.s.status << 10) | (1 << 11);

        }
        else if (priv->port == 1)
        {
            mdio = ((1<<13) /* Full duplex */
                    | (2<<14) /* 1Gbps */
                    | (1<<11) /* resolved */
                    | (1<<10)); /* link up */
        }
    }
    else
    {
#ifndef WN302T_PHY_FIX
        /* Make sure inband status match MDIO */
        link.u64 = cvmx_read_csr(CVMX_GMXX_RXX_RX_INBND(index, interface));
        if (!link.s.status || (link.s.duplex != ((mdio>>13)&1)) || (link.s.speed != ((mdio>>14)&3)))
        {
            /* Either the inband status says the link is down, the duplex doesn't
                match, or the speed doesn't match. Try again later */
            spin_unlock_irqrestore(&global_register_lock, flags);
            printk("%s: Inband status doesn't match the PHY, waiting mdio=0x%x link=0x%lx\n", dev->name, mdio, link.u64);
            return;
        }
#endif

    }

    /* Set the duplex to match the PHY (Register 17.13) */
    gmx_cfg.s.duplex = ((mdio & (1<<13)) != 0);

    if (gmx_cfg.s.duplex)
        link_duplex = "Full";
    else
        link_duplex = "Half";

    /* Set the speed (Register 17.15-14) */
    if ((mdio>>14) == 0)
    {
        link_speed = " 10Mbs";
        gmx_cfg.s.slottime = 0;
        gmx_cfg.s.speed = 0;
        cvmx_write_csr(CVMX_GMXX_TXX_CLK(index, interface), 50);
        cvmx_write_csr(CVMX_GMXX_TXX_SLOT(index, interface), 0x40);
        cvmx_write_csr(CVMX_GMXX_TXX_BURST(index, interface), 0);
    }
    else if ((mdio>>14) == 1)
    {
        link_speed = "100Mbs";
        gmx_cfg.s.slottime = 0;
        gmx_cfg.s.speed = 0;
        cvmx_write_csr(CVMX_GMXX_TXX_CLK(index, interface), 5);
        cvmx_write_csr(CVMX_GMXX_TXX_SLOT(index, interface), 0x40);
        cvmx_write_csr(CVMX_GMXX_TXX_BURST(index, interface), 0);
    }
    else if ((mdio>>14) == 2)
    {
        link_speed = "  1Gbs";
        gmx_cfg.s.slottime = 1;
        gmx_cfg.s.speed = 1;
        cvmx_write_csr(CVMX_GMXX_TXX_CLK(index, interface), 1);
        cvmx_write_csr(CVMX_GMXX_TXX_SLOT(index, interface), 0x200);
        cvmx_write_csr(CVMX_GMXX_TXX_BURST(index, interface), 0x2000);
    }
    else
    {
        link_speed = " Rsrvd";
        gmx_cfg.s.slottime = 1;
        gmx_cfg.s.speed = 1;
        cvmx_write_csr(CVMX_GMXX_TXX_CLK(index, interface), 1);
        cvmx_write_csr(CVMX_GMXX_TXX_SLOT(index, interface), 0x200);
        cvmx_write_csr(CVMX_GMXX_TXX_BURST(index, interface), 0x2000);
    }

    /* Update GMX with the final speed settings */
    cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);
    cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));

    /* Enable ASX TX */
    cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(interface),
                   cvmx_read_csr(CVMX_ASXX_TX_PRT_EN(interface)) | (1<<index));

    /* Enable GMX again now that the config is done */
    gmx_cfg.s.en = priv->up;
    cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);

    /* Enable packet output by enabling all queues for this port */
    queue = cvmx_pko_get_base_queue(priv->port);
    stop_queue = queue + cvmx_pko_get_num_queues(priv->port);
    while (queue < stop_queue)
    {
        cvmx_pko_mem_queue_qos_t pko_qos;
        pko_qos.u64 = 0;
        pko_qos.s.pid = priv->port;
        pko_qos.s.qid = queue;
        pko_qos.s.qos_mask = 0xff;
        cvmx_write_csr(CVMX_PKO_REG_READ_IDX, queue);
        cvmx_write_csr(CVMX_PKO_MEM_QUEUE_QOS, pko_qos.u64);
        queue++;
    }

    /* Enable backpressure for this port */
    ovr_bp.u64 = cvmx_read_csr(CVMX_GMXX_TX_OVR_BP(interface));
    ovr_bp.s.en &= ~(1<<index);
    cvmx_write_csr(CVMX_GMXX_TX_OVR_BP(interface), ovr_bp.u64);

    /* Enable the Inband status interrupt to tells us about changes */
    gmx_en.u64 = 0;
    gmx_en.s.phy_dupx = 1;
    gmx_en.s.phy_link = 1;
    gmx_en.s.phy_spd = 1;
    cvmx_write_csr(CVMX_GMXX_RXX_INT_REG(index, interface), gmx_en.u64);
    cvmx_write_csr(CVMX_GMXX_RXX_INT_EN(index, interface), gmx_en.u64);

    spin_unlock_irqrestore(&global_register_lock, flags);

#endif

    /* Link is good to go. Tell Linux */
    /*PATCH3*/
    netif_carrier_on(dev);

    if (priv->queue != -1)
        DEBUGPRINT("%s: %s %s duplex, port %2d, queue %2d\n",
               dev->name, link_speed, link_duplex, priv->port, priv->queue);
    else
        DEBUGPRINT("%s: %s %s duplex, port %2d, POW\n",
               dev->name, link_speed, link_duplex, priv->port);
}


/**
 * Configure an individual RGMII interface
 *
 * @param dev    Linux ethernet device for the RGMII
 */
static void cvm_oct_configure_rgmii_port(struct net_device *dev)
{
    cvmx_gmxx_prtx_cfg_t gmx_cfg;
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    int interface = INTERFACE(priv->port);
    int index = INDEX(priv->port);

    cvm_oct_configure_rgmii_speed(dev);

#ifndef CONFIG_CAVIUM_OCTEON_SIMULATOR
    cvmx_write_csr(CVMX_GMXX_TXX_THRESH(index, interface), 32);

    if (octeon_is_pass1())
    {
        /* Set hi water mark as per errata GMX-4 */
        if (octeon_bootinfo->eclock_hz >= 325000000 && octeon_bootinfo->eclock_hz < 375000000)
            cvmx_write_csr(CVMX_ASXX_TX_HI_WATERX(index, interface), 12);
        else if (octeon_bootinfo->eclock_hz >= 375000000 && octeon_bootinfo->eclock_hz < 437000000)
            cvmx_write_csr(CVMX_ASXX_TX_HI_WATERX(index, interface), 11);
        else if (octeon_bootinfo->eclock_hz >= 437000000 && octeon_bootinfo->eclock_hz < 550000000)
            cvmx_write_csr(CVMX_ASXX_TX_HI_WATERX(index, interface), 10);
        else if (octeon_bootinfo->eclock_hz >= 550000000 && octeon_bootinfo->eclock_hz < 687000000)
            cvmx_write_csr(CVMX_ASXX_TX_HI_WATERX(index, interface), 9);
        else
            printk("Unable to set high water marks. Unsupported frequency %u\n", octeon_bootinfo->eclock_hz);
    }

    /* Use relaxed RGMII preamble checking, pass2 only */
    if (!octeon_is_pass1())
    {
        cvmx_gmxx_rxx_frm_ctl_t frm_ctl;
        frm_ctl.u64 = cvmx_read_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface));
        frm_ctl.s.pre_free = 1;  /* Pass 2 only field */
        cvmx_write_csr(CVMX_GMXX_RXX_FRM_CTL(index, interface), frm_ctl.u64);
    }

    if (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_CN3005_EVB_HS5 && index == 1)
    {
        /* Set RGMII timing for switch port */
        cvmx_write_csr(CVMX_ASXX_TX_CLK_SETX(index, interface), 0);
        cvmx_write_csr(CVMX_ASXX_RX_CLK_SETX(index, interface), 0);
    }
    else if (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_TRANTOR)
    {
        cvmx_write_csr(CVMX_ASXX_TX_CLK_SETX(index, interface), 0);
        cvmx_write_csr(CVMX_ASXX_RX_CLK_SETX(index, interface), 16);
    }
    else
    {
        cvmx_write_csr(CVMX_ASXX_TX_CLK_SETX(index, interface), 24);
        cvmx_write_csr(CVMX_ASXX_RX_CLK_SETX(index, interface), 24);
    }
#endif

#ifdef CONFIG_WN302T_DUAL_BAND_AP
    cvmx_write_csr(CVMX_ASXX_TX_CLK_SETX(index, interface), 0xd);
    cvmx_write_csr(CVMX_ASXX_RX_CLK_SETX(index, interface), 0xe);
#endif

    /* Disable the port in case the bootloader leaves it enabled */
    gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(index, interface));
    gmx_cfg.s.en = 0;
    cvmx_write_csr(CVMX_GMXX_PRTX_CFG(index, interface), gmx_cfg.u64);
}


/**
 * Perform any SPI per port config. Currently there isn't any.
 *
 * @param dev    Device to configure
 */
static void cvm_oct_configure_spi_port(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    if (priv->queue != -1)
        printk("\t\t%s: SPI port %2d, queue %2d\n", dev->name, priv->port, priv->queue);
    else
        printk("\t\t%s: SPI port %2d, POW\n", dev->name, priv->port);
}


/**
 * Perform per port configuration.
 *
 * @param dev    Device to configure
 */
static void cvm_oct_configure_port(struct net_device *dev)
{
    cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;
    const uint64_t priorities[8] = {8,8,8,8,8,8,8,8};
    cvmx_pip_port_cfg_t     port_config;
    cvmx_pip_port_tag_cfg_t tag_config;

    if (priv->isRGMII)
        cvm_oct_configure_rgmii_port(dev);
    else
        cvm_oct_configure_spi_port(dev);

    /* Packet output configures Queue and Ports */
    cvmx_pko_config_port(priv->port, cvmx_pko_get_base_queue(priv->port),
                         cvmx_pko_get_num_queues(priv->port), priorities);

    /* Setup Port input tagging */
    port_config.u64 = 0;
    port_config.s.mode = CVMX_PIP_PORT_CFG_MODE_SKIPL2; /* Process the headers and place the IP header in the work queue */
    port_config.s.qos = priv->port & 0x7;
    port_config.s.crc_en = 1;
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
    /* Configure port for dynamic short */
    port_config.s.dyn_rs = 1;
#endif

    tag_config.u64 = 0;
    tag_config.s.ip6_src_flag  = CVMX_HELPER_INPUT_TAG_IPV6_SRC_IP;
    tag_config.s.ip6_dst_flag  = CVMX_HELPER_INPUT_TAG_IPV6_DST_IP;
    tag_config.s.ip6_sprt_flag = CVMX_HELPER_INPUT_TAG_IPV6_SRC_PORT;
    tag_config.s.ip6_dprt_flag = CVMX_HELPER_INPUT_TAG_IPV6_DST_PORT;
    tag_config.s.ip6_nxth_flag = CVMX_HELPER_INPUT_TAG_IPV6_NEXT_HEADER;
    tag_config.s.ip4_src_flag  = CVMX_HELPER_INPUT_TAG_IPV4_SRC_IP;
    tag_config.s.ip4_dst_flag  = CVMX_HELPER_INPUT_TAG_IPV4_DST_IP;
    tag_config.s.ip4_sprt_flag = CVMX_HELPER_INPUT_TAG_IPV4_SRC_PORT;
    tag_config.s.ip4_dprt_flag = CVMX_HELPER_INPUT_TAG_IPV4_DST_PORT;
    tag_config.s.ip4_pctl_flag = CVMX_HELPER_INPUT_TAG_IPV4_PROTOCOL;
    tag_config.s.inc_prt_flag  = CVMX_HELPER_INPUT_TAG_INPUT_PORT;
    tag_config.s.tcp6_tag_type = CVMX_HELPER_INPUT_TAG_TYPE;
    tag_config.s.tcp4_tag_type = CVMX_HELPER_INPUT_TAG_TYPE;
    tag_config.s.ip6_tag_type = CVMX_HELPER_INPUT_TAG_TYPE;
    tag_config.s.ip4_tag_type = CVMX_HELPER_INPUT_TAG_TYPE;
    tag_config.s.non_tag_type = CVMX_HELPER_INPUT_TAG_TYPE;
    tag_config.s.grp = pow_receive_group;

    cvmx_pip_config_port(priv->port, port_config, tag_config);
}


/**
 * Registers a intercept callback for the names ethernet
 * device. It returns the Linux device structure for the
 * ethernet port. Usign a callback of NULL will remove
 * the callback. Note that this callback must not disturb
 * scratch. It will be called with SYNCIOBDMAs in progress
 * and userspace may be using scratch. It also must not
 * disturb the group mask.
 *
 * @param device_name
 *                 Device name to register for. (Example: "eth0")
 * @param callback Intercept callback to set.
 * @return Device structure for the ethernet port or NULL on failure.
 */
struct net_device *cvm_oct_register_callback(const char *device_name, cvm_oct_callback_t callback)
{
    int port;

    for (port=0; port<TOTAL_NUMBER_OF_PORTS; port++)
    {
        if (cvm_oct_device[port] && (strcmp(device_name, cvm_oct_device[port]->name) == 0))
        {
            cvm_oct_private_t* priv = (cvm_oct_private_t*)cvm_oct_device[port]->priv;
            priv->intercept_cb = callback;
            wmb();
            return cvm_oct_device[port];
        }
    }

    return NULL;
}
EXPORT_SYMBOL(cvm_oct_register_callback);


/**
 * Free a work queue entry received in a intercept callback.
 *
 * @param work_queue_entry
 *               Work queue entry to free
 * @return Zero on success, Negative on failure.
 */
int cvm_oct_free_work(void *work_queue_entry)
{
    cvmx_wqe_t *work = work_queue_entry;

    int segments = work->word2.s.bufs;
    cvmx_buf_ptr_t segment_ptr = work->packet_ptr;

#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
    if (segments > 0)
    {
#endif
    while (segments--)
    {
        cvmx_buf_ptr_t next_ptr = *(cvmx_buf_ptr_t*)cvmx_phys_to_ptr(segment_ptr.s.addr-8);
        if (unlikely(!segment_ptr.s.i))
            cvmx_fpa_free(cvm_oct_get_buffer_ptr(segment_ptr), segment_ptr.s.pool, DONT_WRITEBACK(CVMX_FPA_PACKET_POOL_SIZE/128));
        segment_ptr = next_ptr;
    }
    cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
    }
#endif

    return 0;
}
EXPORT_SYMBOL(cvm_oct_free_work);



/**
 * Transmit a work queue entry out of the ethernet port. Both
 * the work queue entry and the packet data can optionally be
 * freed. The work will be freed on error as well.
 *
 * @param dev     Device to transmit out.
 * @param work_queue_entry
 *                Work queue entry to send
 * @param do_free True if the work queue entry and packet data should be
 *                freed. If false, neither will be freed.
 * @return Zero on success, negative on failure.
 */
int cvm_oct_transmit(struct net_device *dev, void *work_queue_entry, int do_free)
{
    cvmx_buf_ptr_t              hw_buffer;
    cvmx_pko_command_word0_t    pko_command;
    uint64_t                    old_scratch;
    int                         dropped;
    cvm_oct_private_t*          priv = (cvm_oct_private_t*)dev->priv;
    cvmx_wqe_t *                work = work_queue_entry;

    if (!priv->up)
    {
        DEBUGPRINT("%s: Device not up\n", dev->name);
        if (do_free)
            cvm_oct_free_work(work);
        return -1;
    }

    /* Start off assuming no drop */
    dropped = 0;

    spin_lock(&priv->tx_free_list.lock);

    /* Save scratch in case userspace is using it */
    CVMX_SYNCIOBDMA;
    old_scratch = cvmx_scratch_read64(CVMX_SCR_SCRATCH);

    /* Build the PKO buffer pointer */
    hw_buffer.u64 = 0;
    hw_buffer.s.addr = work->packet_ptr.s.addr;
    hw_buffer.s.pool = CVMX_FPA_PACKET_POOL;
    hw_buffer.s.size = CVMX_FPA_PACKET_POOL_SIZE;
    hw_buffer.s.back = work->packet_ptr.s.back;

    /* Build the PKO command */
    pko_command.u64 = 0;
    pko_command.s.dontfree = !do_free;
#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
    if (work->word2.s.bufs == 0)
    {
        pko_command.s.segs = 1;
        do_free = 0;
    }
    else 
        pko_command.s.segs = work->word2.s.bufs;
#else
    pko_command.s.segs = work->word2.s.bufs;
#endif
    pko_command.s.total_bytes = work->len;

    /* Check if we can use the hardware checksumming */
    if (unlikely(work->word2.s.not_IP || work->word2.s.IP_exc))
        pko_command.s.ipoffp1 = 0;
    else
        pko_command.s.ipoffp1 = sizeof(struct ethhdr) + 1;

    /* Get the position for writing the PKO command. This will be used when we
        call cvm_oct_low_level_send. If we drop, this must be decremented */
    cvmx_fau_async_fetch_and_add64(CVMX_SCR_SCRATCH, CVMX_FAU_REG_OQ_ADDR_INDEX + priv->queue*8, 2);

    /* Send the packet to the output queue */
    if (unlikely(cvm_oct_low_level_send(dev, pko_command, hw_buffer)))
    {
        DEBUGPRINT("%s: Failed to send the packet\n", dev->name);
        dropped=-1;
    }

    /* Restore the scratch area */
    cvmx_scratch_write64(CVMX_SCR_SCRATCH, old_scratch);

    if (unlikely(dropped))
    {
        if (do_free)
            cvm_oct_free_work(work);
        cvmx_fau_atomic_add64(CVMX_FAU_REG_OQ_ADDR_INDEX + priv->queue*8, -2);
        priv->stats.tx_dropped++;
    }
    else if (do_free)
        cvmx_fpa_free(work, CVMX_FPA_WQE_POOL, DONT_WRITEBACK(1));

    spin_unlock(&priv->tx_free_list.lock);

    return dropped;
}
EXPORT_SYMBOL(cvm_oct_transmit);


static uint64_t cvm_oct_stats_read_switch(struct net_device *dev, int phy_id, int offset)
{
    cvm_oct_mdio_write(dev, phy_id, 0x1d, 0xcc00 | offset);
    return ((uint64_t)cvm_oct_mdio_read(dev, phy_id, 0x1e)<<16) | (uint64_t)cvm_oct_mdio_read(dev, phy_id, 0x1f);
}

/**
 * User is reading /proc/octeon_ethernet_stats
 *
 * @param m
 * @param v
 * @return
 */
static int cvm_oct_stats_show(struct seq_file *m, void *v)
{
    int port;

    for (port=0; port<TOTAL_NUMBER_OF_PORTS; port++)
    {
        if (cvm_oct_device[port])
        {
            cvm_oct_private_t* priv = (cvm_oct_private_t*)cvm_oct_device[port]->priv;
            seq_printf(m, "\nOcteon Port %d (%s)\n", port, cvm_oct_device[port]->name);
            seq_printf(m,
                       "rx_packets:             %12lu\t"
                       "tx_packets:             %12lu\n",
                       priv->stats.rx_packets,
                       priv->stats.tx_packets);
            seq_printf(m,
                       "rx_bytes:               %12lu\t"
                       "tx_bytes:               %12lu\n",
                       priv->stats.rx_bytes,
                       priv->stats.tx_bytes);
            seq_printf(m,
                       "rx_errors:              %12lu\t"
                       "tx_errors:              %12lu\n",
                       priv->stats.rx_errors,
                       priv->stats.tx_errors);
            seq_printf(m,
                       "rx_dropped:             %12lu\t"
                       "tx_dropped:             %12lu\n",
                       priv->stats.rx_dropped,
                       priv->stats.tx_dropped);
            seq_printf(m,
                       "rx_length_errors:       %12lu\t"
                       "tx_aborted_errors:      %12lu\n",
                       priv->stats.rx_length_errors,
                       priv->stats.tx_aborted_errors);
            seq_printf(m,
                       "rx_over_errors:         %12lu\t"
                       "tx_carrier_errors:      %12lu\n",
                       priv->stats.rx_over_errors,
                       priv->stats.tx_carrier_errors);
            seq_printf(m,
                       "rx_crc_errors:          %12lu\t"
                       "tx_fifo_errors:         %12lu\n",
                       priv->stats.rx_crc_errors,
                       priv->stats.tx_fifo_errors);
            seq_printf(m,
                       "rx_frame_errors:        %12lu\t"
                       "tx_heartbeat_errors:    %12lu\n",
                       priv->stats.rx_frame_errors,
                       priv->stats.tx_heartbeat_errors);
            seq_printf(m,
                       "rx_fifo_errors:         %12lu\t"
                       "tx_window_errors:       %12lu\n",
                       priv->stats.rx_fifo_errors,
                       priv->stats.tx_window_errors);
            seq_printf(m,
                       "rx_missed_errors:       %12lu\t"
                       "multicast:              %12lu\n",
                       priv->stats.rx_missed_errors,
                       priv->stats.multicast);
        }
    }

    if (cvm_oct_is_gmii_mode() || (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_KODAMA))
    {
        static const int ports[] = {0,1,2,3,9,-1};
        struct net_device *dev = cvm_oct_device[0];
        int index = 0;

        while (ports[index] != -1)
        {
            /* Latch port */
            cvm_oct_mdio_write(dev, 0x1b, 0x1d, 0xdc00 | ports[index]);
            seq_printf(m, "\nSwitch Port %d\n", ports[index]);
            seq_printf(m, "InGoodOctets:   %12lu\t"
                       "OutOctets:      %12lu\t"
                       "64 Octets:      %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x00) | (cvm_oct_stats_read_switch(dev, 0x1b, 0x01)<<32),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x0E) | (cvm_oct_stats_read_switch(dev, 0x1b, 0x0F)<<32),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x08));

            seq_printf(m, "InBadOctets:    %12lu\t"
                       "OutUnicast:     %12lu\t"
                       "65-127 Octets:  %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x02),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x10),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x09));

            seq_printf(m, "InUnicast:      %12lu\t"
                       "OutBroadcasts:  %12lu\t"
                       "128-255 Octets: %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x04),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x13),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x0A));

            seq_printf(m, "InBroadcasts:   %12lu\t"
                       "OutMulticasts:  %12lu\t"
                       "256-511 Octets: %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x06),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x12),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x0B));

            seq_printf(m, "InMulticasts:   %12lu\t"
                       "OutPause:       %12lu\t"
                       "512-1023 Octets:%12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x07),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x15),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x0C));

            seq_printf(m, "InPause:        %12lu\t"
                       "Excessive:      %12lu\t"
                       "1024-Max Octets:%12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x16),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x11),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x0D));

            seq_printf(m, "InUndersize:    %12lu\t"
                       "Collisions:     %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x18),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x1E));

            seq_printf(m, "InFragments:    %12lu\t"
                       "Deferred:       %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x19),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x05));

            seq_printf(m, "InOversize:     %12lu\t"
                       "Single:         %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x1A),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x14));

            seq_printf(m, "InJabber:       %12lu\t"
                       "Multiple:       %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x1B),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x17));

            seq_printf(m, "In RxErr:       %12lu\t"
                       "OutFCSErr:      %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x1C),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x03));

            seq_printf(m, "InFCSErr:       %12lu\t"
                       "Late:           %12lu\n",
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x1D),
                       cvm_oct_stats_read_switch(dev, 0x1b, 0x1F));
            index++;
        }
    }
    return 0;
}


/**
 * /proc/octeon_ethernet_stats was openned. Use the single_open iterator
 *
 * @param inode
 * @param file
 * @return
 */
static int cvm_oct_stats_open(struct inode *inode, struct file *file)
{
	return single_open(file, cvm_oct_stats_show, NULL);
}


static struct file_operations cvm_oct_stats_operations = {
	.open		= cvm_oct_stats_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};


/**
 * Module/ driver initialization. Creates the linux network
 * devices.
 *
 * @return Zero on success
 */
static int __init cvm_oct_init_module(void)
{
    int max_devices = num_ethernet_devices;
    int interface;
    int fau = FAU_NUM_PACKET_BUFFERS_TO_FREE - sizeof(uint32_t);
    int phy_id = 0;
    int i;
    struct proc_dir_entry *entry;

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    int j;
    cvm_ipfwd_flow_bucket_t *bucket;
    cvm_ipfwd_flow_entry_t *flow_entry;
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

    printk("Octeon ethernet driver version: %s\n", cvm_oct_get_version());

    entry = create_proc_entry("octeon_ethernet_stats", 0, NULL);
    if (entry)
        entry->proc_fops = &cvm_oct_stats_operations;

    /* Initialize all of the tasklets */
    for (i=0; i<NR_CPUS; i++)
        tasklet_init(cvm_oct_tasklet + i, cvm_oct_tasklet_rx, 0);

    memset(cvm_oct_device, 0, sizeof(cvm_oct_device));

    /* Initialize the FAU used for counting packet buffers that need to be freed */
    cvmx_fau_atomic_write32(FAU_NUM_PACKET_BUFFERS_TO_FREE, 0);

    cvm_oct_configure_common_hw();

    if ((pow_send_group != -1))
    {
        struct net_device *dev;
        printk("\tConfiguring device for POW only access\n");
        dev = alloc_etherdev(sizeof(cvm_oct_private_t));
        if (dev)
        {
            int err;
            cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

            SET_MODULE_OWNER(dev);
            dev->init = cvm_oct_init;

            /* Initialize the device private structure. */
            memset(priv, 0, sizeof(cvm_oct_private_t));
            priv->isRGMII = 0;
            priv->port = CVMX_PIP_NUM_INPUT_PORTS;
            priv->queue = -1;
            strcpy(dev->name, "pow%d");

            err = register_netdev(dev);
            if (err<0)
            {
                printk("\t\tFailed to register ethernet device for POW\n");
                kfree(dev);
            }
            else
            {
                cvm_oct_device[CVMX_PIP_NUM_INPUT_PORTS] = dev;
                skb_queue_head_init(&priv->tx_free_list);
                printk("\t\t%s: POW send group %d, receive group %d\n",
                       dev->name, pow_send_group, pow_receive_group);
            }
        }
        else
        {
            printk("\t\tFailed to allocate ethernet device for POW\n");
        }
    }

    if (num_ethernet_devices)
    {
        for (interface=0; interface<2; interface++)
        {
            cvmx_gmxx_inf_mode_t mode;
            mode.u64 = cvmx_read_csr(CVMX_GMXX_INF_MODE(interface));
            if (mode.s.en)
            {
                const char *name;
                int num_ports;
                int port;
                int isRGMII;

                if (cvm_oct_is_gmii_mode())
                {
                    printk("\tInterface %d is GMII\n", interface);
                    name = "eth%d";
                    num_ports = 2;
                    isRGMII = 1;
                }
                else if (mode.s.type)
                {
                    printk("\tInterface %d is SPI4\n", interface);
                    name = "spi%d";
                    num_ports = 10;
                    isRGMII = 0;

                    if (skip_spi_interfaces)
                    {
		        printk("\t\tSkipping SPI4 initialization\n");
                        num_ports = 0;
                    }
                }
                else
                {
                    printk("\tInterface %d is RGMII\n", interface);
                    name = "eth%d";
                    isRGMII = 1;
                    if (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_KODAMA)
                        num_ports = 2;
                    else if (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_CN3005_EVB_HS5)
                        num_ports = 2;
                    else if (cvm_oct_has_4ports())
                        num_ports = 4;
                    else
                        num_ports = 3;
                }

                if (num_ports)
                {
                    if (cvm_oct_configure_interface(interface, isRGMII))
                        num_ports = 0;
                }
                for (port=interface*16; port<interface*16+num_ports; port++)
                {
                    struct net_device *dev;
                    if (max_devices <= 0)
                        continue;

                    dev = alloc_etherdev(sizeof(cvm_oct_private_t));
                    if (dev)
                    {
                        int err;
                        cvm_oct_private_t* priv = (cvm_oct_private_t*)dev->priv;

                        SET_MODULE_OWNER(dev);
                        dev->init = cvm_oct_init;

                        /* Initialize the device private structure. */
                        memset(priv, 0, sizeof(cvm_oct_private_t));
                        priv->isRGMII = isRGMII;
                        priv->port = port;
                        priv->link_status = -1;
                        priv->queue = cvmx_pko_get_base_queue(priv->port);
                        priv->intercept_cb = NULL;
                        priv->mii_info.dev = dev;
                        if (isRGMII)
			{
                            /*PATCH4*/
			    netif_carrier_off(dev);
                            priv->mii_info.phy_id = phy_id++;
			}
                        if (octeon_bootinfo->board_type == CVMX_BOARD_TYPE_KODAMA)
                            priv->mii_info.phy_id = (priv->port == 0) ? 9 : 4;
                        else if (cvm_oct_is_gmii_mode()) {
#ifdef WN302T_PHY_FIX
	 		unsigned short volatile phy_val=0x00;
	       		phy_val = cvm_oct_mdio_read(dev, 0, 0x02);
			if (phy_val == 0x143) {
				first_phy_id = 0;
				/* printk("PHYID 0 \n"); */
			} else {
        		phy_val = cvm_oct_mdio_read(dev, 1, 0x02);
				if (phy_val == 0x143) {
					first_phy_id = 1;
					/* printk("PHYID 1 \n"); */
				}
			}

                            priv->mii_info.phy_id = first_phy_id;
#else
                            priv->mii_info.phy_id = (priv->port == 1) ? 9 : 4;
#endif
                        }
                        priv->mii_info.phy_id_mask = 0xff;
                        priv->mii_info.supports_gmii = 1;
                        priv->mii_info.reg_num_mask = 0x1f;
                        priv->mii_info.mdio_read = cvm_oct_mdio_read;
                        priv->mii_info.mdio_write = cvm_oct_mdio_write;
                        strcpy(dev->name, name);

                        err = register_netdev(dev);
                        if (err<0)
                        {
                            printk("\t\tFailed to register ethernet device for interface %d, port %d\n", interface, priv->port);
                            kfree(dev);
                        }
                        else
                        {
                            cvm_oct_device[priv->port] = dev;
                            /* Force the interface to use the POW send if always_use_pow was
                                specified or it is in the pow send list */
                            if ((pow_send_group != -1) && (always_use_pow || strstr(pow_send_list, dev->name)))
                            {
                                dev->hard_start_xmit = cvm_oct_xmit_pow;
                                priv->queue = -1;
                            }
                            skb_queue_head_init(&priv->tx_free_list);
                            priv->fau = fau;
                            cvmx_fau_atomic_write32(priv->fau, 0);
                            cvm_oct_configure_port(dev);
                            cvm_oct_change_mtu(dev, dev->mtu);
                            fau -= sizeof(uint32_t);
                            max_devices--;
                        }
                    }
                    else
                    {
                        printk("\t\tFailed to allocate ethernet device for port %d\n", port);
                    }
                }
            }
            else
            {
                printk("\tInterface %d is disabled\n", interface);
            }
        }
    }

#if INTERRUPT_LIMIT
    /* Set the POW timer rate to give an interrupt at most INTERRUPT_LIMIT times per second */
    cvmx_write_csr(CVMX_POW_WQ_INT_PC, octeon_bootinfo->eclock_hz/(INTERRUPT_LIMIT*16*256)<<8);

    /* Enable POW timer interrupt. It will count when there are packets available */
    cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0x1ful<<24);
#else
    /* Enable POW interrupt when our port has at least one packet */
    cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0x1001);
#endif

    /* Enable the poll timer for checking RGMII status */
    init_timer(&cvm_oct_poll_timer);
    cvm_oct_poll_timer.data = 0;
    cvm_oct_poll_timer.function = cvm_do_timer;
    mod_timer(&cvm_oct_poll_timer, jiffies + HZ*2);

#if (defined(CVM_IPFWD_DRIVER_OFFLOAD) && defined(CVM_IPFWD_DYNAMIC_SHORT))
    cvmx_write_csr(CVMX_PIP_IP_OFFSET, CVM_IPFWD_IP_OFFSET/8);
#endif

    cvmx_ipd_enable();

#if FIX_IPD_PTR_ALIGNMENT   /* { */
    /* Workaround for PKI-302 Errata on CN30XX pass 1 and CN31XX pass 1 */
    if (OCTEON_IS_MODEL(OCTEON_CN30XX_PASS1) || OCTEON_IS_MODEL(OCTEON_CN31XX_PASS1))
    {
        const int FIX_IPD_FIRST_BUFF_PAYLOAD_BYTES = CVMX_FPA_PACKET_POOL_SIZE - 8 - CVMX_HELPER_FIRST_MBUFF_SKIP;
        const int FIX_IPD_NON_FIRST_BUFF_PAYLOAD_BYTES = CVMX_FPA_PACKET_POOL_SIZE - 8 - CVMX_HELPER_NOT_FIRST_MBUFF_SKIP;
        const int FIX_IPD_OUTPORT = 1;
        struct sk_buff *skb;
        uint64_t *p64;
        cvmx_buf_ptr_t packet;
        cvmx_pko_command_word0_t pko_command;
        cvmx_wqe_t *work;
        int size, num_segs, wqe_pcnt, pkt_pcnt;
        cvmx_gmxx_prtx_cfg_t gmx_cfg;
        int retry_cnt;
        int retry_loop_cnt;
        int mtu;

        //DEBUGPRINT("Aligning IPD prefetch for work and packet buffers\n");
        cvmx_wait(100000000ull);
        for (retry_loop_cnt = 0;retry_loop_cnt < 10;retry_loop_cnt++)
        {
            wqe_pcnt = cvmx_read_csr(CVMX_IPD_PTR_COUNT);
            pkt_pcnt = (wqe_pcnt >> 7) & 0x7f;
            wqe_pcnt &= 0x7f;

            num_segs = 1 + ((2 + pkt_pcnt - wqe_pcnt) & 3);
            if (num_segs == 1)
                break;

            size = FIX_IPD_FIRST_BUFF_PAYLOAD_BYTES + ((num_segs-1)*FIX_IPD_NON_FIRST_BUFF_PAYLOAD_BYTES) -
                (FIX_IPD_NON_FIRST_BUFF_PAYLOAD_BYTES / 2);

            skb = dev_alloc_skb(size);
            if (skb == NULL)
            {
                DEBUGPRINT("WARNING: FIX_IPD_PTR_ALIGNMENT failed to allocate memory.\n");
                break;
            }

            /* Build the packet */
            p64 = (uint64_t*)skb_put(skb, size);
            memset(p64, 0, size);
            p64[0] = 0xffffffffffff0000ull;
            p64[1] = 0x08004510ull;
            p64[2] = ((uint64_t)(size-14) << 48) | 0x5ae740004000ull;
            p64[3] = 0x3a5fc0a81073c0a8ull;

            /* Build a packet pointer */
            packet.u64 = 0;
            packet.s.addr = cvmx_ptr_to_phys(p64);
            packet.s.pool = CVMX_FPA_PACKET_POOL;
            packet.s.size = size + 1024;

            /* Build the PKO command */
            pko_command.u64 = 0;
            pko_command.s.segs = 1;
            pko_command.s.total_bytes = size;
            pko_command.s.dontfree = 1;

            /* Enable ASX packet loopback */
            cvmx_write_csr(CVMX_ASXX_PRT_LOOP(INTERFACE(FIX_IPD_OUTPORT)), 1 << INDEX(FIX_IPD_OUTPORT));

            /* Enable the port */
            gmx_cfg.u64 = cvmx_read_csr(CVMX_GMXX_PRTX_CFG(INDEX(FIX_IPD_OUTPORT), INTERFACE(FIX_IPD_OUTPORT)));
            gmx_cfg.s.en = 1;
            cvmx_write_csr(CVMX_GMXX_PRTX_CFG(INDEX(FIX_IPD_OUTPORT), INTERFACE(FIX_IPD_OUTPORT)), gmx_cfg.u64);
            cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(INTERFACE(FIX_IPD_OUTPORT)), 1 << INDEX(FIX_IPD_OUTPORT));

            /* Increase the MTU so our packet won't be truncated */
            mtu = cvmx_read_csr(CVMX_GMXX_RXX_JABBER(INDEX(FIX_IPD_OUTPORT), INTERFACE(FIX_IPD_OUTPORT)));
            cvmx_write_csr(CVMX_GMXX_RXX_JABBER(INDEX(FIX_IPD_OUTPORT), INTERFACE(FIX_IPD_OUTPORT)), 65392-14-4);

            /* Make sure everything is done before we send the packet */
            CVMX_SYNC;
            cvm_oct_xmit(skb, cvm_oct_device[FIX_IPD_OUTPORT]);

            /* Read the packet back */
            retry_cnt = 1000;
            do {
                work = cvmx_pow_work_request_sync(CVMX_POW_WAIT);
                retry_cnt--;
            } while ((work == NULL) && (retry_cnt > 0));

            /* Free packet */
            if (work)
                cvm_oct_free_work(work);
            else
                DEBUGPRINT("WARNING: FIX_IPD_PTR_ALIGNMENT get_work() timeout occured.\n");

            gmx_cfg.s.en = 0;
            cvmx_write_csr(CVMX_GMXX_PRTX_CFG(INDEX(FIX_IPD_OUTPORT), INTERFACE(FIX_IPD_OUTPORT)), gmx_cfg.u64);
            cvmx_write_csr(CVMX_ASXX_TX_PRT_EN(INTERFACE(FIX_IPD_OUTPORT)), 0);
            cvmx_write_csr(CVMX_GMXX_RXX_JABBER(INDEX(FIX_IPD_OUTPORT), INTERFACE(FIX_IPD_OUTPORT)), mtu);
            cvmx_write_csr(CVMX_ASXX_PRT_LOOP(INTERFACE(FIX_IPD_OUTPORT)), 0);
            CVMX_SYNC;
        }
        if (num_segs != 1)
            DEBUGPRINT("WARNING: FIX_IPD_PTR_ALIGNMENT failed.\n");
    }
#endif /* FIX_IPD_PTR_ALIGNMENT } */

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    /* Setup forwarding table -- align on cache-line boundary */
    cvm_gbl_ipfwd_cache =
#if 0
        (cvm_ipfwd_flow_bucket_t*) cvmx_bootmem_alloc_named(sizeof(cvm_ipfwd_flow_bucket_t) * CVM_IPFWD_LUT_BUCKETS,
                                                            128, CVM_IPFWD_LUT_NAME);
#else
    (cvm_ipfwd_flow_bucket_t*) cvmx_bootmem_alloc(sizeof(cvm_ipfwd_flow_bucket_t) * CVM_IPFWD_LUT_BUCKETS, 128);
#endif

    if (cvm_gbl_ipfwd_cache == NULL)
        panic("Unable to allocate %ld bytes of memory for IP forwarding cache\n",
                (sizeof(cvm_ipfwd_flow_bucket_t) * CVM_IPFWD_LUT_BUCKETS));

    /* Initialize all entries to zero */
    for (i=0;i<CVM_IPFWD_LUT_BUCKETS;i++) {
        bucket = &cvm_gbl_ipfwd_cache[i];
        spin_lock_init(&bucket->hdr.wr_lock);
        for (j=0;j<CVM_IPFWD_ENTRIES_PER_BUCKET;j++) {
            flow_entry = &bucket->entry[j];
            flow_entry->u.word0 = 0;
            flow_entry->u.word1 = 0;
            flow_entry->u.word2 = 0;
        }
    }
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
    return 0;
}


/**
 * Module / driver shutdown
 *
 * @return Zero on success
 */
static void __exit cvm_oct_cleanup_module(void)
{
    cvm_oct_private_t*  priv;
    int                 port;
    int                 i;
#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    int j;
    cvm_ipfwd_flow_bucket_t *bucket;
    cvm_ipfwd_flow_entry_t *flow_entry;
#endif

    cvmx_ipd_disable();

    del_timer(&cvm_oct_poll_timer);

    /* Disable POW interrupt */
    cvmx_write_csr(CVMX_POW_WQ_INT_THRX(pow_receive_group), 0);

    /* Shutdown all of the tasklets */
    for (i=0; i<NR_CPUS; i++)
        tasklet_kill(cvm_oct_tasklet + i);

    /* Free the interrupt handler */
    free_irq(8 + pow_receive_group, cvm_oct_device);
    free_irq(8 + 46, cvm_oct_device);

#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    /* Flush the ptr fifo */
    cvm_ipfwd_ptr_flush();

    /* Free flow info buffers */
    for (i=0;i<CVM_IPFWD_LUT_BUCKETS;i++) {
        bucket = &cvm_gbl_ipfwd_cache[i];
        for (j=0;j<CVM_IPFWD_ENTRIES_PER_BUCKET;j++) {
            flow_entry = &bucket->entry[j];
            if (flow_entry->s.flow_info)
                cvmx_fpa_free(cvmx_phys_to_ptr((uint64_t)flow_entry->s.flow_info << 7), CVMX_FPA_FLOW_INFO_POOL, 0);
        }
    }

    /* Free flow LUT */
#if 0
    cvmx_bootmem_free_named(CVM_IPFWD_LUT_NAME);
#else
    DEBUGPRINT("ipfwd: Cannot free the flow cache because it was allocated with cvmx_bootmem_alloc()\n");
#endif
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

    /* Free the ethernet devices */
    for (port=0; port<TOTAL_NUMBER_OF_PORTS; port++)
    {
        if (cvm_oct_device[port])
        {
            priv = (cvm_oct_private_t*)cvm_oct_device[port]->priv;

            /* Free buffers */
            while (skb_queue_len(&priv->tx_free_list))
                dev_kfree_skb(skb_dequeue(&priv->tx_free_list));

            unregister_netdev(cvm_oct_device[port]);
            kfree(cvm_oct_device[port]);
            cvm_oct_device[port] = NULL;
        }
    }

    cvmx_pko_shutdown();

    /* Free the HW pools */
#if USE_SKBUFFS_IN_HW
    cvm_oct_free_hw_skbuff(CVMX_FPA_PACKET_POOL, CVMX_FPA_PACKET_POOL_SIZE, NUM_PACKET_BUFFERS);
#else
    cvm_oct_free_hw_memory(CVMX_FPA_PACKET_POOL, CVMX_FPA_PACKET_POOL_SIZE, NUM_PACKET_BUFFERS);
#endif

#ifndef CVM_IPFWD_DRIVER_OFFLOAD
    cvm_oct_free_hw_memory(CVMX_FPA_WQE_POOL, CVMX_FPA_WQE_POOL_SIZE, NUM_PACKET_BUFFERS);
#else
#ifdef CVM_IPFWD_DYNAMIC_SHORT
    cvm_oct_free_hw_skbuff(CVMX_FPA_WQE_POOL, CVMX_FPA_WQE_POOL_SIZE, NUM_PACKET_BUFFERS);
#else
    cvm_oct_free_hw_memory(CVMX_FPA_WQE_POOL, CVMX_FPA_WQE_POOL_SIZE, NUM_PACKET_BUFFERS);
#endif
    /* Free flow info pool */
    cvm_oct_free_hw_memory(CVMX_FPA_FLOW_INFO_POOL, CVMX_FPA_FLOW_INFO_POOL_SIZE, 
	  (CVM_IPFWD_LUT_BUCKETS*CVM_IPFWD_ENTRY_RETRIES));
#endif

#if CVMX_FPA_OUTPUT_BUFFER_POOL != CVMX_FPA_PACKET_POOL
    cvm_oct_free_hw_memory(CVMX_FPA_OUTPUT_BUFFER_POOL, CVMX_FPA_OUTPUT_BUFFER_POOL_SIZE, 128);
#endif

    remove_proc_entry("octeon_ethernet_stats", NULL);
}

/* Ok, this is bad to do. CVMX isn't GPL compatable and is outside of the
    kernel tree. This needs to be improved sometime */
#include "cvmx-spi.c"
#include "cvmx-spi4000.c"
#include "cvmx-pko.c"

MODULE_LICENSE("Proprietary");
MODULE_AUTHOR("Cavium Networks <support@caviumnetworks.com>");
MODULE_DESCRIPTION("Cavium Networks Octeon ethernet driver.");
module_init(cvm_oct_init_module);
module_exit(cvm_oct_cleanup_module);

